/* $Filename:       WormWars/Source/system.h $
 * $VER:            WormWars 5.21 for Amiga $
 * $Description:    Header file for system.c $
 *
 *  Copyright 1999 James R. Jacobs. Freely distributable.

#INCLUDES -------------------------------------------------------------- */

#include <exec/exec.h>
#include <intuition/intuition.h>
#include <intuition/gadgetclass.h>		// STRINGA_ReplaceMode
#include <libraries/asl.h>				// ASL_FileRequest
#include <libraries/gadtools.h>			// struct StringInfo
#include <devices/gameport.h>			// struct GamePortTrigger
#include <dos/dosextens.h>				// struct Process
#include <graphics/displayinfo.h>		// PAL_MONITOR_ID, etc.
#include <exec/memory.h>
#include <devices/audio.h>
#include <dos/dos.h>
#include <libraries/iffparse.h>
#include <utility/date.h>				// struct DateStamp
#include <workbench/workbench.h>		// struct DiskObject
#include <intuition/intuitionbase.h>	// struct IntuitionBase

#include <stdio.h>						// FILE, printf()
#include <stdlib.h>						// EXIT_SUCCESS, EXIT_FAILURE
#include <hardware/cia.h>				// struct CIA

#include "diff.h"
#include "same.h"
#include "libproto.h"

// joystick 1

#define JOYUP					 1
#define JOYDOWN					 2
#define JOYLEFT					 4
#define JOYRIGHT				 8
#define JOYFIRE1				16
#define JOYFIRE2				32
#define POTGOR					*(UWORD *)0xDFF016     

// Easter eggs

#define  FIRSTLISA              "Only Amiga makes it possible!"
#define SECONDLISA              "Xenia rules! :-)"

// statistics text

#define STATSINGLE              "Single"
#define STATTRIPLE              "Triple"
#define STATQUINTUPLE           "Quint."
#define STATSEPTUPLE            "Sept. "
#define STATFAST                "Fast  "
#define STATNORMAL              "Normal"
#define STATSLOW                "Slow  "

#define ALTJUMP					   5
#define LIFEMODULO				1000

/* #DEFINES ---------------------------------------------------------------

Project|About... window */

#define ABOUTLINES			  7
#define ABOUTSHADOW			 DARKGREY
#define ABOUTSHINE			LIGHTGREY
#define ABOUTXPIXEL			312
#define ABOUTYPIXEL			122
#define PROJECTX			128
#define PROJECTY			107

// miscellaneous

#define ANIMDELAY			2000L
#define SQUAREX				9
#define SQUAREY				6
#define BITMAP1WIDTH		(SQUAREX * 16)
#define BITMAP1HEIGHT		(SQUAREY * ((ARRAYSIZE + 1) / 16))
#define BITMAP2WIDTH		(SQUAREX * (FIELDX + 1))
#define BITMAP2HEIGHT		(SQUAREX * (FIELDY + 1))
#define DEPTH				4
#define ISOSQUAREX			9
#define ISOSQUAREY			6
#define QUEUELIMIT			15
#define UNDERLINEOFFSET		(40 + STARTYPIXEL)

// constant strings

#define ASLOPENHAIL     	"Open Fieldset"
#define ASLSAVEHAIL			"Save Fieldset"
#define TITLEBAR        	"Worm Wars 5.21"
#define MEDMODULE			"PROGDIR:WormWars.MED"
#define PATTERN				"(#?.fset)"
#define OLDKICKSTART		"Worm Wars: You need Kickstart R2.04+!\n(Can't open intuition.library V37+!)\n"
#define ALERTTIMERVALUE     "\0\20\20Worm Wars: Can't allocate timer value structure!\0"

// object descriptions (in pixels)

#define FIRSTDESCX			120
#define SECONDDESCX			(SCREENXPIXEL - FIRSTDESCX)
#define DESCY				3

// pseudo-gadgets

#define GADGETX				(-3)
#define GOLDGADGET			(FIELDY / 2 - 12)
#define SILVERGADGET		(FIELDY / 2 -  9)
#define EMPTYGADGET			(FIELDY / 2 -  6)
#define WOODGADGET			(FIELDY / 2 -  3)
#define STONEGADGET			(FIELDY / 2)
#define METALGADGET			(FIELDY / 2 +  3)
#define ONEGADGET			(FIELDY / 2 +  6)
#define TWOGADGET			(FIELDY / 2 +  9)
#define STARTGADGET			(FIELDY / 2 + 12)

// menus

#define PROJECTSTART	0		// array offsets of menu headers
#define LEVELSTART		9

#define MN_PROJECT		0
#define MN_EDIT			1
#define MN_LEVEL		2
#define IN_NEW			0
#define IN_OPEN			1
#define IN_SAVE			2
#define IN_SAVEAS		3
#define IN_ABOUT		5
#define IN_QUIT			7
#define IN_CUT			0
#define IN_COPY			1
#define IN_PASTE		2
#define IN_INSERT		0
#define IN_DELETE		1
#define IN_ERASE		2
#define IN_APPEND		3

// ARRAYS -----------------------------------------------------------------

UWORD chip CustomPointer[] =
{	0x0000, 0x0000,		// reserved

	0xF000, 0xF800,		// 1st row 1st plane, 1st row 2nd plane
	0xF000, 0x8800,		// 2nd row 1st plane, 2nd row 2nd plane
	0xF000, 0x8800,		// 3rd row 1st plane, 3rd row 2nd plane
	0xF000, 0x8800,		// 4th row 1st plane, 4th row 2nd plane
	0xF000, 0x8800,		// 5th row 1st plane, 5th row 2nd plane
	0x8000, 0xF800,		// 6th row 1st plane, 6th row 2nd plane

	0x0000, 0x0000		// reserved
};

STRPTR objectdesc[LASTOBJECT + 1] =
{	"AFFIXER: Stops protectors rotating.",
	"AMMO: 2-6 bullets.",
	"ARMOUR: Immune to most damage.",
    "BIAS: Rules changes in your favour.",
	"BOMB: An explosion.",
	"BONUS: Random letter.",
	"CLOCK: Extra time.",
	"GROWER: Enlarges silver and gold.",
	"HEALER: 100/200 lives.",
	"ICE: Freezes everything except you.",
	"LIFE: 2-6 lives.",
    "LIGHTNING: Flashes squares near tail.",
	"MISSILE: A guided missile.",
	"MULTIPLIER: More points.",
	"NITRO: Diagonal movement.",
    "POWER: Thicker bullets.",
	"PROTECTOR: An orbiting companion.",
	"PULSE: Fragment explosion.",
	"REMNANTS: 2-6 remnants.",
	"SLAYER: Smart bomb.",
	"SLOWER: Slows creatures down.",
	"SWITCHER: Changes tail colours.",
	"TONGUE: Convert tails.",
    "TREASURE: Treasure level.",
	"UMBRELLA: Skip 1-3 levels."
};

UWORD chip IsoImageData[ARRAYSIZE + 1][ISOSQUAREY * 4] =
{ {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// AFFIXER
	0x1FC0,	0x2720,	0x7FC0,	0x6EC0,	0x6100,	0x3E00,
	0x3FE0,	0x2720,	0x7FC0,	0x6EC0,	0xE180,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3BE0,	0x3BE0,	0x7BC0,	0x7BC0,	0xFB80,	0xF180,		// AMMO
	0x0400,	0x0400,	0x0400,	0x0400,	0x0400,	0x0E00,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0400,	0x0000,	0x0400,	0x0000,	0x0400,	0x0000 },
  {	0x1FC0,	0x0F80,	0x0400,	0x3180,	0x6300,	0x0000,		// ARMOUR
	0x2020,	0x3060,	0x7BC0,	0x4E40,	0x9C80,	0xFF80,
	0x3FE0,	0x3FE0,	0x7FC0,	0x4E40,	0xFF80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3F80,	0x3F20,	0x7E40,	0x1CC0,	0xC980,	0xE380,		// BIAS
	0x0060,	0x00C0,	0x0180,	0x6300,	0x3600,	0x1C00,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0060,	0x00C0,	0x0180,	0x6300,	0x3600,	0x1C00 },
  {	0x3BE0,	0x37E0,	0x7FC0,	0x41C0,	0xC180,	0xFF80,		// BOMB
	0x0600,	0x0800,	0x3E00,	0x4100,	0x4100,	0x3E00,
	0x3FE0,	0x37E0,	0x41C0,	0x3EC0,	0xBE80,	0xC180,
	0x0600,	0x0800,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3F80,	0x3F20,	0x7E40,	0x1CC0,	0xC980,	0xE380,		// BONUS
	0x0060,	0x00C0,	0x0180,	0x6300,	0x3600,	0x1C00,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3C20,	0x3FE0,	0x46C0,	0x7CC0,	0xC080,	0xFF80,		// CLOCK
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3C20,	0x03E0,	0x3F40,	0x7F40,	0xBF00,	0xC080,
	0x03C0,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// GROWER
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3020,	0x2DE0,	0x60C0,	0x7B40,	0x8080,	0xF780,
	0x0DC0,	0x1000,	0x1B00,	0x0080,	0x7700,	0x0000 },
  {	0x3FE0,	0x0840,	0x7FC0,	0x4E40,	0xFF80,	0x4200,		// HEALER
	0x0000,	0x28E0,	0x0000,	0x6EC0,	0x0000,	0xE280,
	0x3FE0,	0x3FC0,	0x7FC0,	0x7FC0,	0xFF80,	0x7F80,
	0x0000,	0x2880,	0x0000,	0x2E80,	0x0000,	0x2280 },
  {	0x3FE0,	0x3F60,	0x7E40,	0x7E40,	0xFE80,	0xFF80,		// ICE
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x3FE0,	0x41C0,	0x41C0,	0xC180,	0xFF80,
	0x0000,	0x1F80,	0x0180,	0x0180,	0x0100,	0x0000 },
  {	0x3FE0,	0x2EE0,	0x5540,	0x6AC0,	0xF580,	0xFB80,		// LIFE
	0x0000,	0x1100,	0x3B80,	0x1F00,	0x0E00,	0x0400,
	0x3FE0,	0x2EE0,	0x4440,	0x60C0,	0xF180,	0xFB80,
	0x0000,	0x1100,	0x3B80,	0x1F00,	0x0E00,	0x0400 },
  {	0x3FFF,	0x2FFF,	0x7FDF,	0x7FDF,	0xFF9F,	0xFF9F,		// LIGHTNING
	0x0400,	0x0800, 0x1F00,	0x0200,	0x0400,	0x0000,
	0x3FE0,	0x2FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0400,	0x0800, 0x1F00,	0x0200,	0x0400,	0x0000 },
  {	0x1E00,	0x0CE0,	0x4040,	0x7B00,	0xE700,	0x0F00,		// MISSILE
	0x21E0,	0x3300,	0x3F80,	0x04C0,	0x1880,	0xF080,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x30E0,	0x3F60,	0x7EC0, 0x7DC0,	0xF380,	0xF000,		// MULTIPLIER
	0x0F00,	0x2880,	0x1100,	0x2A00,	0x0C00,	0x0F80,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// NITRO
	0x0000,	0x0000,	0x0A00,	0x0A00,	0x0A00,	0x0A00,
	0x3BE0,	0x31E0,	0x60C0,	0x71C0,	0xF180,	0xF180,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x39E0,	0x30E0,	0x4000,	0x6040,	0xF080,	0xF980,		// POWER
	0x0600,	0x0F00,	0x3FC0,	0x1F80,	0x0F00,	0x0600,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0200,	0x0300,	0x03C0,	0x1C00,	0x0C00,	0x0400 },
  {	0x2020,	0x18C0,	0x0000,	0x1100,	0x9E80,	0xC180,		// PROTECTOR
	0x1FC0,	0x2720,	0x7FC0,	0x6EC0,	0x6100,	0x3E00,
	0x3FE0,	0x2720,	0x7FC0,	0x6EC0,	0xE180,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// PULSE
	0x2440,	0x0000, 0x4000,	0x0040,	0x0000,	0x4480,
	0x3FE0,	0x3FE0,	0x71C0,	0x71C0,	0xFF80,	0xFF80,
	0x2440,	0x0000, 0x4E00, 0x0E40,	0x0000,	0x4480 },
  {	0x3FE0,	0x3FE0,	0x73C0,	0x71C0,	0xF980,	0xFF80,		// REMNANT
	0x0000,	0x0000,	0x0C00,	0x0E00,	0x0600,	0x0000,
	0x3FE0,	0x3FE0, 0x7FC0,	0x7DC0,	0xF980,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x3FE0,	0x71C0,	0x71C0,	0xFF80,	0xFF80,		// SLAYER
	0x2440,	0x0000,	0x4E00,	0x0E40,	0x0000,	0x4480,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x01E0,	0x3060,	0x76C0,	0x6DC0,	0xC180,	0xF000,		// SLOWER
	0x3E00,	0x0F80, 0x0900,	0x1200,	0x3E00,	0x0F80,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0000,	0x0F80, 0x0100,	0x0200,	0x0400,	0x0F80 },
  {	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,		// SWITCHER
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000, 0x0000,	0x0000 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// TONGUE
	0x2020,	0x3060,	0x7BC0,	0x4E40,	0x9C80,	0xFF80,
	0x3FE0,	0x3FE0,	0x7FC0,	0x4E40,	0xFF80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3220,	0x2FE0,	0x64C0,	0x7F40,	0x8880,	0xFF80,		// TREASURE
	0x0DC0,	0x1000,	0x1B00,	0x0080,	0x7700,	0x0000,
	0x3DE0,	0x3DE0,	0x7BC0,	0x7BC0,	0xF780,	0xF780,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x21E0,	0x00E0,	0x0040,	0x77C0,	0xF780,	0xF980,		// UMBRELLA
	0x1E00,	0x3F00,	0x7F80,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x2020,	0x4040,	0x4040,	0x8080,	0xFF80,		// EMPTY
	0x3FC0,	0x3FC0,	0x7F80,	0x7F80,	0xFF00,	0x8000,
	0x0020,	0x0020,	0x0040,	0x0040, 0x0080,	0x7F80,
	0x0000, 0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x2020,	0x4040,	0x4040,	0x8080,	0xFF80,		// SILVER
	0x0020,	0x0020,	0x0040,	0x0040,	0x0080,	0x7F80,
	0x0000,	0x1FC0,	0x3F80,	0x3F80,	0x7F00,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x2020,	0x4040,	0x4040,	0x8080,	0xFF80,		// GOLD
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x3FC0,	0x3FC0,	0x7F80,	0x7F80,	0xFF00,	0x8000 },
  {	0x0000,	0x0440,	0x5DC0,	0x7740,	0x5500,	0xFF80,		// TELEPORT
	0x3FE0,	0x3FE0,	0x7FC0,	0x7740,	0x4400,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0880,	0xAA80,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x3FE0,	0x7FC0, 0x7DC0,	0xF980,	0xFF80,		// FRAGMENT
	0x0000,	0x0000,	0x0000,	0x0200,	0x0600,	0x0000,
	0x3FE0,	0x3FE0,	0x73C0,	0x71C0,	0xF980,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x0000,	0x0E00,	0x1F00,	0x1F80,	0x0F80,	0x0700,		// ORB
	0x3FE0,	0x3FE0,	0x7FC0,	0x7F40,	0xFE00,	0xF880,
	0x0000,	0x0E00,	0x1F00,	0x1F80,	0x0F80,	0x0700,
	0x0000,	0x0E00,	0x1F00,	0x1F00,	0x0E00,	0x0000 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// Green Protector
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x2020,	0x18C0,	0x0000,	0x1100,	0x9E80,	0xC180,
	0x1FC0,	0x2720,	0x7FC0,	0x6EC0,	0x6100,	0x3E00 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// Red Protector
	0x1FC0,	0x2720,	0x7FC0,	0x6EC0,	0x6100,	0x3E00,
	0x2020,	0x18C0,	0x0000,	0x1100,	0x9E80,	0xC180,
	0x1FC0,	0x2720,	0x7FC0,	0x6EC0,	0x6100,	0x3E00 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// Blue Protector
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x1FC0,	0x2720,	0x7FC0,	0x6EC0,	0x6100,	0x3E00 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// Yellow Protector
	0x1FC0,	0x2720,	0x7FC0,	0x6EC0,	0x6100,	0x3E00,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x1FC0,	0x2720,	0x7FC0,	0x6EC0,	0x6100,	0x3E00 },
  {	0x1F80,	0x2FC0,	0x3000,	0x3000,	0x1F80,	0x0F80,		// Green C
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0FC0,	0x1000,	0x1000,	0x0000,	0x0F80,
	0x2060,	0x1020,	0x4FC0,	0x4FC0,	0xE000,	0xF000 },
  {	0x1F00,	0x2F80,	0x30C0,	0x30C0,	0x1F00,	0x0F80,		// Red O
	0x20E0,	0x1060,	0x4F00,	0x4F00,	0xE080,	0xF000,
	0x0000,	0x0F00,	0x1040,	0x1040,	0x0000,	0x0F80,
	0x20E0,	0x1060,	0x4F00,	0x4F00,	0xE080,	0xF000 },
  {	0x1B00,	0x2D80,	0x36C0,	0x32C0,	0x3080,	0x1000,		// Blue M
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x24E0,	0x1B60,	0x5B40,	0x5F40,	0xDF00,	0xFF80,
	0x24E0,	0x1260,	0x4900,	0x4D00,	0xCF00,	0xEF80 },
  {	0x3F00,	0x3F80,	0x3F40,	0x3F80,	0x3000,	0x1000,		// Yellow P
	0x00E0,	0x0060,	0x4080,	0x4040,	0xCF80,	0xEF80,
	0x00E0,	0x1F60,	0x40C0,	0x5FC0,	0xDF80,	0xFF80,
	0x00E0,	0x0060,	0x4080,	0x4040,	0xCF80,	0xEF80 },
  {	0x2000,	0x3000,	0x3000,	0x3000,	0x3F80,	0x1F80,		// Green L
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x1000,	0x1000,	0x1000,	0x0000,	0x1F80,
	0x1FE0,	0x0FE0,	0x4FC0,	0x4FC0,	0xC000,	0xE000 },
  {	0x3F80,	0x3FC0,	0x3C00,	0x3E00,	0x3F80,	0x1F80,		// Red E
	0x0060,	0x0020,	0x43C0,	0x41C0,	0xC000,	0xE000,
	0x0000,	0x1FC0,	0x0000,	0x1E00,	0x0000,	0x1F80,
	0x0060,	0x0020,	0x43C0,	0x41C0,	0xC000,	0xE000 },
  {	0x3F80,	0x1FC0,	0x0600,	0x0600,	0x0600,	0x0200,		// Blue T
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0060,	0x3BC0,	0x7BC0,	0x7BC0,	0xFB80,	0xFF80,
	0x0060,	0x2000,	0x79C0,	0x79C0,	0xF980,	0xFD80 },
  {	0x3F80,	0x3FC0,	0x3C00,	0x3E00,	0x3F00,	0x1F80,		// Yellow E
	0x0060,	0x0000,	0x43C0,	0x41C0,	0xC080,	0xE000,
	0x0060,	0x1FC0,	0x43C0,	0x5FC0,	0xC080,	0xFF80,
	0x0060,	0x0000,	0x43C0,	0x41C0,	0xC080,	0xE000 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// SKULL
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3060,	0x2DA0,	0x4040,	0x4E40,	0xDD80,	0xE380,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x2520,	0x0000,	0x4040,	0x4040,	0x0000,	0x9480,		// SLIME
	0x1240,	0x2AA0,	0x2A80,	0x1500,	0xD500,	0x2A00,
	0x3760,	0x2AA0,	0x6AC0,	0x5540,	0xD500,	0xBE80,
	0x0880,	0x1540,	0x1500,	0x2A80,	0x2A80,	0x4100 },
  {	0x3FC0,	0x3FC0,	0x7F80,	0x7F80,	0xFF00,	0x8000,		// STONE
	0x3FE0,	0x2020,	0x4040,	0x4040,	0x8080,	0xFF80,
	0x0000,	0x1FC0,	0x3F80,	0x3F80,	0x7F00,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  { 0x23E0, 0x23E0, 0x63C0, 0x00C0, 0x6300, 0x8080,     // GOAT
	0x1C00,	0x1C00,	0x1C00,	0x7F00,	0x9C80,	0x7F00,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x1C00,	0x1C00,	0x1C00,	0x7F00,	0x9C80,	0x7F00 },
  {	0x0000,	0x1FC0,	0x3F80,	0x3F80,	0x7F00,	0x0000,		// WOOD
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x3FC0,	0x3FC0,	0x7F80,	0x7F80,	0xFF00,	0x8000,
	0x3FC0,	0x2000,	0x4000,	0x4000,	0x8000,	0x8000 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// Green Tail
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// Red Tail
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// Blue Tail
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0, 0x7FC0,	0xFF80,	0xFF80,		// Yellow Tail
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80 },
  {	0x0000,	0x0E00,	0x1F00,	0x1F80,	0x0F80,	0x0700,		// ORBTONGUE
	0x3FE0,	0x31E0,	0x60C0,	0x6040,	0xF000,	0xF880,		// (blue)
	0x0000,	0x0E00,	0x1F00,	0x1F80,	0x0F80,	0x0700,
	0x0000,	0x0E00,	0x1F00,	0x1F00,	0x0E00,	0x0000 },
  {	0x0000,	0x0000,	0x0000,	0x0080,	0x0180,	0x0700,		// ORBARMOUR
	0x3FE0,	0x3FE0,	0x7FC0,	0x7F40,	0xFE00,	0xF880,		// (red)
	0x0000,	0x0000,	0x0000,	0x0080,	0x0180,	0x0700,
	0x0000,	0x0E00,	0x1F00,	0x1F00,	0x0E00,	0x0000 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7DC0,	0xF980,	0xFF80,		// Green Fire
	0x0000,	0x0000,	0x0000,	0x0200,	0x0600,	0x0000,
	0x3FE0,	0x3FE0,	0x73C0,	0x71C0,	0xF980,	0xFF80,
	0x0000,	0x0000,	0x0C00,	0x0C00,	0x0000,	0x0000 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7DC0,	0xF980,	0xFF80,		// Red Fire
	0x0000,	0x0000,	0x0C00,	0x0E00,	0x0600,	0x0000,
	0x3FE0,	0x3FE0,	0x73C0,	0x71C0,	0xF980,	0xFF80,
	0x0000,	0x0000,	0x0C00,	0x0C00,	0x0000,	0x0000 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7DC0,	0xF980,	0xFF80,		// Blue Fire
	0x0000,	0x0000,	0x0000,	0x0200,	0x0600,	0x0000,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7DC0,	0xF980,	0xFF80,
	0x0000,	0x0000,	0x0C00,	0x0C00,	0x0000,	0x0000 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7DC0,	0xF980,	0xFF80,		// Yellow Fire
	0x0000,	0x0000,	0x0C00,	0x0E00,	0x0600,	0x0000,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7DC0,	0xF980,	0xFF80,
	0x0000,	0x0000,	0x0C00,	0x0C00,	0x0000,	0x0000 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// Green Missile
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x1E00,	0x0CE0,	0x4040,	0x7B00,	0xE700,	0x0F00,
	0x21E0,	0x3300,	0x3F80,	0x04C0,	0x1880,	0xF080 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// Red Missile
	0x21E0,	0x3300,	0x3F80,	0x04C0,	0x1880,	0xF080,
	0x1E00,	0x0CE0,	0x4040,	0x7B00,	0xE700,	0x0F00,
	0x21E0,	0x3300,	0x3F80,	0x04C0,	0x1880,	0xF080 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// Blue Missile
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x21E0,	0x3300,	0x3F80,	0x04C0,	0x1880,	0xF080 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// Yellow Missile
	0x21E0,	0x3300,	0x3F80,	0x04C0,	0x1880,	0xF080,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x21E0,	0x3300,	0x3F80,	0x04C0,	0x1880,	0xF080 },
  {	0x38E0,	0x38E0,	0x0E00,	0x0E00,	0xE380,	0xE380,		// START
	0x0000,	0x0000,	0x01C0,	0x01C0,	0x1C00,	0x1C00,
	0x3FE0,	0x3FE0,	0x0E00,	0x0E00,	0xFF80,	0xFF80,
	0x0700,	0x0700,	0x71C0,	0x71C0,	0x1C00,	0x1C00 },
  {	0x2060,	0x2F60,	0x6040,	0x6F40,	0x8C00,	0x8C00,		// MUSIC
	0x1F80,	0x1080,	0x1F80,	0x1080,	0x7380,	0x7380,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3BE0,	0x3BE0,	0x7BC0,	0x77C0,	0xF480,	0xFB80,		// STICKY
	0x0000,	0x0400,	0x0400,	0x0800,	0x0B00,	0x0400,
	0x3FE0, 0x3BE0,	0x7BC0,	0x77C0,	0xF480,	0xF380,
	0x0000,	0x0400,	0x0400,	0x0800,	0x0800,	0x0000 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// FX
	0x0220,	0x0140,	0x0180,	0x0300,	0x0500,	0x0880,
	0x3DC0,	0x3EA0,	0x7E40,	0x7CC0,	0xFA80,	0xF700,
	0x1A20,	0x2140,	0x4180,	0x7300,	0x8500,	0x8880 },
  {	0x3FE0,	0x38E0,	0x7780,	0x7780,	0xFF80,	0xF800,		// ZERO
	0x0000,	0x0700,	0x0840,	0x0840,	0x0000,	0x0780,
	0x30E0,	0x2860,	0x6700,	0x6700,	0xF080,	0xF800,
	0x0000, 0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x3EE0,	0x7EC0,	0x7EC0,	0xFF80,	0xFC00,		// ONE
	0x0000,	0x0100,	0x0100,	0x0100,	0x0000,	0x0380,
	0x3DE0,	0x38E0,	0x7CC0,	0x7CC0,	0xF880,	0xFC00,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x30E0,	0x7F80,	0x7840,	0xFF80,	0xF000,		// TWO
	0x0000,	0x0F00,	0x0040,	0x0780,	0x0000,	0x0F80,
	0x20E0,	0x3060,	0x7080,	0x6840,	0xE080,	0xF000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x30E0,	0x7F80,	0x7CC0,	0xFF80,	0xF000,		// Three
	0x0000,	0x0F00,	0x0040,	0x0300,	0x0000,	0x0F80,
	0x20E0,	0x3060,	0x7880,	0x7C40,	0xE080,	0xF000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x3760,	0x7FC0,	0x7100,	0xFF00,	0xFF00,		// Four
	0x0000,	0x0880,	0x0000,	0x0EC0,	0x0080,	0x0080,
	0x2EE0,	0x2660,	0x6040,	0x7000,	0xFE00,	0xFF00,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x3020,	0x7FC0,	0x70C0,	0xFF80,	0xF000,		// Five
	0x0000,	0x0FC0,	0x0000,	0x0F00,	0x0000,	0x0F80,
	0x2060,	0x2020,	0x60C0,	0x7040,	0xE080,	0xF000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x3820,	0x7FC0,	0x70C0,	0xFF80,	0xF800,		// Six
	0x0000,	0x07C0,	0x0000,	0x0F00,	0x0000,	0x0780,
	0x3060,	0x2820,	0x60C0,	0x6040,	0xF080,	0xF800,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x30A0,	0x7F80,	0x7F40,	0xFF00,	0xFE80,		// Seven
	0x0000,	0x0F40,	0x0040,	0x0080,	0x0080,	0x0100,
	0x2060,	0x3020,	0x7E80,	0x7E40,	0xFD00,	0xFE80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x38E0,	0x7F80,	0x78C0,	0xFF00,	0xF800,		// Eight
	0x0000,	0x0700,	0x0040,	0x0700,	0x0080,	0x0780,
	0x30E0,	0x2860,	0x7080,	0x6840,	0xF000,	0xF800,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x38A0,	0x7F80,	0x7880,	0xFF80,	0xFF80,		// Nine
	0x0000,	0x0740,	0x0040,	0x0740,	0x0000,	0x0000,
	0x3060,	0x2820,	0x7000,	0x7800,	0xFF00,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// GREENHEADUP
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x38E0,	0x3060,	0x4040,	0x0000,	0x6300,	0x0000,
	0x0700,	0x0F80,	0x3F80,	0x4E40,	0x9C80,	0xFF80 },
  {	0x3FE0, 0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// REDHEADUP
	0x0700,	0x0F80,	0x3F80,	0x4E40,	0x9C80,	0xFF80,
	0x38E0,	0x3060,	0x4040,	0x0000,	0x6300,	0x0000,
	0x0700,	0x0F80,	0x3F80,	0x4E40,	0x9C80,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// BLUEHEADUP
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x3FE0,	0x7FC0,	0x4E40,	0xFF80,	0xFF80,
	0x0700,	0x0F80,	0x3F80,	0x4E40,	0x9C80,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// YELLOWHEADUP
	0x0700,	0x0F80,	0x3F80,	0x4E40,	0x9C80,	0xFF80,
	0x3FE0,	0x3FE0,	0x7FC0,	0x4E40,	0xFF80,	0xFF80,
	0x0700,	0x0F80,	0x3F80,	0x4E40,	0x9C80,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// GREENHEADDOWN
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x18C0,	0x0040,	0x4040,	0xC180,	0xE380,
	0x3FE0,	0x2720,	0x4E00,	0x3F80,	0x3E00,	0x1C00 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// REDHEADDOWN
	0x3FE0,	0x2720,	0x4E40,	0x3F80,	0x3E00,	0x1C00,
	0x0000,	0x18C0,	0x0000,	0x4040,	0xC180,	0xE380,
	0x3FE0,	0x2720,	0x4E40,	0x3F80,	0x3E00,	0x1C00 },
  {	0x3FE0, 0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// BLUEHEADDOWN
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x3FE0,	0x4E40,	0x7FC0,	0xFF80,	0xFF80,
	0x3FE0,	0x2720,	0x4E40,	0x3F80,	0x3E00,	0x1C00 },
  {	0x3FE0, 0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// YELLOWHEADDOWN
	0x3FE0,	0x2720,	0x4E40,	0x3F80,	0x3E00,	0x1C00,
	0x3FE0,	0x3FE0,	0x4E40,	0x7FC0,	0xFF80,	0xFF80,
	0x3FE0,	0x2720,	0x4E40,	0x3F80,	0x3E00,	0x1C00 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// GREENHEADLEFT
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3C00,	0x30C0,	0x0000,	0x0000,	0xC300,	0xF000,
	0x03E0,	0x0C20,	0x7FC0,	0x7FC0,	0x3080,	0x0F80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// REDHEADLEFT
	0x03E0,	0x0C20,	0x7FC0,	0x7FC0,	0x3080,	0x0F80,
	0x3C00,	0x30C0,	0x0000,	0x0000,	0xC300,	0xF000,
	0x03E0,	0x0C20,	0x7FC0,	0x7FC0,	0x3080,	0x0F80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// BLUEHEADLEFT
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x3CE0,	0x7FC0,	0x7FC0,	0xF380,	0xFF80,
	0x03E0,	0x0C20,	0x7FC0,	0x7FC0,	0x3080,	0x0F80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// YELLOWHEADLEFT
	0x03E0,	0x0C20,	0x7FC0,	0x7FC0,	0x3080,	0x0F80,
	0x3FE0,	0x3CE0,	0x7FC0,	0x7FC0,	0xF380,	0xFF80,
	0x03E0,	0x0C20,	0x7FC0,	0x7FC0,	0x3080,	0x0F80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// GREENHEADRIGHT
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x01E0,	0x1860,	0x0000,	0x0000,	0x6180,	0x0780,
	0x3E00,	0x2180,	0x7FC0,	0x7FC0,	0x8600,	0xF800 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// REDHEADRIGHT
	0x3E00,	0x2180,	0x7FC0,	0x7FC0,	0x8600,	0xF800,
	0x01E0,	0x1860,	0x0000,	0x0000,	0x6180,	0x0780,
	0x3E00,	0x2180,	0x7FC0,	0x7FC0,	0x8600,	0xF800 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// BLUEHEADRIGHT
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x39E0,	0x7FC0,	0x7FC0,	0xE780,	0xFF80,
	0x3E00,	0x2180,	0x7FC0,	0x7FC0,	0x8600,	0xF800 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// YELLOWHEADRIGHT
	0x3E00,	0x2180,	0x7FC0,	0x7FC0,	0x8600,	0xF800,
	0x3FE0,	0x39E0,	0x7FC0,	0x7FC0,	0xE780,	0xFF80,
	0x3E00,	0x2180,	0x7FC0,	0x7FC0,	0x8600,	0xF800 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// GREENMODEUP
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x1FC0,	0x0F80,	0x0400,	0x0000,	0x6300,	0x0000,
	0x2020,	0x3060,	0x7BC0,	0x4E40,	0x9C80,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// REDMODEUP
	0x2020,	0x3060,	0x7BC0,	0x4E40,	0x9C80,	0xFF80,
	0x1FC0,	0x0F80,	0x0400,	0x0000,	0x6300,	0x0000,
	0x2020,	0x3060,	0x7BC0,	0x4E40,	0x9C80,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// BLUEMODEUP
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x3FE0,	0x7FC0,	0x4E40,	0xFF80,	0xFF80,
	0x2020,	0x3060,	0x7BC0,	0x4E40,	0x9C80,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// YELLOWMODEUP
	0x2020,	0x3060,	0x7BC0,	0x4E40,	0x9C80,	0xFF80,
	0x3FE0,	0x3FE0,	0x7FC0,	0x4E40,	0xFF80,	0xFF80,
	0x2020,	0x3060,	0x7BC0,	0x4E40,	0x9C80,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// GREENMODEDOWN
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x18C0,	0x0000,	0x0400,	0x3E00,	0x7F00,
	0x3FE0,	0x2720,	0x4E40,	0x7BC0,	0xC180,	0x8080 },
  {	0x3FE0, 0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// REDMODEDOWN
	0x3FE0,	0x2720,	0x4E40,	0x7BC0,	0xC180,	0x8080,
	0x0000,	0x18C0,	0x0000,	0x0400,	0x3E00,	0x7F00,
	0x3FE0,	0x2720,	0x4E40,	0x7BC0,	0xC180,	0x8080 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// BLUEMODEDOWN
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x3FE0,	0x4E40,	0x7FC0,	0xFF80,	0xFF80,
	0x3FE0,	0x2720,	0x4E40,	0x7BC0,	0xC180,	0x8080 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// YELLOWMODEDOWN
	0x3FE0,	0x2720,	0x4E40,	0x7BC0,	0xC180,	0x8080,
	0x3FE0,	0x3FE0,	0x4E40,	0x7FC0,	0xFF80,	0xFF80,
	0x3FE0,	0x2720,	0x4E40,	0x7BC0,	0xC180,	0x8080 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// GREENMODELEFT
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x20C0,	0x7000,	0x7000,	0x8300,	0x0000,
	0x3FE0,	0x1C20,	0x0FC0,	0x0FC0,	0x7080,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// REDMODELEFT
	0x3FE0,	0x1C20,	0x0FC0,	0x0FC0,	0x7080,	0xFF80,
	0x0000,	0x20C0,	0x7000,	0x7000,	0x8300,	0x0000,
	0x3FE0,	0x1C20,	0x0FC0,	0x0FC0,	0x7080,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// BLUEMODELEFT
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x3CE0,	0x7FC0,	0x7FC0,	0xF380,	0xFF80,
	0x3FE0,	0x1C20,	0x0FC0,	0x0FC0,	0x7080,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// YELLOWMODELEFT
	0x3FE0, 0x1C20,	0x0FC0,	0x0FC0,	0x7080,	0xFF80,
	0x3FE0,	0x3CE0,	0x7FC0,	0x7FC0,	0xF380,	0xFF80,
	0x3FE0,	0x1C20,	0x0FC0,	0x0FC0,	0x7080,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// GREENMODERIGHT
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x1820,	0x01C0,	0x01C0,	0x6080,	0x0000,
	0x3FE0,	0x21C0,	0x7E00,	0x7E00,	0x8700,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// REDMODERIGHT
	0x3FE0,	0x21C0,	0x7E00,	0x7E00,	0x8700,	0xFF80,
	0x0000,	0x1820,	0x01C0,	0x01C0,	0x6080,	0x0000,
	0x3FE0,	0x21C0,	0x7E00,	0x7E00,	0x8700,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// BLUEMODERIGHT
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x39E0,	0x7FC0,	0x7FC0,	0xE780,	0xFF80,
	0x3FE0,	0x21C0,	0x7E00,	0x7E00,	0x8700,	0xFF80 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// YELLOWMODERIGHT
	0x3FE0,	0x21C0,	0x7E00,	0x7E00,	0x8700,	0xFF80,
	0x3FE0,	0x39E0,	0x7FC0,	0x7FC0,	0xE780,	0xFF80,
	0x3FE0,	0x21C0,	0x7E00,	0x7E00,	0x8700,	0xFF80 },
  {	0x0000,	0x1FC0,	0x3F80,	0x3F80,	0x7F00,	0x0000,		// METAL
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0020,	0x1FE0,	0x3FC0,	0x3FC0,	0x7F80,	0x7F80 },
  {	0x3BE0,	0x3BE0,	0x71C0,	0x60C0,	0xE080,	0xF180,		// Green Drip
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3BE0,	0x3BE0,	0x71C0,	0x60C0,	0xE080,	0xF180,
	0x0400,	0x0400,	0x0E00,	0x1F00,	0x1F00,	0x0E00 },
  {	0x3BE0,	0x3BE0,	0x71C0,	0x60C0,	0xE080,	0xF180,		// Red Drip
	0x0400,	0x0400,	0x0E00,	0x1F00,	0x1F00,	0x0E00,
	0x3BE0,	0x3BE0,	0x71C0,	0x60C0,	0xE080,	0xF180,
	0x0400,	0x0400,	0x0E00,	0x1F00,	0x1F00,	0x0E00 },
  {	0x3BE0,	0x3BE0,	0x71C0,	0x60C0,	0xE080,	0xF180,		// Blue Drip
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0400,	0x0400,	0x0E00,	0x1F00,	0x1F00,	0x0E00 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// Yellow Drip
	0x0400,	0x0400,	0x0E00,	0x1F00,	0x1F00,	0x0E00,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0400,	0x0400,	0x0E00,	0x1F00,	0x1F00,	0x0E00 },
  {	0x3FE0, 0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// BLACKENED
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FE0,	0x3FE0,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,		// WHITENED
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000, 0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x3FC0,	0x3CC0, 0x07C0,	0x7FC0,	0xE600,	0xCE00,		// PENGUIN
	0x0000,	0x3800,	0x7800,	0x0000,	0x1980,	0x3180,
	0x3C00,	0x3C00,	0x7FC0,	0x7FC0,	0xFF80,	0xFF80,
	0x0400,	0x3C00,	0x7F80,	0x0F80,	0x0000,	0x0000 }
};

UWORD chip ImageData[ARRAYSIZE + 1][SQUAREY * 4] =
{ {	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,		// AFFIXER
	0x7F00,	0x9C80,	0xFF80,	0xDD80,	0x6300,	0x3E00,
	0xFF80,	0x9C80,	0xFF80,	0xDD80,	0xE380,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0xF780,	0xF780,	0xF780,	0xF780,	0xF780,	0xE380,		// AMMO
	0x0800,	0x0800,	0x0800,	0x0800,	0x0800,	0x1C00,
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0x0800,	0x0000,	0x0800,	0x0000,	0x0800,	0x0000 },
  {	0x0000,	0x0000,	0x0000,	0x6300,	0x0000,	0x0000,		// ARMOUR
	0x8080,	0xC180,	0xF780,	0x9C80,	0x9C80,	0xFF80,
	0x8080,	0xC180,	0xF780,	0x9C80,	0x9C80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0xFE00,	0xFE00,	0xFC80,	0x3980,	0x9380,	0xC780,		// BIAS
	0x0180,	0x0180,	0x0300,	0xC600,	0x6C00,	0x3800,
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0x0180,	0x0180,	0x0300,	0xC600,	0x6C00,	0x3800 },
  {	0xFB80,	0xF780,	0xFF80,	0xE380,	0xC180,	0xFF80,		// BOMB
	0x0600,	0x0800,	0x1C00,	0x6300,	0x4100,	0x3E00,
	0xFF80,	0xF780,	0xE380,	0x9C80,	0xBE80,	0xC180,
	0x0600,	0x0800,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,		// BONUS
	0x0180,	0x0180, 0x0300,	0xC600,	0x6C00,	0x3800,
	0x0180,	0x0180,	0x0300,	0xC600,	0x6C00,	0x3800,
	0x0000,	0x0000, 0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x007F,	0x3E7F,	0xC0FF,	0x00FF,	0xC1FF,	0x3E7F,		// CLOCK
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x3800,	0x0700,	0x3E00,	0x0000,
	0x07FF,	0x007F,	0x007F,	0x007F,	0x007F,	0x007F },
  {	0x3F00,	0x4800,	0x3E00,	0x0900,	0x7E00,	0x0800,		// GROWER
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000, 0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x3700,	0x4000,	0x3600, 0x0100,	0x7600,	0x0000 },
  {	0xFF80,	0x2100,	0xFF80,	0x9C80,	0xFF80,	0x4200,		// HEALER
	0x0000,	0xA380,	0x0000,	0xDD80,	0x0000,	0xE280,
	0xFF80,	0xFF00,	0xFF80,	0xFF80,	0xFF80,	0x7F80,
	0x0000,	0xA200,	0x0000,	0x5D00,	0x0000,	0x2280 },
  {	0x0000,	0x3E00,	0x7C00,	0x7C00,	0x7C00,	0x0000,		// ICE
	0x0000,	0x0000, 0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x3F00,	0x0300,	0x0300,	0x0200,	0x0000,
	0x0000,	0x3F00, 0x0300,	0x0300,	0x0200,	0x0000 },
  {	0x007F,	0x637F,	0x777F,	0x1C7F,	0x087F,	0x007F,		// LIFE
	0x6300,	0xF780,	0xFF80,	0x7F00,	0x1C00,	0x0800,
	0x0000,	0x0000, 0x0000,	0x0000,	0x0000,	0x0000,
	0x637F,	0xF7FF,	0xFFFF,	0x7F7F,	0x1C7F,	0x087F },
  {	0x087F,	0x107F,	0x3E7F,	0x047F,	0x087F,	0x007F,		// LIGHTNING
	0x0800,	0x1000,	0x3E00,	0x0400,	0x0800,	0x0000,
	0x0800,	0x1000, 0x3E00,	0x0400,	0x0800,	0x0000,
	0x0800,	0x1000,	0x3E00,	0x0400,	0x0800,	0x0000 },
  {	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,		// MISSILE
	0x8780,	0xCC00, 0x7F00,	0x0980,	0x1880,	0xF080,
	0x8780,	0xCC00,	0x7F00,	0x0980,	0x1880,	0xF080,
	0x0000,	0x0000, 0x0000,	0x0000,	0x0000,	0x0000 },
  {	0xF080,	0xFF00,	0xFE80,	0xFD80,	0xFB80,	0xF000,		// MULTIPLIER
	0x0F00,	0x9080,	0x6100,	0x9200,	0x0400,	0x0F80,
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x0800,	0x1C00,	0x2A00,	0x0800,	0x1C00,	0x0800,		// NITRO
	0x0000,	0x0000, 0x0000,	0x0000,	0x1400,	0x1400,
	0x0000,	0x0000,	0x1400,	0x1400, 0x0000,	0x0000,
	0x0000,	0x0000, 0x0000,	0x0000,	0x0000,	0x0000 },
  {	0xF780,	0xC180,	0x0000,	0x8080,	0xC180,	0xF780,		// POWER
	0x0800,	0x3E00,	0xFF80,	0x7F00,	0x3E00,	0x0800,
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0x0800,	0x0E00,	0x0F80,	0x7800,	0x3800,	0x0800 },
  {	0x0000,	0x6300,	0x0000,	0x2200,	0x1C00,	0x0000,		// PROTECTOR
	0x7F00,	0x9C80,	0xFF80,	0xDD80,	0x6300,	0x3E00,
	0x7F00,	0x9C80,	0xFF80,	0xDD80,	0x6300,	0x3E00,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,		// PULSE
	0x4880,	0x0000, 0x8000,	0x0080,	0x0000,	0x8900,
	0xFF80,	0xFF80,	0xE380,	0xE380,	0xFF80,	0xFF80,
	0x4880,	0x0000, 0x9C00,	0x1C80,	0x0000,	0x8900 },	// REMNANTS
  {	0xFF80,	0xFF80,	0xE780,	0xE380,	0xF380,	0xFF80,
	0x0000,	0x0000,	0x1800,	0x1C00,	0x0C00,	0x0000,
	0xFF80,	0xFF80,	0xFF80,	0xFB80,	0xF380,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0xFF80,	0xFF80,	0xE380,	0xE380,	0xFF80,	0xFF80,		// SLAYER
	0x4880,	0x0000,	0x9C00,	0x1C80,	0x0000,	0x8900,
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,		// SLOWER
	0xF800,	0x1F80, 0x2100,	0x4200,	0xFC00,	0x0F80,
	0xF800,	0x1F80,	0x2100,	0x4200,	0xFC00,	0x0F80,
	0x0000,	0x0F80, 0x0100,	0x0200,	0x0400,	0x0F80 },
  {	0x007F,	0x007F,	0x007F,	0x007F,	0x007F,	0x007F,		// SWITCHER
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0x007F,	0x007F,	0x007F,	0x007F,	0x007F,	0x007F },
  {	0x8080,	0xC180,	0xF780,	0xFF80,	0x9C80,	0xFF80,		// TONGUE
	0x8080,	0xC180,	0xF780,	0x9C80,	0x9C80,	0xFF80,
	0x8080,	0xC180,	0xF780,	0x9C80,	0x9C80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x087F,	0x087F,	0x087F,	0x087F,	0x087F,	0x087F,		// TREASURE
	0x3700,	0x4000, 0x3600,	0x0100,	0x7600,	0x0000,
	0x3700,	0x4000,	0x3600,	0x0100, 0x7600,	0x0000,
	0x007F,	0x007F, 0x007F,	0x007F,	0x007F,	0x007F },
  {	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,		// UMBRELLA
	0x3C00,	0x7E00, 0xFF00,	0x0000,	0x0000,	0x0000,
	0x3C00,	0x7E00,	0xFF00,	0x1000, 0x1000,	0x0C00,
	0x0000,	0x0000, 0x0000,	0x0000,	0x0000,	0x0000 },
  {	0xFF00,	0x8000,	0x8000,	0x8000,	0x8000,	0x8000,		// EMPTY
	0xFF00,	0xFF00, 0xFF00,	0xFF00,	0xFF00,	0x8000,
	0x0000,	0x0000,	0x0000,	0x0000, 0x0000,	0x0000,
	0x0000,	0x0000, 0x0000,	0x0000,	0x0000,	0x0000 },
  {	0xFF80,	0x8080,	0x8080,	0x8080,	0x8080,	0xFF80,		// SILVER
	0x0080,	0x0080, 0x0080,	0x0080,	0x0080,	0x7F80,
	0x0000,	0x7F00,	0x7F00,	0x7F00, 0x7F00,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0xFF80,	0x8080,	0x8080,	0x8080,	0x8080,	0xFF80,		// GOLD
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0xFF80,	0xFF80, 0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0xFF00,	0xFF00,	0xFF00,	0xFF00,	0xFF00,	0x8000 },
  {	0x0000,	0x1100,	0xBB80,	0xEE80,	0x5500,	0xFF80,		// TELEPORT
	0xFF80,	0xFF80,	0xFF80,	0xEE80,	0x4400,	0x0000,
	0x0000,	0x0000,	0x0000,	0x1100,	0xAA80,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x0000,	0x0000,	0x1800,	0x1800,	0x0000,	0x0000,		// FRAGMENT
	0x0000,	0x0000,	0x0000,	0x0400,	0x0C00,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x007F,	0x1C7F,	0x3E7F,	0x3E7F,	0x1C7F,	0x007F,		// ORB (yellow)
	0xFF80,	0xFF80, 0xFF80,	0xFE80,	0xFC80,	0xF180,
	0x0000,	0x1C00,	0x3E00,	0x3E00,	0x1C00,	0x0000,
	0x007F,	0x1C7F, 0x3E7F,	0x3E7F,	0x1C7F,	0x007F },
  {	0x7F00,	0x9C80,	0xFF80,	0xDD80,	0x6300,	0x3E00,		// Green Protector
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x7F00,	0x9C80,	0xFF80,	0xDD80,	0x6300,	0x3E00 },
  {	0x7F00,	0x9C80,	0xFF80,	0xDD80,	0x6300,	0x3E00,		// Red Protector
	0x7F00,	0x9C80,	0xFF80,	0xDD80,	0x6300,	0x3E00,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x7F00,	0x9C80,	0xFF80,	0xDD80,	0x6300,	0x3E00 },
  {	0x7F00,	0x9C80,	0xFF80,	0xDD80,	0x6300,	0x3E00,		// Blue Protector
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x7F00,	0x9C80,	0xFF80,	0xDD80,	0x6300,	0x3E00,
	0x7F00,	0x9C80,	0xFF80,	0xDD80,	0x6300,	0x3E00 },
  {	0x7F00,	0x9C80,	0xFF80,	0xDD80,	0x6300,	0x3E00,		// Yellow Protector
	0x7F00,	0x9C80, 0xFF80,	0xDD80,	0x6300,	0x3E00,
	0x7F00,	0x9C80,	0xFF80,	0xDD80,	0x6300,	0x3E00,
	0x7F00,	0x9C80, 0xFF80,	0xDD80,	0x6300,	0x3E00 },
  {	0x3F00,		// Green C
	0x4000,
	0x4000,
	0x4000,
	0x3F00,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0xC080,
	0xA000,
	0x9F80,
	0x9F80,
	0xC080,
	0xE000 },
  {	0x3E00,		// Red O
	0x4100,
	0x4100,
	0x4100,
	0x3E00,
	0x0000,
	0xC180,
	0xA080,
	0x9E00,
	0x9E00,
	0xC100,
	0xE080,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0xC180,
	0xA080,
	0x9E00,
	0x9E00,
	0xC100,
	0xE080 },
  {	0x3600,		// Blue M
	0x4900,
	0x4900,
	0x4900,
	0x4100,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0xC980,
	0xA480,
	0x9200,
	0x9200,
	0x9A00,
	0xDF00,
	0xC980,
	0xA480,
	0x9200,
	0x9200,
	0x9A00,
	0xDF00 },
  {	0x7E00,		// Yellow P
	0x4100,
	0x7E00,
	0x4000,
	0x4000,
	0x0000,
	0x8180,
	0x8080,
	0x8100,
	0x8080,
	0x9F80,
	0xDF80,
	0x8180,
	0x8080,
	0x8100,
	0x8080,
	0x9F80,
	0xDF80,
	0x8180,
	0x8080,
	0x8100,
	0x8080,
	0x9F80,
	0xDF80 },
  {	0x4000,		// Green L
	0x4000,
	0x4000,
	0x4000,
	0x7F00,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0xBF80,
	0x9F80,
	0x9F80,
	0x9F80,
	0x8080,
	0xC000 },
  {	0x7F00,		// Red E
	0x4000,
	0x7800,
	0x4000,
	0x7F00,
	0x0000,
	0x8080,
	0x8000,
	0x8700,
	0x8300,
	0x8000,
	0xC000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x8080,
	0x8000,
	0x8700,
	0x8300,
	0x8000,
	0xC000 },
  {	0x7F00,		// Blue T
	0x0800,
	0x0800,
	0x0800,
	0x0800,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x8080,
	0xC000,
	0xF380,
	0xF380,
	0xF380,
	0xFB80,
	0x8080,
	0xC000,
	0xF380,
	0xF380,
	0xF380,
	0xFB80 },
  {	0x7F00,		// Yellow E
	0x4000,
	0x7800,
	0x4000,
	0x7F00,
	0x0000,
	0x8080,
	0x8000,
	0x8780,
	0x8380,
	0x8080,
	0xC000,
	0x8080,
	0x8000,
	0x8780,
	0x8380,
	0x8080,
	0xC000,
	0x8080,
	0x8000,
	0x8780,
	0x8380,
	0x8080,
	0xC000 },
  {	0x3E00, 0x4900,	0x7F00,	0x6300,	0x2200,	0x1C00,		// SKULL
	0x0000,	0x0000, 0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000, 0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x0000, 0x0000,	0x0000,	0x0000,	0x0000,	0x0000,		// SLIME
	0x4900,	0xAA80, 0x5500,	0x2A00,	0xD500,	0x2A00,
	0x4900,	0xAA80,	0x5500,	0x2A00,	0xD500,	0x2A00,
	0x2200,	0x5500, 0x2A00,	0x5500,	0x2A80,	0x4100 },
  {	0xFF00,	0x8000,	0x8000,	0x8000,	0x8000,	0x8000,		// STONE
	0xFF80,	0x8080, 0x8080,	0x8080,	0x8080,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000, 0x0000,	0x0000,
	0x0000,	0x0000, 0x0000,	0x0000,	0x0000,	0x0000 },
  { 0xE380, 0xE380, 0xE380, 0x8080, 0x6300, 0x8080,     // GOAT
	0x1C00,	0x1C00,	0x1C00,	0x7F00,	0x9C80,	0x7F00,
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0x1C00,	0x1C00,	0x1C00,	0x7F00,	0x9C80,	0x7F00 },
  {	0x0000,	0x7F00,	0x7F00,	0x7F00,	0x7F00,	0x0000,		// WOOD
	0xFF80,	0xFF80, 0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0xFF00,	0xFF00,	0xFF00,	0xFF00,	0xFF00,	0x8000,
	0xFF00,	0x8000, 0x8000,	0x8000,	0x8000,	0x8000 },
  {	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,		// Green Tail
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80 },
  {	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,		// Red Tail
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80 },
  {	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,		// Blue Tail
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80 },
  {	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,		// Yellow Tail
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80 },
  {	0x007F,	0x1C7F,	0x3E7F,	0x3E7F,	0x1C7F,	0x007F,		// ORBTONGUE (blue)
	0xFF80,	0xE380, 0xC180,	0xC080,	0xE080,	0xF180,
	0x0000,	0x1C00,	0x3E00,	0x3E00,	0x1C00,	0x0000,
	0x007F,	0x1C7F, 0x3E7F,	0x3E7F,	0x1C7F,	0x007F },
  {	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,		// ORBARMOUR (red)
	0xFF80,	0xFF80, 0xFF80,	0xFE80,	0xFC80,	0xF180,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x1C00, 0x3E00,	0x3E00,	0x1C00,	0x0000 },
  {	0x0000, 0x0000,	0x1800,	0x1800,	0x0000,	0x0000,     // Green Fire
    0x0000,	0x0000, 0x0000,	0x0400,	0x0C00,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000, 0x0000,	0x0000,
    0x0000,	0x0000, 0x1800,	0x1800,	0x0000,	0x0000 },
  {	0x0000,	0x0000,	0x1800,	0x1800,	0x0000,	0x0000,     // Red Fire
    0x0000,	0x0000, 0x1800,	0x1C00,	0x0C00,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
    0x0000,	0x0000, 0x1800,	0x1800,	0x0000,	0x0000 },
  {	0x0000,	0x0000,	0x1800,	0x1800,	0x0000,	0x0000,     // Blue Fire
	0x0000,	0x0000, 0x0000,	0x0400,	0x0C00,	0x0000,
    0x0000,	0x0000, 0x1800,	0x1800,	0x0000,	0x0000,
	0x0000,	0x0000,	0x1800,	0x1800,	0x0000,	0x0000 },
  {	0x0000,	0x0000,	0x1800,	0x1800,	0x0000,	0x0000,     // Yellow Missile
    0x0000,	0x0000, 0x1800,	0x1C00,	0x0C00,	0x0000,
	0x0000,	0x0000,	0x1800,	0x1800, 0x0000,	0x0000,
    0x0000,	0x0000, 0x1800,	0x1800,	0x0000,	0x0000 },
  {	0x8780,		// Green Missile
	0xCC00,
	0x7F00,
	0x0980,
	0x1880,
	0xF080,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x8780,
	0xCC00,
	0x7F00,
	0x0980,
	0x1880,
	0xF080 },
  {	0x8780,		// Red Missile
	0xCC00,
	0x7F00,
	0x0980,
	0x1880,
	0xF080,
	0x8780,
	0xCC00,
	0x7F00,
	0x0980,
	0x1880,
	0xF080,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x8780,
	0xCC00,
	0x7F00,
	0x0980,
	0x1880,
	0xF080 },
  {	0x8780,		// Blue Missile
	0xCC00,
	0x7F00,
	0x0980,
	0x1880,
	0xF080,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x8780,
	0xCC00,
	0x7F00,
	0x0980,
	0x1880,
	0xF080,
	0x8780,
	0xCC00,
	0x7F00,
	0x0980,
	0x1880,
	0xF080 },
  {	0x8780,		// Yellow Missile
	0xCC00,
	0x7F00,
	0x0980,
	0x1880,
	0xF080,
	0x8780,
	0xCC00,
	0x7F00,
	0x0980,
	0x1880,
	0xF080,
	0x8780,
	0xCC00,
	0x7F00,
	0x0980,
	0x1880,
	0xF080,
	0x8780,
	0xCC00,
	0x7F00,
	0x0980,
	0x1880,
	0xF080 },
  {	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,		// START
	0x0000,	0x0000, 0x0380,	0x0380,	0x1C00,	0x1C00,
	0x1C00,	0x1C00,	0x0000,	0x0000,	0x1C00,	0x1C00,
	0x1C00,	0x1C00, 0xE380,	0xE380,	0x1C00,	0x1C00 },
  {	0xC000,	0xDF00,	0xC000,	0xDF00,	0x1C00,	0x1C00,		// MUSIC
	0x3F80,	0x2080,	0x3F80,	0x2080,	0xE380,	0xE380,
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x2000,		// STICKY
	0x0000,	0x1000,	0x1000,	0x2000,	0x2700,	0x1800,
	0x1000,	0x0000, 0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x1000,	0x1000,	0x2000,	0x2000,	0x0000 },
  {	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,		// FX
	0x0880,	0x0500,	0x0200,	0x0200,	0x0500,	0x0880,
	0xF700,	0xFA80,	0xFD80,	0xFD80,	0xFA80,	0xF700,
	0x6880,	0x8500,	0x8200,	0xE200,	0x8500,	0x8880 },
  {	0x7C00,		// ZERO
	0x8200,
	0x8200,
	0x8200,
	0x7C00,
	0x0000,
	0x0000,
	0x3C00,
	0x4100,
	0x4100,
	0x0100,
	0x3E00,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000 },
  {	0x1000,		// ONE
	0x3000,
	0x1000,
	0x1000,
	0x3800,
	0x0000,
	0x0000,
	0x0800,
	0x0800,
	0x0800,
	0x0000,
	0x1C00,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000 },
  {	0xFC00,		// TWO
	0x0200,
	0x7C00,
	0x8000,
	0xFE00,
	0x0000,
	0x0000,
	0x7C00,
	0x0100,
	0x3E00,
	0x0000,
	0x7F00,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000 },
  {	0xFC00,		// Three
	0x0200,
	0x1C00,
	0x0200,
	0xFC00,
	0x0000,
	0x0000,
	0x7C00,
	0x0100,
	0x0C00,
	0x0100,
	0x7E00,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000 },
  {	0x8800,		// Four
	0x8800,
	0x8800,
	0xFE00,
	0x0800,
	0x0000,
	0x0000,
	0x4400,
	0x4400,
	0x0000,
	0x7700,
	0x0400,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000 },
  {	0xFE00,		// Five
	0x8000,
	0xFC00,
	0x0200,
	0xFC00,
	0x0000,
	0x0000,
	0x7F00,
	0x0000,
	0x7C00,
	0x0100,
	0x7E00,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000 },
  {	0x7E00,		// Six
	0x8000,
	0xFC00,
	0x8200,
	0x7C00,
	0x0000,
	0x0000,
	0x3F00,
	0x0000,
	0x7C00,
	0x0100,
	0x3E00,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000 },
  {	0xFC00,		// Seven
	0x0400,
	0x0800,
	0x1000,
	0x2000,
	0x0000,
	0x0000,
	0x7A00,
	0x0200,
	0x0400,
	0x0800,
	0x1000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000 },
  {	0x7C00,		// Eight
	0x8200,
	0x7C00,
	0x8200,
	0x7C00,
	0x0000,
	0x0000,
	0x3C00,
	0x0100,
	0x3C00,
	0x0100,
	0x3E00,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000 },
  {	0x7C00,		// Nine
	0x8200,
	0x7E00,
	0x0200,
	0x0200,
	0x0000,
	0x0000,
	0x3C00,
	0x0100,
	0x3D00,
	0x0100,
	0x0100,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000,
	0x0000 },
  {	0x1C00,	0x3E00,	0x7F00,	0xFF80,	0x9C80,	0xFF80,		// GREENHEADUP
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x1C00,	0x3E00,	0x7F00,	0x9C80,	0x9C80,	0xFF80 },
  {	0x1C00,	0x3E00, 0x7F00,	0xFF80,	0x9C80,	0xFF80,		// REDHEADUP
	0x1C00,	0x3E00,	0x7F00,	0x9C80,	0x9C80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x1C00,	0x3E00,	0x7F00,	0x9C80,	0x9C80,	0xFF80 },
  {	0x1C00,	0x3E00,	0x7F00,	0xFF80,	0x9C80,	0xFF80,		// BLUEHEADUP
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x1C00,	0x3E00,	0x7F00,	0x9C80,	0x9C80,	0xFF80,
	0x1C00,	0x3E00,	0x7F00,	0x9C80,	0x9C80,	0xFF80 },
  {	0x1C00,	0x3E00,	0x7F00,	0xFF80,	0x9C80,	0xFF80,		// YELLOWHEADUP
	0x1C00,	0x3E00,	0x7F00,	0x9C80,	0x9C80,	0xFF80,
	0x1C00,	0x3E00,	0x7F00,	0x9C80,	0x9C80,	0xFF80,
	0x1C00,	0x3E00,	0x7F00,	0x9C80,	0x9C80,	0xFF80 },
  {	0xFF80,	0x9C80,	0xFF00,	0x7F00,	0x3E00,	0x1C00,		// GREENHEADDOWN
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0x9C80,	0x9C00,	0x7F00,	0x3E00,	0x1C00 },
  {	0xFF80,	0x9C80,	0xFF80,	0x7F00,	0x3E00,	0x1C00,		// REDHEADDOWN
	0xFF80,	0x9C80,	0x9C80,	0x7F00,	0x3E00,	0x1C00,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0x9C80,	0x9C80,	0x7F00,	0x3E00,	0x1C00 },
  {	0xFF80,	0x9C80, 0xFF80,	0x7F00,	0x3E00,	0x1C00,		// BLUEHEADDOWN
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0x9C80,	0x9C80,	0x7F00,	0x3E00,	0x1C00,
	0xFF80,	0x9C80,	0x9C80,	0x7F00,	0x3E00,	0x1C00 },
  {	0xFF80,	0x9C80,	0xFF80,	0x7F00,	0x3E00,	0x1C00,		// YELLOWHEADDOWN
	0xFF80,	0x9C80,	0x9C80,	0x7F00,	0x3E00,	0x1C00,
	0xFF80,	0x9C80,	0x9C80,	0x7F00,	0x3E00,	0x1C00,
	0xFF80,	0x9C80,	0x9C80,	0x7F00,	0x3E00,	0x1C00 },
  {	0x0F80,	0x3C80,	0xFF80,	0xFF80,	0x3C80,	0x0F80,		// GREENHEADLEFT
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0F80,	0x3080,	0xFF80,	0xFF80,	0x3080,	0x0F80 },
  {	0x0F80,	0x3C80,	0xFF80,	0xFF80,	0x3C80,	0x0F80,		// REDHEADLEFT
	0x0F80,	0x3080,	0xFF80,	0xFF80,	0x3080,	0x0F80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0F80,	0x3080,	0xFF80,	0xFF80,	0x3080,	0x0F80 },
  {	0x0F80,	0x3C80,	0xFF80,	0xFF80,	0x3C80,	0x0F80,		// BLUEHEADLEFT
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0F80,	0x3080,	0xFF80,	0xFF80,	0x3080,	0x0F80,
	0x0F80,	0x3080,	0xFF80,	0xFF80,	0x3080,	0x0F80 },
  {	0x0F80,	0x3C80,	0xFF80,	0xFF80,	0x3C80,	0x0F80,		// YELLOWHEADLEFT
	0x0F80,	0x3080,	0xFF80,	0xFF80,	0x3080,	0x0F80,
	0x0F80,	0x3080,	0xFF80,	0xFF80,	0x3080,	0x0F80,
	0x0F80,	0x3080,	0xFF80,	0xFF80,	0x3080,	0x0F80 },
  {	0xF800,	0x9E00,	0xFF80,	0xFF80,	0x9E00,	0xF800,		// GREENHEADRIGHT
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xF800,	0x8600,	0xFF80,	0xFF80,	0x8600,	0xF800 },
  {	0xF800,	0x9E00,	0xFF80,	0xFF80,	0x9E00,	0xF800,		// REDHEADRIGHT
	0xF800,	0x8600, 0xFF80,	0xFF80,	0x8600,	0xF800,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xF800,	0x8600,	0xFF80,	0xFF80,	0x8600,	0xF800 },
  {	0xF800,	0x9E00,	0xFF80,	0xFF80,	0x9E00,	0xF800,		// BLUEHEADRIGHT
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xF800,	0x8600,	0xFF80,	0xFF80,	0x8600,	0xF800,
	0xF800,	0x8600,	0xFF80,	0xFF80,	0x8600,	0xF800 },
  {	0xF800,	0x9E00,	0xFF80,	0xFF80,	0x9E00,	0xF800,		// YELLOWHEADRIGHT
	0xF800,	0x8600,	0xFF80,	0xFF80,	0x8600,	0xF800,
	0xF800,	0x8600,	0xFF80,	0xFF80,	0x8600,	0xF800,
	0xF800,	0x8600,	0xFF80,	0xFF80,	0x8600,	0xF800 },
  {	0x8080,	0xC180,	0xF780,	0xFF80,	0x9C80,	0xFF80,		// GREENMODEUP
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x8080,	0xC180,	0xF780,	0x9C80,	0x9C80,	0xFF80 },
  {	0x8080,	0xC180,	0xF780,	0xFF80,	0x9C80,	0xFF80,		// REDMODEUP
	0x8080,	0xC180,	0xF780,	0x9C80,	0x9C80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x8080,	0xC180,	0xF780,	0x9C80,	0x9C80,	0xFF80 },
  {	0x8080,	0xC180,	0xF780,	0xFF80,	0x9C80,	0xFF80,		// BLUEMODEUP
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x8080,	0xC180,	0xF780,	0x9C80,	0x9C80,	0xFF80,
	0x8080,	0xC180,	0xF780,	0x9C80,	0x9C80,	0xFF80 },
  {	0x8080,	0xC180,	0xF780,	0xFF80,	0x9C80,	0xFF80,		// YELLOWMODEUP
	0x8080,	0xC180,	0xF780,	0x9C80,	0x9C80,	0xFF80,
	0x8080,	0xC180,	0xF780,	0x9C80,	0x9C80, 0xFF80,
	0x8080,	0xC180,	0xF780,	0x9C80,	0x9C80,	0xFF80 },
  {	0xFF80,	0x9C80,	0xFF80,	0xF780,	0xC180,	0x8080,		// GREENMODEDOWN
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0x9C80,	0x9C80,	0xF780,	0xC180,	0x8080 },
  {	0xFF80,	0x9C80,	0xFF80,	0xF780,	0xC180,	0x8080,		// REDMODEDOWN
	0xFF80,	0x9C80,	0x9C80,	0xF780,	0xC180,	0x8080,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0x9C80,	0x9C80,	0xF780,	0xC180,	0x8080 },
  {	0xFF80,	0x9C80,	0xFF80,	0xF780,	0xC180,	0x8080,		// BLUEMODEDOWN
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0x9C80,	0x9C80,	0xF780,	0xC180,	0x8080,
	0xFF80,	0x9C80,	0x9C80,	0xF780,	0xC180,	0x8080 },
  {	0xFF80,	0x9C80,	0xFF80,	0xF780,	0xC180,	0x8080,		// YELLOWMODEDOWN
	0xFF80,	0x9C80,	0x9C80,	0xF780,	0xC180,	0x8080,
	0xFF80,	0x9C80,	0x9C80,	0xF780,	0xC180,	0x8080,
	0xFF80,	0x9C80,	0x9C80,	0xF780,	0xC180,	0x8080 },
  {	0xFF80, 0x7C80, 0x1F80,	0x1F80,	0x7C80,	0xFF80,		// GREENMODELEFT
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0x7080, 0x1F80,	0x1F80,	0x7080,	0xFF80 },
  {	0xFF80,	0x7C80,	0x1F80,	0x1F80,	0x7C80,	0xFF80,		// REDMODELEFT
	0xFF80,	0x7080,	0x1F80,	0x1F80,	0x7080,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0x7080, 0x1F80,	0x1F80,	0x7080,	0xFF80 },
  {	0xFF80,	0x7C80,	0x1F80,	0x1F80,	0x7C80,	0xFF80,		// BLUEMODELEFT
	0x0000,	0x0000, 0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0x7080,	0x1F80,	0x1F80, 0x7080,	0xFF80,
	0xFF80,	0x7080, 0x1F80,	0x1F80,	0x7080,	0xFF80 },
  {	0xFF80,	0x7C80,	0x1F80,	0x1F80,	0x7C80,	0xFF80,		// YELLOWMODELEFT
	0xFF80,	0x7080, 0x1F80,	0x1F80,	0x7080,	0xFF80,
	0xFF80,	0x7080,	0x1F80,	0x1F80,	0x7080,	0xFF80,
	0xFF80,	0x7080, 0x1F80,	0x1F80,	0x7080,	0xFF80 },
  {	0xFF80,	0x9F00,	0xFC00,	0xFC00,	0x9F00,	0xFF80,		// GREENMODERIGHT
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0x8700, 0xFC00,	0xFC00,	0x8700,	0xFF80 },
  {	0xFF80,	0x9F00,	0xFC00,	0xFC00,	0x9F00,	0xFF80,		// REDMODERIGHT
	0xFF80,	0x8700, 0xFC00,	0xFC00,	0x8700,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0x8700, 0xFC00,	0xFC00,	0x8700,	0xFF80 },
  {	0xFF80,	0x9F00,	0xFC00,	0xFC00,	0x9F00,	0xFF80,		// BLUEMODERIGHT
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0x8700, 0xFC00,	0xFC00,	0x8700,	0xFF80,
	0xFF80,	0x8700,	0xFC00,	0xFC00,	0x8700,	0xFF80 },
  {	0xFF80,	0x9F00,	0xFC00,	0xFC00,	0x9F00,	0xFF80,		// YELLOWMODERIGHT
	0xFF80,	0x8700,	0xFC00,	0xFC00,	0x8700,	0xFF80,
	0xFF80,	0x8700,	0xFC00,	0xFC00,	0x8700,	0xFF80,
	0xFF80,	0x8700,	0xFC00,	0xFC00,	0x8700,	0xFF80 },
  {	0x0000, 0x7F00,	0x7F00,	0x7F00,	0x7F00,	0x0000,		// METAL
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0x0080,	0x7F80,	0x7F80,	0x7F80,	0x7F80,	0x7F80 },
  {	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,		// Green Drip
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0800,	0x0800,	0x1C00,	0x3E00,	0x3E00,	0x1C00 },
  {	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,		// Red Drip
	0x0800,	0x0800,	0x1C00,	0x3E00,	0x3E00,	0x1C00,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0800,	0x0800,	0x1C00,	0x3E00,	0x3E00,	0x1C00 },
  {	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,		// Blue Drip
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0800,	0x0800,	0x1C00,	0x3E00,	0x3E00,	0x1C00,
	0x0800,	0x0800,	0x1C00,	0x3E00,	0x3E00,	0x1C00 },
  {	0x0800,	0x0800,	0x1C00,	0x3E00,	0x3E00,	0x1C00,		// Yellow Drip
	0x0800,	0x0800,	0x1C00,	0x3E00,	0x3E00,	0x1C00,
	0x0800,	0x0800,	0x1C00,	0x3E00,	0x3E00,	0x1C00,
	0x0800,	0x0800,	0x1C00,	0x3E00, 0x3E00,	0x1C00 },
  {	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,		// BLACKENED
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,	0xFF80,		// WHITENED
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000, 0x0000,	0x0000,	0x0000,	0x0000,
	0x0000,	0x0000,	0x0000,	0x0000,	0x0000,	0x0000 },
  {	0x0F80, 0xF980,	0x0F80,	0x1F80,	0x0000,	0x0000,		// PENGUIN
	0x0000,	0xF000,	0xF000,	0x0000,	0x3980,	0x7180,
	0x0800,	0xF800,	0xFF80,	0x1F80,	0x3980,	0x7180,
	0x0800,	0xF800,	0xFF80,	0x1F80,	0x0000,	0x0000 }
};

UWORD chip LogoData[4320] =
{	/* Plane 0 */
	0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFF0,
	0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFF0,
	0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFF0,
	0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFF0,
	0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFF0,
	0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x00C0,0x0000,0x4000,0x0070,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0020,0x0000,0x0030,0x3FF0,
	0xFFC0,0x0040,0x0000,0x2000,0x0060,0x0001,0xFE00,0x0000,0x0000,0x0000,0x0800,0x0010,0x0000,0x0030,0x3FF0,
	0xFFC0,0x0060,0x0000,0x3000,0x00E0,0x001F,0x8060,0x0000,0x007F,0xFFF8,0x0400,0x0018,0x0000,0x0030,0x3FF0,
	0xFFC0,0x0020,0x0000,0x1000,0x01C0,0x00F8,0x0004,0x0000,0x0060,0x0000,0x0200,0x0008,0x0000,0x0030,0x3FF0,
	0xFFC0,0x0030,0x0008,0x1800,0x0180,0x01C0,0x0000,0x0080,0x0060,0x0000,0x0300,0x0004,0x0000,0x0030,0x3FF0,
	0xFFC0,0x0010,0x0008,0x0800,0x0380,0x0380,0x0000,0x0040,0x0060,0x0000,0x0300,0x0206,0x0000,0x2030,0x3FF0,
	0xFFC0,0x0018,0x001C,0x0C00,0x0300,0x0700,0x0000,0x0060,0x0060,0x0000,0x0300,0x0302,0x0000,0x6030,0x3FF0,
	0xFFC0,0x0008,0x0018,0x0400,0x0700,0x0E00,0x0000,0x0020,0x0060,0x0000,0x0700,0x0301,0x0000,0x6030,0x3FF0,
	0xFFC0,0x000C,0x0038,0x0600,0x0600,0x0C00,0x0000,0x0030,0x0060,0x0000,0x0E00,0x0301,0x8000,0xE030,0x3FF0,
	0xFFC0,0x0004,0x0030,0x0200,0x0E00,0x0C00,0x0000,0x0030,0x0000,0x0000,0x7C00,0x0300,0x8000,0xC030,0x3FF0,
	0xFFC0,0x0006,0x0070,0x0100,0x0C00,0x0C00,0x0000,0x0030,0x0000,0x0000,0x6000,0x0300,0xC001,0xC030,0x3FF0,
	0xFFC0,0x0002,0x00E0,0x0180,0x1C00,0x0C00,0x0000,0x0030,0x0000,0x0000,0x0000,0x0300,0x4003,0x8030,0x3FF0,
	0xFFC0,0x0003,0x00C0,0x0080,0x1800,0x0400,0x0000,0x0030,0x007F,0xFFF0,0x0400,0x0300,0x2003,0x0030,0x3FF0,
	0xFFC0,0x0001,0x01C0,0x00C0,0x3800,0x0600,0x0000,0x0070,0x0060,0x0000,0x0200,0x0300,0x3007,0x0030,0x3FF0,
	0xFFC0,0x0001,0x0180,0x0040,0x3000,0x0200,0x0000,0x0060,0x0060,0x0000,0x0300,0x0300,0x100E,0x0030,0x3FF0,
	0xFFC0,0x0000,0x0380,0x0000,0xF000,0x0100,0x0000,0x00E0,0x0060,0x0000,0x0300,0x0300,0x080C,0x0030,0x3FF0,
	0xFFC0,0x0000,0x0300,0x0000,0xC000,0x0080,0x0000,0x01C0,0x0060,0x0000,0x0300,0x0300,0x081C,0x0030,0x3FF0,
	0xFFC0,0x0000,0x0700,0x0001,0xC000,0x0000,0x0000,0x0780,0x0060,0x0000,0x0300,0x0300,0x0038,0x0030,0x3FF0,
	0xFFC0,0x0000,0x0600,0x0003,0x8000,0x0000,0x0000,0x1E00,0x0060,0x0000,0x0300,0x0300,0x0030,0x0030,0x3FF0,
	0xFFC0,0x0000,0x0E00,0x0003,0x0000,0x0000,0x0000,0x7800,0x0060,0x0000,0x0300,0x0300,0x0070,0x0030,0x3FF0,
	0xFFC0,0x0000,0x1C00,0x0007,0x0000,0x0100,0x0003,0xE000,0x0060,0x0000,0x0100,0x0300,0x00E0,0x0030,0x3FF0,
	0xFFC0,0x0003,0xF800,0x00FE,0x0000,0x003C,0x00FF,0x0000,0x0FE0,0x0000,0x7F80,0x7F00,0x1FC0,0x07F0,0x3FF0,
	0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x3000,0x0010,0x0000,0x1C00,0x0003,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x1000,0x0008,0x0000,0x1800,0x0001,0x0000,0x0000,0x0000,0x0200,0x0003,0xF800,0x0000,0x3FF0,
	0xFFC0,0x0000,0x1800,0x000C,0x0000,0x3800,0x0000,0x8000,0x001F,0xFFFE,0x0100,0x003F,0x0180,0x2000,0x3FF0,
	0xFFC0,0x0000,0x0800,0x0004,0x0000,0x7000,0x0000,0xC000,0x0018,0x0000,0x0080,0x0070,0x0000,0x3000,0x3FF0,
	0xFFC0,0x0000,0x0C00,0x0206,0x0000,0x6000,0x00C0,0x4000,0x0018,0x0000,0x00C0,0x00E0,0x0000,0x1000,0x3FF0,
	0xFFC0,0x0000,0x0400,0x0202,0x0000,0xE000,0x01C0,0x2000,0x0018,0x0000,0x00C0,0x00C0,0x0000,0x1800,0x3FF0,
	0xFFC0,0x0000,0x0600,0x0703,0x0000,0xC000,0x0180,0x1000,0x0018,0x0000,0x00C0,0x0040,0x0003,0xF800,0x3FF0,
	0xFFC0,0x0000,0x0200,0x0601,0x0001,0xC000,0x0380,0x1800,0x0018,0x0000,0x01C0,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0300,0x0E01,0x8001,0x8000,0x0700,0x0800,0x0018,0x0000,0x0380,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0100,0x0C00,0x8003,0x8000,0x0600,0x0400,0x0000,0x0000,0x1F00,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0180,0x1C00,0x0003,0x0000,0x0E00,0x0600,0x0000,0x0000,0x1800,0x01C0,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0080,0x3800,0x3007,0x0000,0x1C00,0x0000,0x0000,0x0000,0x0000,0x000F,0xE000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x00C0,0x3000,0x1006,0x0000,0x0000,0x0080,0x001F,0xFFFC,0x0100,0x0000,0x0780,0x1000,0x3FF0,
	0xFFC0,0x0000,0x0040,0x7000,0x180E,0x0000,0x0000,0x0040,0x0018,0x0000,0x0080,0x0000,0x0010,0x0800,0x3FF0,
	0xFFC0,0x0000,0x0040,0x6000,0x080C,0x0000,0x0000,0x0060,0x0018,0x0000,0x00C0,0x0000,0x0000,0x0C00,0x3FF0,
	0xFFC0,0x0000,0x0000,0xE000,0x001C,0x0000,0x7FFF,0x8020,0x0018,0x0000,0x00C0,0x0100,0x0000,0x0C00,0x3FF0,
	0xFFC0,0x0000,0x0000,0xC000,0x0018,0x0000,0xE000,0x0010,0x0018,0x0000,0x00C0,0x0080,0x0000,0x0C00,0x3FF0,
	0xFFC0,0x0000,0x0001,0xC000,0x0038,0x0000,0xC000,0x0018,0x0018,0x0000,0x00C0,0x0000,0x0000,0x1C00,0x3FF0,
	0xFFC0,0x0000,0x0001,0x8000,0x0070,0x0001,0xC000,0x0008,0x0018,0x0000,0x00C0,0x0000,0x0000,0x3800,0x3FF0,
	0xFFC0,0x0000,0x0003,0x8000,0x0060,0x0003,0x8000,0x0004,0x0018,0x0000,0x00C0,0x0000,0x0000,0xF000,0x3FF0,
	0xFFC0,0x0000,0x0007,0x0000,0x00E0,0x0003,0x0000,0x0006,0x0018,0x0000,0x0040,0x0000,0x0007,0xC000,0x3FF0,
	0xFFC0,0x0000,0x00FE,0x0000,0x3FC0,0x00FF,0x0000,0x03FE,0x03F8,0x0000,0x1FE0,0x00F0,0x00FE,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0003,0xFFC0,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FF0,
	0xFFC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FF0,
	0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFC00,
	0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xF9CC,0x001C,0x1B06,0x79FE,0x01E0,0xFFFF,0xFFFF,0xFFFF,0xFFFC,0xFF00,
	0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xF9C8,0xF9C9,0xF272,0x33FE,0x7FFE,0x7FFF,0xFFFF,0xFFFF,0xFFFC,0xFFC0,
	0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFC98,0x385C,0x3272,0x53FE,0x03F9,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFF0,
	0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFC39,0xF93F,0x9272,0x63FF,0xF9E7,0xFFFF,0xFFFF,0xFFFF,0xFFFC,0xFFC0,
	0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFC78,0x0980,0x3306,0x73FE,0x0320,0x7FFF,0xFFFF,0xFFFF,0xFFFC,0xFF00,
	0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFC00,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	/* Plane 1 */
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0003,0xFF00,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x00FF,0xFFFC,0x0000,0x0000,0x0000,0x0001,0xFF80,0x0000,0xFFC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x07FF,0xFFFF,0x8000,0x0000,0x0000,0x0001,0xFFC0,0x0001,0xFFC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x1FFE,0x01FF,0xE000,0x0000,0x0000,0x0001,0xFFE0,0x0003,0xFFC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x7FE0,0x001F,0xF800,0x0000,0x0000,0x0001,0xFFE0,0x0003,0xFFC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0xFF00,0x0003,0xFE00,0x0000,0x0000,0x0001,0xFFF0,0x0007,0xFFC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0001,0xFE00,0x0001,0xFF00,0x0000,0x0000,0x0001,0xFFF8,0x000F,0xFFC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0003,0xFC00,0x0000,0xFF80,0x0000,0x0000,0x0001,0xFDF8,0x000F,0xDFC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0003,0xF800,0x0000,0x7F80,0x0000,0x0000,0x0001,0xFCFC,0x001F,0x9FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0007,0xF000,0x0000,0x3FC0,0x0000,0x0000,0x0001,0xFCFE,0x003F,0x9FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0007,0xF000,0x0000,0x3FC0,0x0000,0x0000,0x0001,0xFC7E,0x003F,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0007,0xF000,0x0000,0x3FC0,0x0000,0x0000,0x0001,0xFC7F,0x007F,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0007,0xF000,0x0000,0x3FC0,0x0000,0x0000,0x0001,0xFC3F,0x007E,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0007,0xF000,0x0000,0x3FC0,0x0000,0x0000,0x0001,0xFC1F,0x80FC,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0007,0xF800,0x0000,0x7FC0,0x0000,0x0000,0x0001,0xFC1F,0xC1FC,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0003,0xF800,0x0000,0x7F80,0x0000,0x0000,0x0001,0xFC0F,0xC1F8,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0003,0xFC00,0x0000,0xFF80,0x0000,0x0000,0x0001,0xFC07,0xE3F0,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0001,0xFE00,0x0001,0xFF00,0x0000,0x0000,0x0001,0xFC07,0xF7F0,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0xFF00,0x0003,0xFE00,0x0000,0x0000,0x0001,0xFC03,0xF7E0,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x7FE0,0x001F,0xF800,0x0000,0x0000,0x0001,0xFC01,0xFFC0,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x1FFE,0x01FF,0xE000,0x0000,0x0000,0x0001,0xFC01,0xFFC0,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x07FF,0xFFFF,0x8000,0x0000,0x0000,0x0001,0xFC00,0xFF80,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x00FF,0xFFFC,0x0000,0x0000,0x0000,0x0001,0xFC00,0x7F00,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0003,0xFF00,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x003F,0xC000,0x1FC0,0x000F,0xF000,0x0000,0x0000,0x0FFF,0xFFFF,0xC000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x001F,0xC000,0x3FE0,0x001F,0xE000,0x0000,0x0000,0x0FFF,0xFFFF,0xF800,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x001F,0xE000,0x7FF0,0x001F,0xE000,0x0000,0x0000,0x0FFF,0xFFFF,0xFC00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x000F,0xE000,0x7FF0,0x001F,0xC000,0x0000,0x0000,0x0FE0,0x0001,0xFE00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0007,0xF000,0xFFF8,0x003F,0x8000,0x0000,0x0000,0x0FE0,0x0000,0xFF00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0007,0xF000,0xFDF8,0x003F,0x8000,0x0000,0x0000,0x0FE0,0x0000,0x7F00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0003,0xF801,0xFDFC,0x007F,0x0000,0x0000,0x0000,0x0FE0,0x0000,0x7F00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0003,0xF801,0xF8FC,0x007F,0x0000,0x0000,0x0000,0x0FE0,0x0000,0x7F00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0001,0xFC03,0xF8FE,0x00FE,0x0000,0x0000,0x0000,0x0FE0,0x0001,0xFE00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0001,0xFC03,0xF07E,0x00FE,0x0000,0x0000,0x0000,0x0FE0,0x001F,0xFC00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0xFE07,0xF07F,0x01FC,0x0000,0x0000,0x0000,0x0FFF,0xFFFF,0xE000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0xFE0F,0xE03F,0xC1FC,0x0000,0x0000,0x0000,0x0FFF,0xFFFF,0xE000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x7F0F,0xC01F,0xC3F8,0x0000,0x0000,0x0000,0x0FFF,0xFFFF,0xFC00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x7F1F,0xC01F,0xE3F8,0x0000,0x0000,0x0000,0x0FE0,0x0003,0xFE00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x3F9F,0x800F,0xE7F0,0x0000,0x0000,0x0000,0x0FE0,0x0000,0xFF00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x3FBF,0x800F,0xF7F0,0x0000,0x0000,0x0000,0x0FE0,0x0000,0x7F00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x1FFF,0x0007,0xFFE0,0x0000,0x0000,0x0000,0x0FE0,0x0000,0x7F00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x1FFF,0x0007,0xFFE0,0x0000,0x0000,0x0000,0x0FE0,0x0000,0x7F00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0FFE,0x0003,0xFFC0,0x0000,0x0000,0x0000,0x0FE0,0x0000,0x7F00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x07FE,0x0003,0xFF80,0x0000,0x0000,0x0000,0x0FE0,0x0000,0x7F00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x07FC,0x0001,0xFF80,0x0000,0x0000,0x0000,0x0FE0,0x0000,0x7F00,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x03F8,0x0000,0xFF00,0x0000,0x0000,0x0000,0x0FE0,0x0000,0x7F80,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	/* Plane 2 */
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0008,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FFF,0xFFFF,0x0001,0xFF80,0x0000,0xFFC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FFF,0xFFFF,0xE001,0xFFC0,0x0001,0xFFC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FFF,0xFFFF,0xF001,0xFFE0,0x0003,0xFFC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3F80,0x0007,0xF801,0xFFE0,0x0003,0xFFC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3F80,0x0003,0xFC01,0xFFF0,0x0007,0xFFC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3F80,0x0001,0xFC01,0xFFF8,0x000F,0xFFC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3F80,0x0001,0xFC01,0xFDF8,0x000F,0xDFC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3F80,0x0001,0xFC01,0xFCFC,0x001F,0x9FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3F80,0x0007,0xF801,0xFCFE,0x003F,0x9FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3F80,0x007F,0xF001,0xFC7E,0x003F,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FFF,0xFFFF,0x8001,0xFC7F,0x007F,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FFF,0xFFFF,0x8001,0xFC3F,0x007E,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3FFF,0xFFFF,0xF001,0xFC1F,0x80FC,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3F80,0x000F,0xF801,0xFC1F,0xC1FC,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3F80,0x0003,0xFC01,0xFC0F,0xC1F8,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3F80,0x0001,0xFC01,0xFC07,0xE3F0,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3F80,0x0001,0xFC01,0xFC07,0xF7F0,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3F80,0x0001,0xFC01,0xFC03,0xF7E0,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3F80,0x0001,0xFC01,0xFC01,0xFFC0,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3F80,0x0001,0xFC01,0xFC01,0xFFC0,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3F80,0x0001,0xFC01,0xFC00,0xFF80,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3F80,0x0001,0xFE01,0xFC00,0x7F00,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x003F,0xC000,0x1FC0,0x000F,0xF000,0x0FFC,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x001F,0xC000,0x3FE0,0x001F,0xE000,0x0FFC,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x001F,0xE000,0x7FF0,0x001F,0xE000,0x1FFE,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x000F,0xE000,0x7FF0,0x001F,0xC000,0x3FFF,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0007,0xF000,0xFFF8,0x003F,0x8000,0x3FFF,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0007,0xF000,0xFDF8,0x003F,0x8000,0x7F3F,0x8000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0003,0xF801,0xFDFC,0x007F,0x0000,0xFE1F,0xC000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0003,0xF801,0xF8FC,0x007F,0x0000,0xFE1F,0xE000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0001,0xFC03,0xF8FE,0x00FE,0x0001,0xFC0F,0xE000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0001,0xFC03,0xF07E,0x00FE,0x0003,0xF807,0xF000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0xFE07,0xF07F,0x01FC,0x0007,0xF807,0xF800,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0xFE0F,0xE03F,0xC1FC,0x0007,0xF003,0xF800,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x7F0F,0xC01F,0xC3F8,0x000F,0xE001,0xFE00,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x7F1F,0xC01F,0xE3F8,0x001F,0xFFFF,0xFF00,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x3F9F,0x800F,0xE7F0,0x001F,0xFFFF,0xFF80,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x3FBF,0x800F,0xF7F0,0x003F,0xFFFF,0xFF80,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x1FFF,0x0007,0xFFE0,0x007F,0x8000,0x7FC0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x1FFF,0x0007,0xFFE0,0x00FF,0x0000,0x3FE0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0FFE,0x0003,0xFFC0,0x00FF,0x0000,0x3FE0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x07FE,0x0003,0xFF80,0x01FE,0x0000,0x1FF0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x07FC,0x0001,0xFF80,0x03FC,0x0000,0x0FF8,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x03F8,0x0000,0xFF00,0x03FC,0x0000,0x0FF8,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x00F8,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0038,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0008,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0018,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0078,
	0x8000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x01F8,
	/* Plane 3 */
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0008,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0003,0xFF00,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0xFF00,0x007F,0x0000,0x3FC0,0x00FF,0xFFFC,0x0000,0x3FFF,0xFFFF,0x0001,0xFF80,0x0000,0xFFC0,0x0000,
	0x0000,0x7F00,0x00FF,0x8000,0xFF80,0x07FF,0xFFFF,0x8000,0x3FFF,0xFFFF,0xE001,0xFFC0,0x0001,0xFFC0,0x0000,
	0x0000,0x7F80,0x01FF,0xC000,0xFF80,0x1FFE,0x01FF,0xE000,0x3FFF,0xFFFF,0xF001,0xFFE0,0x0003,0xFFC0,0x0000,
	0x0000,0x3F80,0x01FF,0xC000,0xFF00,0x7FE0,0x001F,0xF800,0x3F80,0x0007,0xF801,0xFFE0,0x0003,0xFFC0,0x0000,
	0x0000,0x1FC0,0x03FF,0xE001,0xFE00,0xFF00,0x0003,0xFE00,0x3F80,0x0003,0xFC01,0xFFF0,0x0007,0xFFC0,0x0000,
	0x0000,0x1FC0,0x03F7,0xE001,0xFE01,0xFE00,0x0001,0xFF00,0x3F80,0x0001,0xFC01,0xFFF8,0x000F,0xFFC0,0x0000,
	0x0000,0x0FE0,0x07F7,0xF003,0xFC03,0xFC00,0x0000,0xFF80,0x3F80,0x0001,0xFC01,0xFDF8,0x000F,0xDFC0,0x0000,
	0x0000,0x0FE0,0x07E3,0xF003,0xFC03,0xF800,0x0000,0x7F80,0x3F80,0x0001,0xFC01,0xFCFC,0x001F,0x9FC0,0x0000,
	0x0000,0x07F0,0x0FE3,0xF807,0xF807,0xF000,0x0000,0x3FC0,0x3F80,0x0007,0xF801,0xFCFE,0x003F,0x9FC0,0x0000,
	0x0000,0x07F0,0x0FC1,0xF807,0xF807,0xF000,0x0000,0x3FC0,0x3F80,0x007F,0xF001,0xFC7E,0x003F,0x1FC0,0x0000,
	0x0000,0x03F8,0x1FC1,0xFC0F,0xF007,0xF000,0x0000,0x3FC0,0x3FFF,0xFFFF,0x8001,0xFC7F,0x007F,0x1FC0,0x0000,
	0x0000,0x03F8,0x3F80,0xFE0F,0xF007,0xF000,0x0000,0x3FC0,0x3FFF,0xFFFF,0x8001,0xFC3F,0x007E,0x1FC0,0x0000,
	0x0000,0x01FC,0x3F00,0x7E1F,0xE007,0xF000,0x0000,0x3FC0,0x3FFF,0xFFFF,0xF001,0xFC1F,0x80FC,0x1FC0,0x0000,
	0x0000,0x01FC,0x7F00,0x7F1F,0xE007,0xF800,0x0000,0x7FC0,0x3F80,0x000F,0xF801,0xFC1F,0xC1FC,0x1FC0,0x0000,
	0x0000,0x00FE,0x7E00,0x3F3F,0xC003,0xF800,0x0000,0x7F80,0x3F80,0x0003,0xFC01,0xFC0F,0xC1F8,0x1FC0,0x0000,
	0x0000,0x00FE,0xFE00,0x3FBF,0xC003,0xFC00,0x0000,0xFF80,0x3F80,0x0001,0xFC01,0xFC07,0xE3F0,0x1FC0,0x0000,
	0x0000,0x007F,0xFC00,0x1FFF,0x0001,0xFE00,0x0001,0xFF00,0x3F80,0x0001,0xFC01,0xFC07,0xF7F0,0x1FC0,0x0000,
	0x0000,0x007F,0xFC00,0x1FFF,0x0000,0xFF00,0x0003,0xFE00,0x3F80,0x0001,0xFC01,0xFC03,0xF7E0,0x1FC0,0x0000,
	0x0000,0x003F,0xF800,0x0FFE,0x0000,0x7FE0,0x001F,0xF800,0x3F80,0x0001,0xFC01,0xFC01,0xFFC0,0x1FC0,0x0000,
	0x0000,0x001F,0xF800,0x0FFC,0x0000,0x1FFE,0x01FF,0xE000,0x3F80,0x0001,0xFC01,0xFC01,0xFFC0,0x1FC0,0x0000,
	0x0000,0x001F,0xF000,0x07FC,0x0000,0x07FF,0xFFFF,0x8000,0x3F80,0x0001,0xFC01,0xFC00,0xFF80,0x1FC0,0x0000,
	0x0000,0x000F,0xE000,0x03F8,0x0000,0x00FF,0xFFFC,0x0000,0x3F80,0x0001,0xFE01,0xFC00,0x7F00,0x1FC0,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0003,0xFF00,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x000F,0xFE00,0x0000,0x0000,
	0x0000,0x003F,0xC000,0x1FC0,0x000F,0xF000,0x0FFC,0x0000,0x0FFF,0xFFFF,0xC000,0x01FF,0xFFF0,0x0000,0x0000,
	0x0000,0x001F,0xC000,0x3FE0,0x001F,0xE000,0x0FFC,0x0000,0x0FFF,0xFFFF,0xF800,0x07FF,0xFFFE,0x0000,0x0000,
	0x0000,0x001F,0xE000,0x7FF0,0x001F,0xE000,0x1FFE,0x0000,0x0FFF,0xFFFF,0xFC00,0x1FFC,0x07FF,0x8000,0x0000,
	0x0000,0x000F,0xE000,0x7FF0,0x001F,0xC000,0x3FFF,0x0000,0x0FE0,0x0001,0xFE00,0x3FC0,0x007F,0xC000,0x0000,
	0x0000,0x0007,0xF000,0xFFF8,0x003F,0x8000,0x3FFF,0x0000,0x0FE0,0x0000,0xFF00,0x7F80,0x001F,0xC000,0x0000,
	0x0000,0x0007,0xF000,0xFDF8,0x003F,0x8000,0x7F3F,0x8000,0x0FE0,0x0000,0x7F00,0x7F00,0x000F,0xE000,0x0000,
	0x0000,0x0003,0xF801,0xFDFC,0x007F,0x0000,0xFE1F,0xC000,0x0FE0,0x0000,0x7F00,0x7F00,0x000F,0xE000,0x0000,
	0x0000,0x0003,0xF801,0xF8FC,0x007F,0x0000,0xFE1F,0xE000,0x0FE0,0x0000,0x7F00,0x7F80,0x0000,0x0000,0x0000,
	0x0000,0x0001,0xFC03,0xF8FE,0x00FE,0x0001,0xFC0F,0xE000,0x0FE0,0x0001,0xFE00,0x3FF8,0x0000,0x0000,0x0000,
	0x0000,0x0001,0xFC03,0xF07E,0x00FE,0x0003,0xF807,0xF000,0x0FE0,0x001F,0xFC00,0x1FFF,0xFC00,0x0000,0x0000,
	0x0000,0x0000,0xFE07,0xF07F,0x01FC,0x0007,0xF807,0xF800,0x0FFF,0xFFFF,0xE000,0x07FF,0xFFF8,0x0000,0x0000,
	0x0000,0x0000,0xFE0F,0xE03F,0xC1FC,0x0007,0xF003,0xF800,0x0FFF,0xFFFF,0xE000,0x003F,0xFFFF,0x0000,0x0000,
	0x0000,0x0000,0x7F0F,0xC01F,0xC3F8,0x000F,0xE001,0xFE00,0x0FFF,0xFFFF,0xFC00,0x0000,0x1FFF,0xC000,0x0000,
	0x0000,0x0000,0x7F1F,0xC01F,0xE3F8,0x001F,0xFFFF,0xFF00,0x0FE0,0x0003,0xFE00,0x0000,0x007F,0xE000,0x0000,
	0x0000,0x0000,0x3F9F,0x800F,0xE7F0,0x001F,0xFFFF,0xFF80,0x0FE0,0x0000,0xFF00,0x0000,0x000F,0xF000,0x0000,
	0x0000,0x0000,0x3FBF,0x800F,0xF7F0,0x003F,0xFFFF,0xFF80,0x0FE0,0x0000,0x7F01,0xFC00,0x0007,0xF000,0x0000,
	0x0000,0x0000,0x1FFF,0x0007,0xFFE0,0x007F,0x8000,0x7FC0,0x0FE0,0x0000,0x7F01,0xFE00,0x0007,0xF000,0x0000,
	0x0000,0x0000,0x1FFF,0x0007,0xFFE0,0x00FF,0x0000,0x3FE0,0x0FE0,0x0000,0x7F00,0xFF00,0x000F,0xF000,0x0000,
	0x0000,0x0000,0x0FFE,0x0003,0xFFC0,0x00FF,0x0000,0x3FE0,0x0FE0,0x0000,0x7F00,0x7FC0,0x001F,0xE000,0x0000,
	0x0000,0x0000,0x07FE,0x0003,0xFF80,0x01FE,0x0000,0x1FF0,0x0FE0,0x0000,0x7F00,0x3FFC,0x03FF,0xC000,0x0000,
	0x0000,0x0000,0x07FC,0x0001,0xFF80,0x03FC,0x0000,0x0FF8,0x0FE0,0x0000,0x7F00,0x0FFF,0xFFFF,0x0000,0x0000,
	0x0000,0x0000,0x03F8,0x0000,0xFF00,0x03FC,0x0000,0x0FF8,0x0FE0,0x0000,0x7F80,0x03FF,0xFFF8,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x000F,0xFF00,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x00F8,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0038,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0008,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0018,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0078,
	0x8000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x01F8
};

/* About bitmap has 2 blank lines to left and right, and 1 blank line
above and below. */

UWORD chip AboutData[240] =
{	/* Plane 0 */
	0x0000,0x0000,0x0000,
	0x0200,0x0000,0x0000,
	0x1AC0,0x0000,0x0000,
	0x0700,0x0C00,0x0000,
	0x1AC1,0xE600,0x0400,
	0x0203,0xC300,0x0400,
	0x0000,0x0180,0x0400,
	0x0000,0x001F,0xFBE0,
	0x1E00,0x0060,0x1C00,
	0x0000,0x0030,0x6400,
	0x07F8,0x0018,0x0400,
	0x03F0,0x0004,0x0000,
	0x01E0,0x0000,0x0000,
	0x0000,0x0000,0x0000,
	0x0000,0x1000,0x0000,
	0x0000,0x1000,0x0000,
	0x0000,0x1000,0x0000,
	0x0000,0x1000,0x0000,
	0x0000,0x1000,0x0000,
	0x0000,0x0000,0x0000,
	/* Plane 1 */
	0x0000,0x0000,0x0000,
	0x0000,0x3800,0x0000,
	0x0000,0x7C00,0x0000,
	0x0000,0xFE00,0x0000,
	0x0001,0xE700,0x0000,
	0x0003,0xC380,0x00C0,
	0x0007,0x81C0,0x0300,
	0x0000,0x0000,0x0400,
	0x001E,0x0070,0x1C00,
	0x003C,0x0038,0x6400,
	0x0078,0x001C,0x0400,
	0x00F0,0x0006,0x0000,
	0x01E0,0x0000,0x0000,
	0x0000,0x0000,0x0000,
	0x0000,0x11C0,0x0000,
	0x0000,0x1200,0x0000,
	0x0000,0x1460,0x0000,
	0x0000,0x1210,0x0000,
	0x0000,0x11E0,0x0000,
	0x0000,0x0000,0x0000,
	/* Plane 2 */
	0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,
	0x0000,0x0800,0x0000,
	0x0000,0x0400,0x6000,
	0x0007,0x8200,0x1800,
	0x3C0F,0xFFE0,0x07E0,
	0x1E1F,0xE080,0x1800,
	0x003C,0x1C40,0x6000,
	0x0078,0x03A0,0x0000,
	0x00F0,0x0078,0x0000,
	0x01E0,0x000F,0x0000,
	0x0000,0x0000,0x0000,
	0x0600,0x1001,0x8000,
	0x0900,0x1002,0x4000,
	0x1100,0x1004,0x4000,
	0x1F80,0x1003,0xE000,
	0x2080,0x1004,0x2000,
	0x0000,0x0000,0x0000,
	/* Plane 3 */
	0x0000,0x0000,0x0000,
	0x0000,0x3800,0x0000,
	0x0000,0x7800,0x0000,
	0x0000,0xF000,0x0000,
	0x0001,0xE000,0x0400,
	0x0003,0xC000,0x64C0,
	0x0007,0x8000,0x1F00,
	0x3C00,0x0000,0x03E0,
	0x1E1E,0x0000,0x1F00,
	0x0F3C,0x0000,0x64C0,
	0x07F8,0x0000,0x0400,
	0x03F0,0x0000,0x0000,
	0x01E0,0x0000,0x0000,
	0x0000,0x0000,0x0000,
	0x061B,0x11C1,0x8440,
	0x0924,0x9202,0x4A40,
	0x1124,0x9464,0x4A40,
	0x1FA0,0x9213,0xE8C0,
	0x20A0,0x91E4,0x2840,
	0x0000,0x0000,0x0000
};

UWORD chip BackgroundData[10240] =
{	/* Plane 0 */
	0x820F,0x808C,0x0000,0x3380,0x0730,0x0000,0xC407,0xC104,0x107C,0x0460,0x0001,0x9C00,0x3980,0x0001,0x9C00,0x3980,0x0006,0x203E,0x0820,0x83E0,0x2040,0x7C10,0x4107,0xC046,0x0000,0x19C0,0x0203,0xE082,0x083E,0x0230,0x0000,0xCE00,0x1CC0,0x0000,0xCE00,0x1CC0,


0x0003,0x101F,0x0410,0x41F0,
	0x8187,0x8084,0x1000,0x0000,0x0000,0x0020,0x8407,0x8604,0x0C3C,0x0420,0x8000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0104,0x203C,0x3020,0x61E0,0x2040,0x7860,0x40C3,0xC042,0x0800,0x0000,0x0203,0xC302,0x061E,0x0210,0x4000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0082,0x101E,0x1810,0x30F0,
	0x81C7,0x80C0,0x1040,0x0000,0x0000,0x0820,0x0C07,0x8E04,0x0E3C,0x0600,0x8200,0x0000,0x0000,0x4200,0x0000,0x0000,0x4100,0x603C,0x7020,0x71E0,0x30C0,0x78E0,0x40E3,0xC060,0x0820,0x0000,0x0603,0xC702,0x071E,0x0300,0x4100,0x0000,0x0000,0x2100,0x0000,0x0000,


0x2080,0x301E,0x3810,0x38F0,
	0x81E7,0x80C0,0x3040,0x0000,0x0000,0x0830,0x0C07,0x9E04,0x0F3C,0x0601,0x8200,0x0000,0x0000,0x4200,0x0000,0x0000,0x4180,0x603C,0xF020,0x79E0,0x30C0,0x79E0,0x40F3,0xC060,0x1820,0x0000,0x0603,0xCF02,0x079E,0x0300,0xC100,0x0000,0x0000,0x2100,0x0000,0x0000,


0x20C0,0x301E,0x7810,0x3CF0,
	0x80CF,0x00C0,0x3000,0x0000,0x0000,0x0030,0x0C03,0xCC04,0x0678,0x0601,0x8000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0180,0x601E,0x6020,0x33C0,0x30C0,0x3CC0,0x4067,0x8060,0x1800,0x0000,0x0601,0xE602,0x033C,0x0300,0xC000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x00C0,0x300F,0x3010,0x19E0,
	0x808F,0x0088,0x1800,0x0000,0x0000,0x0060,0x4403,0xC404,0x0478,0x0440,0xC000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0302,0x201E,0x2020,0x23C0,0x2040,0x3C40,0x4047,0x8044,0x0C00,0x0000,0x0201,0xE202,0x023C,0x0220,0x6000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0181,0x100F,0x1010,0x11E0,
	0x800F,0x018C,0x1800,0x0040,0x0800,0x0060,0xC603,0xC004,0x0078,0x0C60,0xC000,0x0200,0x4000,0x0000,0x0200,0x4000,0x0306,0x301E,0x0020,0x03C0,0x6060,0x3C00,0x4007,0x80C6,0x0C00,0x0020,0x0301,0xE002,0x003C,0x0630,0x6000,0x0100,0x2000,0x0000,0x0100,0x2000,


0x0183,0x180F,0x0010,0x01E0,
	0xA027,0x438C,0x3800,0xC0C0,0x0C0C,0x0070,0xC70B,0x9015,0x013A,0x1C61,0xC006,0x0600,0x6060,0x0006,0x0600,0x6060,0x0386,0x385C,0x80A8,0x09D0,0xE070,0xB901,0x5013,0xA1C6,0x1C00,0x6060,0x0385,0xC80A,0x809D,0x0E30,0xE003,0x0300,0x3030,0x0003,0x0300,0x3030,


0x01C3,0x1C2E,0x4054,0x04E8,
	0x8013,0x8386,0x1800,0xC000,0x000C,0x0061,0x8707,0x2004,0x009C,0x1C30,0xC006,0x0000,0x0060,0x0006,0x0000,0x0060,0x030C,0x3839,0x0020,0x04E0,0xE070,0x7200,0x4009,0xC1C3,0x0C00,0x6000,0x0383,0x9002,0x004E,0x0E18,0x6003,0x0000,0x0030,0x0003,0x0000,0x0030,


0x0186,0x1C1C,0x8010,0x0270,
	0x8011,0x8386,0x0C00,0xC000,0x000C,0x00C1,0x8706,0x2004,0x008C,0x1C30,0x6006,0x0000,0x0060,0x0006,0x0000,0x0060,0x060C,0x3831,0x0020,0x0460,0xE070,0x6200,0x4008,0xC1C3,0x0600,0x6000,0x0383,0x1002,0x0046,0x0E18,0x3003,0x0000,0x0030,0x0003,0x0000,0x0030,


0x0306,0x1C18,0x8010,0x0230,
	0xB011,0xC784,0x0C00,0x8000,0x0004,0x00C0,0x878E,0x2035,0x808E,0x3C20,0x6004,0x0000,0x0020,0x0004,0x0000,0x0020,0x0604,0x3C71,0x01AC,0x0471,0xE078,0xE203,0x5808,0xE3C2,0x0600,0x4000,0x03C7,0x101A,0xC047,0x1E10,0x3002,0x0000,0x0010,0x0002,0x0000,0x0010,


0x0302,0x1E38,0x80D6,0x0238,
	0x3010,0xCF02,0x1802,0x0000,0x0001,0x0061,0x03CC,0x2031,0x8086,0x7810,0xC010,0x0000,0x0008,0x0010,0x0000,0x0008,0x0308,0x1E61,0x018C,0x0433,0xC03C,0xC203,0x1808,0x6781,0x0C01,0x0000,0x01E6,0x1018,0xC043,0x3C08,0x6008,0x0000,0x0004,0x0008,0x0000,0x0004,


0x0184,0x0F30,0x80C6,0x0219,
	0x1810,0xEF02,0x3883,0x0020,0x1003,0x0471,0x03DC,0x2060,0xC087,0x7811,0xC418,0x0100,0x8018,0x2418,0x0100,0x8018,0x2388,0x1EE1,0x0306,0x043B,0xC03D,0xC206,0x0C08,0x7781,0x1C41,0x8010,0x01EE,0x1030,0x6043,0xBC08,0xE20C,0x0080,0x400C,0x120C,0x0080,0x400C,


0x11C4,0x0F70,0x8183,0x021D,
	0x1810,0x6F02,0x3A83,0x0070,0x3803,0x0571,0x03D8,0x2060,0xC083,0x7811,0xD418,0x0381,0xC018,0x2418,0x0381,0xC018,0x2B88,0x1EC1,0x0306,0x041B,0xC03D,0x8206,0x0C08,0x3781,0x1D41,0x8038,0x01EC,0x1030,0x6041,0xBC08,0xEA0C,0x01C0,0xE00C,0x120C,0x01C0,0xE00C,


0x15C4,0x0F60,0x8183,0x020D,
	0x0010,0x7FC7,0x1A09,0x0070,0x3802,0x4163,0x8FF8,0x2000,0x0083,0xFE38,0xD048,0x0381,0xC012,0x0048,0x0381,0xC012,0x0B1C,0x7FC1,0x0000,0x041F,0xF0FF,0x8200,0x0008,0x3FE3,0x8D04,0x8038,0x07FC,0x1000,0x0041,0xFF1C,0x6824,0x01C0,0xE009,0x0024,0x01C0,0xE009,


0x058E,0x3FE0,0x8000,0x020F,
	0xA010,0x3FD7,0x9E08,0x0070,0x3800,0x41E7,0xAFF0,0x2015,0x0081,0xFEBC,0xF040,0x0381,0xC002,0x0040,0x0381,0xC002,0x0F3D,0x7F81,0x00A8,0x040F,0xF6FF,0x0201,0x5008,0x1FEB,0xCF04,0x0038,0x17F8,0x100A,0x8040,0xFF5E,0x7820,0x01C0,0xE001,0x0020,0x01C0,0xE001,


0x079E,0xBFC0,0x8054,0x0207,
	0x9018,0x3EFF,0xFE22,0x0030,0x3001,0x11FF,0xFDF0,0x6024,0x80C1,0xF7FF,0xF110,0x0181,0x8008,0x8110,0x0181,0x8008,0x8FFF,0xEF83,0x0124,0x060F,0xBFDF,0x0602,0x480C,0x1F7F,0xFF11,0x0018,0x1EF8,0x3012,0x4060,0xFBFF,0xF888,0x00C0,0xC004,0x4088,0x00C0,0xC004,


0x47FF,0xF7C1,0x8092,0x0307,
	0x9000,0x3E7F,0xFE3E,0x0010,0x2001,0xF1FF,0xF9F0,0x0024,0x8001,0xF3FF,0xF1F0,0x0081,0x000F,0x81F0,0x0081,0x000F,0x8FFF,0xCF80,0x0124,0x000F,0x9F9F,0x0002,0x4800,0x1F3F,0xFF1F,0x0008,0x1CF8,0x0012,0x4000,0xF9FF,0xF8F8,0x0040,0x8007,0xC0F8,0x0040,0x8007,


0xC7FF,0xE7C0,0x0092,0x0007,
	0x9800,0x3E3F,0xFE1F,0x0000,0x0003,0xE1FF,0xF1F0,0x0064,0xC001,0xF1FF,0xF0F8,0x0000,0x001F,0x00F8,0x0000,0x001F,0x0FFF,0x8F80,0x0326,0x000F,0x8F1F,0x0006,0x4C00,0x1F1F,0xFF0F,0x8000,0x18F8,0x0032,0x6000,0xF8FF,0xF87C,0x0000,0x000F,0x807C,0x0000,0x000F,


0x87FF,0xC7C0,0x0193,0x0007,
	0x9800,0x1F1D,0xFF1F,0x0000,0x0003,0xE3FE,0xE3E0,0x0064,0xC000,0xF8EF,0xF8F8,0x0000,0x001F,0x18F8,0x0000,0x001F,0x1FF7,0x1F00,0x0326,0x0007,0xC63E,0x0006,0x4C00,0x0F8E,0xFF8F,0x8000,0x11F0,0x0032,0x6000,0x7C77,0xFC7C,0x0000,0x000F,0x8C7C,0x0000,0x000F,


0x8FFB,0x8F80,0x0193,0x0003,
	0x9800,0x7784,0xFF1F,0x0000,0x0003,0xE3FC,0x87B8,0x0064,0xC003,0xBC27,0xF8F8,0x0000,0x001F,0x18F8,0x0000,0x001F,0x1FE4,0x3DC0,0x0326,0x001D,0xE07B,0x8006,0x4C00,0x3BC2,0x7F8F,0x8000,0x03DC,0x0032,0x6001,0xDE13,0xFC7C,0x0000,0x000F,0x8C7C,0x0000,0x000F,


0x8FF2,0x1EE0,0x0193,0x000E,
	0x1800,0xF3E4,0x7E1F,0x0010,0x2003,0xE1F8,0x9F3C,0x0060,0xC007,0x9F23,0xF0F8,0x0081,0x001F,0x00F8,0x0081,0x001F,0x0FC4,0xF9E0,0x0306,0x003C,0xF9F3,0xC006,0x0C00,0x79F2,0x3F0F,0x8008,0x0F9E,0x0030,0x6003,0xCF91,0xF87C,0x0040,0x800F,0x807C,0x0040,0x800F,


0x87E2,0x7CF0,0x0183,0x001E,
	0x0800,0xE1E0,0x3E3E,0x0008,0x4001,0xF1F0,0x1E1C,0x0040,0x4007,0x0F01,0xF1F0,0x0042,0x000F,0x81F0,0x0042,0x000F,0x8F80,0xF0E0,0x0202,0x0038,0x79E1,0xC004,0x0400,0x70F0,0x1F1F,0x0004,0x2F0E,0x0020,0x2003,0x8780,0xF8F8,0x0021,0x0007,0xC0F8,0x0021,0x0007,


0xC7C0,0x7870,0x0101,0x001C,
	0x0880,0x40E0,0x3E3E,0x0008,0x4001,0xF1F0,0x1C08,0x0440,0x4402,0x0701,0xF1F0,0x0042,0x000F,0x81F0,0x0042,0x000F,0x8F80,0xE040,0x2202,0x2010,0x39C0,0x8044,0x0440,0x2070,0x1F1F,0x0004,0x2E04,0x0220,0x2201,0x0380,0xF8F8,0x0021,0x0007,0xC0F8,0x0021,0x0007,


0xC7C0,0x7020,0x1101,0x1008,
	0x8080,0x00C0,0x3F3F,0x0000,0x0003,0xF3F0,0x0C00,0x0404,0x0400,0x0601,0xF9F8,0x0000,0x001F,0x99F8,0x0000,0x001F,0x9F80,0x6000,0x2020,0x2000,0x30C0,0x0040,0x4040,0x0060,0x1F9F,0x8000,0x0600,0x0202,0x0200,0x0300,0xFCFC,0x0000,0x000F,0xCCFC,0x0000,0x000F,


0xCFC0,0x3000,0x1010,0x1000,
	0x8000,0x00C0,0x1F1F,0x0004,0x8003,0xE3E0,0x0C00,0x0004,0x0000,0x0600,0xF8F8,0x0024,0x001F,0x18F8,0x0024,0x001F,0x1F00,0x6000,0x0020,0x0000,0x30C0,0x0000,0x4000,0x0060,0x0F8F,0x8002,0x4600,0x0002,0x0000,0x0300,0x7C7C,0x0012,0x000F,0x8C7C,0x0012,0x000F,


0x8F80,0x3000,0x0010,0x0000,
	0x8000,0x0000,0x0DFF,0x8008,0x4007,0xFEC0,0x0000,0x0004,0x0000,0x0000,0x6FFC,0x0042,0x003F,0xFFFC,0x0042,0x003F,0xF600,0x0000,0x0020,0x0000,0x0000,0x0000,0x4000,0x0000,0x06FF,0xC004,0x2000,0x0002,0x0000,0x0000,0x37FE,0x0021,0x001F,0xFFFE,0x0021,0x001F,


0xFB00,0x0000,0x0010,0x0000,
	0xC000,0x0000,0x007F,0x0000,0x0003,0xF800,0x0000,0x000E,0x0000,0x0000,0x03F8,0x0000,0x001F,0xC3F8,0x0000,0x001F,0xC000,0x0000,0x0070,0x0000,0x0000,0x0000,0xE000,0x0000,0x003F,0x8000,0x0000,0x0007,0x0000,0x0000,0x01FC,0x0000,0x000F,0xE1FC,0x0000,0x000F,


0xE000,0x0000,0x0038,0x0000,
	0xE000,0x0180,0x001F,0x0000,0x0003,0xE000,0x0600,0x001F,0x0000,0x0C00,0x00F8,0x0000,0x001F,0x00F8,0x0000,0x001F,0x0000,0x3000,0x00F8,0x0000,0x6060,0x0001,0xF000,0x00C0,0x000F,0x8000,0x0300,0x000F,0x8000,0x0600,0x007C,0x0000,0x000F,0x807C,0x0000,0x000F,


0x8000,0x1800,0x007C,0x0000,
	0xE000,0x0380,0x000F,0x0010,0x2003,0xC000,0x0700,0x001F,0x0000,0x1C00,0x0078,0x0081,0x001E,0x0078,0x0081,0x001E,0x0000,0x3800,0x00F8,0x0000,0xE070,0x0001,0xF000,0x01C0,0x0007,0x8008,0x0380,0x000F,0x8000,0x0E00,0x003C,0x0040,0x800F,0x003C,0x0040,0x800F,


0x0000,0x1C00,0x007C,0x0000,
	0xE004,0x0300,0x000F,0x8008,0x4007,0xC000,0x0300,0x801F,0x0020,0x1800,0x007C,0x0042,0x003E,0x007C,0x0042,0x003E,0x0000,0x1804,0x00F8,0x0100,0xC030,0x0801,0xF002,0x0180,0x0007,0xC004,0x2180,0x400F,0x8010,0x0C00,0x003E,0x0021,0x001F,0x003E,0x0021,0x001F,


0x0000,0x0C02,0x007C,0x0080,
	0xF004,0x0000,0x000B,0x8008,0x4007,0x4000,0x0000,0x803F,0x8020,0x0000,0x005C,0x0042,0x003A,0x005C,0x0042,0x003A,0x0000,0x0004,0x01FC,0x0100,0x0000,0x0803,0xF802,0x0000,0x0005,0xC004,0x2000,0x401F,0xC010,0x0000,0x002E,0x0021,0x001D,0x002E,0x0021,0x001D,


0x0000,0x0002,0x00FE,0x0080,
	0xB00E,0x0000,0x0003,0x8000,0x0007,0x0000,0x0001,0xC035,0x8070,0x0000,0x001C,0x0000,0x0038,0x001C,0x0000,0x0038,0x0000,0x000E,0x01AC,0x0380,0x0000,0x1C03,0x5807,0x0000,0x0001,0xC000,0x0000,0xE01A,0xC038,0x0000,0x000E,0x0000,0x001C,0x000E,0x0000,0x001C,


0x0000,0x0007,0x00D6,0x01C0,
	0x180C,0x0000,0x0021,0x8010,0x2006,0x1000,0x0000,0xC060,0xC060,0x0000,0x010C,0x0081,0x0030,0x810C,0x0081,0x0030,0x8000,0x0006,0x0306,0x0300,0x0000,0x0C06,0x0C06,0x0000,0x0010,0xC008,0x0000,0x6030,0x6030,0x0000,0x0086,0x0040,0x8018,0x4086,0x0040,0x8018,


0x4000,0x0003,0x0183,0x0180,
	0x183E,0x0000,0x0039,0xC070,0x380E,0x7000,0x0001,0xF060,0xC1F0,0x0000,0x01CE,0x0381,0xC073,0x81CE,0x0381,0xC073,0x8000,0x000F,0x8306,0x0F80,0x0000,0x1F06,0x0C1F,0x0000,0x001C,0xE038,0x0000,0xF830,0x60F8,0x0000,0x00E7,0x01C0,0xE039,0xC0E7,0x01C0,0xE039,


0xC000,0x0007,0xC183,0x07C0,
	0x083F,0x8000,0x001F,0xC020,0x100F,0xE000,0x0007,0xF040,0x41FC,0x0000,0x00FE,0x0100,0x807F,0x00FE,0x0100,0x807F,0x0000,0x003F,0x8202,0x0FE0,0x0000,0x7F04,0x041F,0xC000,0x000F,0xE010,0x0003,0xF820,0x20FE,0x0000,0x007F,0x0080,0x403F,0x807F,0x0080,0x403F,


0x8000,0x001F,0xC101,0x07F0,
	0x007F,0xA000,0x0007,0xE000,0x001F,0x8000,0x0017,0xF800,0x03FD,0x0000,0x003F,0x0000,0x00FC,0x003F,0x0000,0x00FC,0x0000,0x00BF,0xC000,0x1FE8,0x0001,0x7F80,0x003F,0xD000,0x0003,0xF000,0x000B,0xFC00,0x01FE,0x8000,0x001F,0x8000,0x007E,0x001F,0x8000,0x007E,


0x0000,0x005F,0xE000,0x0FF4,
	0x0053,0xE000,0x0001,0xE180,0x061E,0x0000,0x001F,0x2800,0x029F,0x0000,0x000F,0x0C00,0x30F0,0x000F,0x0C00,0x30F0,0x0000,0x00F9,0x4000,0x14F8,0x0001,0xF280,0x0029,0xF000,0x0000,0xF0C0,0x000F,0x9400,0x014F,0x8000,0x0007,0x8600,0x1878,0x0007,0x8600,0x1878,


0x0000,0x007C,0xA000,0x0A7C,
	0x0001,0xE000,0x0000,0xE600,0x019C,0x0000,0x001E,0x0000,0x000F,0x0000,0x0007,0x3000,0x0CE0,0x0007,0x3000,0x0CE0,0x0000,0x00F0,0x0000,0x0078,0x0001,0xE000,0x0000,0xF000,0x0000,0x7300,0x000F,0x0000,0x0007,0x8000,0x0003,0x9800,0x0670,0x0003,0x9800,0x0670,


0x0000,0x0078,0x0000,0x003C,
	0x0003,0xC000,0x0000,0xC600,0x018C,0x0000,0x000F,0x0000,0x001E,0x0000,0x0006,0x3000,0x0C60,0x0006,0x3000,0x0C60,0x0000,0x0078,0x0000,0x00F0,0x0000,0xF000,0x0001,0xE000,0x0000,0x6300,0x0007,0x8000,0x000F,0x0000,0x0003,0x1800,0x0630,0x0003,0x1800,0x0630,


0x0000,0x003C,0x0000,0x0078,
	0x0000,0xE000,0x0000,0x6600,0x0198,0x0000,0x001C,0x0000,0x0007,0x0000,0x0003,0x3000,0x0CC0,0x0003,0x3000,0x0CC0,0x0000,0x00E0,0x0000,0x0038,0x0001,0xC000,0x0000,0x7000,0x0000,0x3300,0x000E,0x0000,0x0003,0x8000,0x0001,0x9800,0x0660,0x0001,0x9800,0x0660,


0x0000,0x0070,0x0000,0x001C,
	0x8000,0x0000,0x0000,0x6600,0x0198,0x0000,0x0000,0x0004,0x0000,0x0000,0x0003,0x3000,0x0CC0,0x0003,0x3000,0x0CC0,0x0000,0x0000,0x0020,0x0000,0x0000,0x0000,0x4000,0x0000,0x0000,0x3300,0x0000,0x0002,0x0000,0x0000,0x0001,0x9800,0x0660,0x0001,0x9800,0x0660,


0x0000,0x0000,0x0010,0x0000,
	0x8000,0x0800,0x6000,0x2E00,0x01D0,0x0018,0x0040,0x0004,0x0000,0x4003,0x0001,0x7000,0x0E80,0x0001,0x7000,0x0E80,0x00C0,0x0200,0x0020,0x0002,0x0004,0x0000,0x4000,0x0400,0x3000,0x1700,0x0020,0x0002,0x0000,0x2001,0x8000,0xB800,0x0740,0x0000,0xB800,0x0740,


0x0060,0x0100,0x0010,0x0001,
	0x0000,0x0000,0x6000,0x1E00,0x01E0,0x0018,0x0000,0x0000,0x0000,0x0003,0x0000,0xF000,0x0F00,0x0000,0xF000,0x0F00,0x00C0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3000,0x0F00,0x0000,0x0000,0x0000,0x0001,0x8000,0x7800,0x0780,0x0000,0x7800,0x0780,


0x0060,0x0000,0x0000,0x0000,
	0x0000,0x0010,0x0000,0x1E00,0x01E0,0x0000,0x2000,0x0000,0x0000,0x0080,0x0000,0xF000,0x0F00,0x0000,0xF000,0x0F00,0x0001,0x0000,0x0000,0x0000,0x0600,0x0000,0x0000,0x0008,0x0000,0x0F00,0x1000,0x0000,0x0000,0x0040,0x0000,0x7800,0x0780,0x0000,0x7800,0x0780,


0x0000,0x8000,0x0000,0x0000,
	0x0180,0x0018,0x0000,0x1E00,0x01E0,0x0000,0x6000,0x0600,0x0C00,0x00C0,0x0000,0xF000,0x0F00,0x0000,0xF000,0x0F00,0x0003,0x0000,0x3000,0x6000,0x0600,0x0060,0x00C0,0x000C,0x0000,0x0F00,0x1000,0x0300,0x0600,0x0060,0x0000,0x7800,0x0780,0x0000,0x7800,0x0780,


0x0001,0x8000,0x1800,0x3000,
	0x0080,0x0010,0x0000,0x1E00,0x01E0,0x0000,0x2000,0x0400,0x0400,0x0080,0x0000,0xF000,0x0F00,0x0000,0xF000,0x0F00,0x0001,0x0000,0x2000,0x2000,0x0600,0x0040,0x0040,0x0008,0x0000,0x0F00,0x1000,0x0200,0x0200,0x0040,0x0000,0x7800,0x0780,0x0000,0x7800,0x0780,


0x0000,0x8000,0x1000,0x1000,
	0x8080,0x0030,0x0000,0x0E00,0x01C0,0x0000,0x3000,0x0404,0x0400,0x0180,0x0000,0x7000,0x0E00,0x0000,0x7000,0x0E00,0x0001,0x8000,0x2020,0x2000,0x0F00,0x0040,0x4040,0x0018,0x0000,0x0700,0x1800,0x0202,0x0200,0x00C0,0x0000,0x3800,0x0700,0x0000,0x3800,0x0700,


0x0000,0xC000,0x1010,0x1000,
	0x8080,0x0400,0x0000,0x0F00,0x03C0,0x0000,0x0080,0x0404,0x0400,0x2000,0x0000,0x7800,0x1E00,0x0000,0x7800,0x1E00,0x0000,0x0400,0x2020,0x2001,0x0008,0x0040,0x4040,0x0200,0x0000,0x0780,0x0040,0x0202,0x0200,0x1000,0x0000,0x3C00,0x0F00,0x0000,0x3C00,0x0F00,


0x0000,0x0200,0x1010,0x1000,
	0x8000,0x0000,0x0000,0x0F00,0x03C0,0x0000,0x0000,0x0004,0x0000,0x0000,0x0000,0x7800,0x1E00,0x0000,0x7800,0x1E00,0x0000,0x0000,0x0020,0x0000,0x0000,0x0000,0x4000,0x0000,0x0000,0x0780,0x0000,0x0002,0x0000,0x0000,0x0000,0x3C00,0x0F00,0x0000,0x3C00,0x0F00,


0x0000,0x0000,0x0010,0x0000,
	0x8000,0x0000,0x0000,0x0F80,0x07C0,0x0000,0x0000,0x0004,0x0000,0x0000,0x0000,0x7C00,0x3E00,0x0000,0x7C00,0x3E00,0x0000,0x0000,0x0020,0x0000,0x0000,0x0000,0x4000,0x0000,0x0000,0x07C0,0x0000,0x0002,0x0000,0x0000,0x0000,0x3E00,0x1F00,0x0000,0x3E00,0x1F00,


0x0000,0x0000,0x0010,0x0000,
	0x8000,0x1100,0x0000,0x0F80,0x07C0,0x0000,0x0220,0x0004,0x0000,0x8800,0x0000,0x7C00,0x3E00,0x0000,0x7C00,0x3E00,0x0000,0x1100,0x0020,0x0004,0x4022,0x0000,0x4000,0x0880,0x0000,0x07C0,0x0110,0x0002,0x0000,0x4400,0x0000,0x3E00,0x1F00,0x0000,0x3E00,0x1F00,


0x0000,0x0880,0x0010,0x0002,
	0x8000,0x137E,0x0000,0x0880,0x0440,0x0001,0xFB20,0x0004,0x0000,0x9BF0,0x0000,0x4400,0x2200,0x0000,0x4400,0x2200,0x000F,0xD900,0x0020,0x0004,0xDFB2,0x0000,0x4000,0x09BF,0x0000,0x0440,0x1D90,0x0002,0x0000,0x4DF8,0x0000,0x2200,0x1100,0x0000,0x2200,0x1100,


0x0007,0xEC80,0x0010,0x0002,
	0x8000,0x3206,0x0000,0x0000,0x0000,0x0001,0x8130,0x0004,0x0001,0x9030,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x000C,0x0980,0x0020,0x000C,0x8013,0x0000,0x4000,0x1903,0x0000,0x0000,0x0098,0x0002,0x0000,0xC818,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0006,0x04C0,0x0010,0x0006,
	0x0000,0x0300,0x0000,0x0000,0x0000,0x0000,0x0300,0x0000,0x0000,0x1800,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x1800,0x0000,0x0000,0xC030,0x0000,0x0000,0x0180,0x0000,0x0000,0x0180,0x0000,0x0000,0x0C00,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0000,0x0C00,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0400,0x0080,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x2000,0x0400,0x0000,0x2000,0x0400,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0200,0x0000,0x0000,0x0000,0x0000,0x0000,0x1000,0x0200,0x0000,0x1000,0x0200,


0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0080,0x0100,0x0200,0x0400,0x0000,0x0000,0x0000,0x0000,0x0400,0x0800,0x1000,0x2400,0x0800,0x1000,0x2000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0040,0x0080,0x0000,0x0000,0x0000,0x0000,0x0200,0x0400,0x0800,0x1200,0x0400,0x0800,


0x1000,0x0000,0x0000,0x0000,
	0x8000,0x1800,0x0020,0x0108,0x4200,0x1000,0x0060,0x0004,0x0000,0xC000,0x0100,0x0842,0x1000,0x8100,0x0842,0x1000,0x8000,0x0300,0x0020,0x0006,0x0006,0x0000,0x4000,0x0C00,0x0010,0x0084,0x2030,0x0002,0x0000,0x6000,0x0080,0x0421,0x0800,0x4080,0x0421,0x0800,


0x4000,0x0180,0x0010,0x0003,
	0x8002,0x1800,0x0000,0x0018,0x6000,0x0000,0x0061,0x0004,0x0010,0xC000,0x0000,0x00C3,0x0000,0x0000,0x00C3,0x0000,0x0000,0x0308,0x0020,0x0086,0x0006,0x1000,0x4001,0x0C00,0x0000,0x000C,0x2030,0x8002,0x0008,0x6000,0x0000,0x0061,0x8000,0x0000,0x0061,0x8000,


0x0000,0x0184,0x0010,0x0043,
	0x8000,0x1800,0x0000,0x0018,0x6000,0x0000,0x0060,0x0004,0x0000,0xC000,0x0000,0x00C3,0x0000,0x0000,0x00C3,0x0000,0x0000,0x0300,0x0020,0x0006,0x0006,0x0000,0x4000,0x0C00,0x0000,0x000C,0x2030,0x0002,0x0000,0x6000,0x0000,0x0061,0x8000,0x0000,0x0061,0x8000,


0x0000,0x0180,0x0010,0x0003,
	0x8000,0x1A00,0x0008,0x413C,0xF208,0x4000,0x0160,0x0004,0x0000,0xD000,0x0042,0x09E7,0x9042,0x0042,0x09E7,0x9042,0x0000,0x0B00,0x0020,0x0006,0x8016,0x0000,0x4000,0x0D00,0x0004,0x209E,0x60B0,0x0002,0x0000,0x6800,0x0021,0x04F3,0xC821,0x0021,0x04F3,0xC821,


0x0000,0x0580,0x0010,0x0003,
	0x0001,0xF800,0x0000,0xFC3C,0xF0FC,0x0000,0x007E,0x0000,0x000F,0xC000,0x0007,0xE1E7,0x87E0,0x0007,0xE1E7,0x87E0,0x0000,0x03F0,0x0000,0x007E,0x0007,0xE000,0x0000,0xFC00,0x0000,0x7E1E,0x603F,0x0000,0x0007,0xE000,0x0003,0xF0F3,0xC3F0,0x0003,0xF0F3,0xC3F0,


0x0000,0x01F8,0x0000,0x003F,
	0x0001,0xF000,0x001D,0xFC3F,0xF0FE,0xE000,0x003E,0x0000,0x000F,0x8000,0x00EF,0xE1FF,0x87F7,0x00EF,0xE1FF,0x87F7,0x0000,0x01F0,0x0000,0x007C,0x0003,0xE000,0x0000,0xF800,0x000E,0xFE1F,0xE01F,0x0000,0x0007,0xC000,0x0077,0xF0FF,0xC3FB,0x8077,0xF0FF,0xC3FB,


0x8000,0x00F8,0x0000,0x003E,
	0x8000,0x7000,0x003F,0xE03F,0xF01F,0xF000,0x0038,0x0004,0x0003,0x8000,0x01FF,0x01FF,0x80FF,0x81FF,0x01FF,0x80FF,0x8000,0x01C0,0x0020,0x001C,0x0003,0x8000,0x4000,0x3800,0x001F,0xF01F,0xE01C,0x0002,0x0001,0xC000,0x00FF,0x80FF,0xC07F,0xC0FF,0x80FF,0xC07F,


0xC000,0x00E0,0x0010,0x000E,
	0xE000,0x3000,0x011F,0xC03F,0xF00F,0xE200,0x0030,0x001F,0x0001,0x8000,0x08FE,0x01FF,0x807F,0x18FE,0x01FF,0x807F,0x1000,0x0180,0x00F8,0x000C,0x0003,0x0001,0xF000,0x1800,0x008F,0xE01F,0xE018,0x000F,0x8000,0xC000,0x047F,0x00FF,0xC03F,0x8C7F,0x00FF,0xC03F,


0x8800,0x00C0,0x007C,0x0006,
	0xE000,0x0000,0x011F,0x803F,0xF007,0xE200,0x0000,0x001F,0x0000,0x0000,0x08FC,0x01FF,0x803F,0x18FC,0x01FF,0x803F,0x1000,0x0000,0x00F8,0x0000,0x0000,0x0001,0xF000,0x0000,0x008F,0xC01F,0xE000,0x000F,0x8000,0x0000,0x047E,0x00FF,0xC01F,0x8C7E,0x00FF,0xC01F,


0x8800,0x0000,0x007C,0x0000,
	0x1000,0x0000,0x001F,0x803F,0xF007,0xE000,0x0000,0x0020,0x8000,0x0000,0x00FC,0x01FF,0x803F,0x00FC,0x01FF,0x803F,0x0000,0x0000,0x0104,0x0000,0x0000,0x0002,0x0800,0x0000,0x000F,0xC01F,0xE000,0x0010,0x4000,0x0000,0x007E,0x00FF,0xC01F,0x807E,0x00FF,0xC01F,


0x8000,0x0000,0x0082,0x0000,
	0x0400,0x0000,0x061F,0x003F,0xF003,0xE180,0x0000,0x0080,0x2000,0x0000,0x30F8,0x01FF,0x801F,0x00F8,0x01FF,0x801F,0x0C00,0x0000,0x0401,0x0000,0x0000,0x0008,0x0200,0x0000,0x030F,0x801F,0xE000,0x0040,0x1000,0x0000,0x187C,0x00FF,0xC00F,0x807C,0x00FF,0xC00F,


0x8600,0x0000,0x0200,0x8000,
	0x8401,0x0000,0x061F,0x403F,0xF00B,0xE180,0x0002,0x0084,0x2008,0x0000,0x30FA,0x01FF,0x805F,0x00FA,0x01FF,0x805F,0x0C00,0x0010,0x0421,0x0040,0x0000,0x2008,0x4200,0x8000,0x030F,0xA01F,0xC001,0x0042,0x1004,0x0000,0x187D,0x00FF,0xC02F,0x807D,0x00FF,0xC02F,


0x8600,0x0008,0x0210,0x8020,
	0x0000,0x0000,0x0C1E,0x006F,0xD801,0xE0C0,0x0000,0x0000,0x0000,0x0000,0x60F0,0x037E,0xC00F,0x00F0,0x037E,0xC00F,0x0600,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x060F,0x0037,0xE000,0x0000,0x0000,0x0000,0x3078,0x01BF,0x6007,0x8078,0x01BF,0x6007,


0x8300,0x0000,0x0000,0x0000,
	0x0080,0x0000,0x001E,0x006F,0xD801,0xE000,0x0000,0x0400,0x0400,0x0000,0x00F0,0x037E,0xC00F,0x00F0,0x037E,0xC00F,0x0000,0x0000,0x2000,0x2000,0x0000,0x0040,0x0040,0x0000,0x000F,0x0037,0xE000,0x0200,0x0200,0x0000,0x0078,0x01BF,0x6007,0x8078,0x01BF,0x6007,


0x8000,0x0000,0x1000,0x1000,
	0x0080,0x1000,0x1003,0x004F,0xC803,0x0020,0x0020,0x0400,0x0400,0x8000,0x8018,0x027E,0x4018,0x0018,0x027E,0x4018,0x0100,0x0100,0x2000,0x2004,0x0002,0x0040,0x0040,0x0800,0x0801,0x8027,0xE010,0x0200,0x0200,0x4000,0x400C,0x013F,0x200C,0x000C,0x013F,0x200C,


0x0080,0x0080,0x1000,0x1002,
	0x0040,0x7000,0x1001,0x800F,0xC006,0x0020,0x0038,0x0800,0x0203,0x8000,0x800C,0x007E,0x0030,0x000C,0x007E,0x0030,0x0100,0x01C0,0x4000,0x101C,0x0003,0x8080,0x0020,0x3800,0x0800,0xC007,0xE01C,0x0400,0x0101,0xC000,0x4006,0x003F,0x0018,0x0006,0x003F,0x0018,


0x0080,0x00E0,0x2000,0x080E,
	0x803F,0x1801,0x1005,0xE007,0x801E,0x8022,0x0063,0xF004,0x01F8,0xC008,0x802F,0x003C,0x00F4,0x002F,0x003C,0x00F4,0x0110,0x031F,0x8020,0x0FC6,0x0006,0x3F00,0x401F,0x8C00,0x8802,0xF003,0xC031,0xF802,0x00FC,0x6004,0x4017,0x801E,0x007A,0x0017,0x801E,0x007A,


0x0088,0x018F,0xC010,0x07E3,
	0x803F,0x0801,0x6004,0xE007,0x801C,0x801A,0x0043,0xF004,0x01F8,0x400B,0x0027,0x003C,0x00E4,0x0027,0x003C,0x00E4,0x00D0,0x021F,0x8020,0x0FC2,0x0004,0x3F00,0x401F,0x8400,0xB002,0x7003,0xC021,0xF802,0x00FC,0x2005,0x8013,0x801E,0x0072,0x0013,0x801E,0x0072,


0x0068,0x010F,0xC010,0x07E1,
	0x801F,0xEC01,0xE004,0xF003,0x003C,0x801E,0x00DF,0xE004,0x00FF,0x600F,0x0027,0x8018,0x01E4,0x0027,0x8018,0x01E4,0x00F0,0x06FF,0x0020,0x07FB,0x000D,0xFE00,0x400F,0xF600,0xF002,0x7801,0x806F,0xF002,0x007F,0xB007,0x8013,0xC00C,0x00F2,0x0013,0xC00C,0x00F2,


0x0078,0x037F,0x8010,0x03FD,
	0x800F,0xBE00,0xF002,0x7C03,0x00F9,0x003C,0x01F7,0xC004,0x007D,0xF007,0x8013,0xE018,0x07C8,0x0013,0xE018,0x07C8,0x01E0,0x0FBE,0x0020,0x03EF,0x801F,0x7C00,0x4007,0xDF00,0x7801,0x3E01,0x80FB,0xE002,0x003E,0xF803,0xC009,0xF00C,0x03E4,0x0009,0xF00C,0x03E4,


0x00F0,0x07DF,0x0010,0x01FF,
	0x8003,0x3E00,0xF003,0x3C00,0x00F3,0x003C,0x01F3,0x0004,0x0019,0xF007,0x8019,0xE000,0x0798,0x0019,0xE000,0x0798,0x01E0,0x0F98,0x0020,0x00CF,0x801F,0x3000,0x4001,0x9F00,0x7801,0x9E00,0x00F9,0x8002,0x000C,0xF803,0xC00C,0xF000,0x03CC,0x000C,0xF000,0x03CC,


0x00F0,0x07CC,0x0010,0x007F,
	0x8001,0x8200,0xE003,0x3E03,0x01F3,0x001C,0x0106,0x0004,0x000C,0x1007,0x0019,0xF018,0x0F98,0x0019,0xF018,0x0F98,0x00E0,0x0830,0x0020,0x0060,0x8010,0x6000,0x4000,0xC100,0x7001,0x9F01,0x8083,0x0002,0x0006,0x0803,0x800C,0xF80C,0x07CC,0x000C,0xF80C,0x07CC,


0x0070,0x0418,0x0010,0x003F,
	0x8000,0xC000,0x6001,0x0E07,0x81C2,0x0018,0x000C,0x0004,0x0006,0x0003,0x0008,0x703C,0x0E10,0x0008,0x703C,0x0E10,0x00C0,0x0060,0x0020,0x0030,0x0000,0xC000,0x4000,0x6000,0x3000,0x8703,0xC006,0x0002,0x0003,0x0001,0x8004,0x381E,0x0708,0x0004,0x381E,0x0708,


0x0060,0x0030,0x0010,0x001F,
	0x8000,0xC040,0x6001,0x0207,0x8102,0x0018,0x080C,0x0004,0x0006,0x0203,0x0008,0x103C,0x0810,0x0008,0x103C,0x0810,0x00C0,0x4060,0x0020,0x0030,0x1080,0xC000,0x4000,0x6020,0x3000,0x8103,0xC406,0x0002,0x0003,0x0101,0x8004,0x081E,0x0408,0x0004,0x081E,0x0408,


0x0060,0x2030,0x0010,0x001F,
	0x8000,0x6040,0x3001,0x8100,0x0206,0x0030,0x0818,0x0004,0x0003,0x0201,0x800C,0x0800,0x1030,0x000C,0x0800,0x1030,0x0180,0x40C0,0x0020,0x0018,0x1081,0x8000,0x4000,0x3020,0x1800,0xC080,0x040C,0x0002,0x0001,0x8100,0xC006,0x0400,0x0818,0x0006,0x0400,0x0818,


0x00C0,0x2060,0x0010,0x000F,
	0x8000,0x2000,0x0003,0x8100,0x0207,0x0000,0x0010,0x0004,0x0001,0x0000,0x001C,0x0800,0x1038,0x001C,0x0800,0x1038,0x0000,0x0080,0x0020,0x0008,0x0001,0x0000,0x4000,0x1000,0x0001,0xC080,0x0008,0x0002,0x0000,0x8000,0x000E,0x0400,0x081C,0x000E,0x0400,0x081C,


0x0000,0x0040,0x0010,0x0007,
	0xC000,0x3000,0x0003,0x8000,0x0007,0x0000,0x0030,0x000E,0x0001,0x8000,0x001C,0x0000,0x0038,0x001C,0x0000,0x0038,0x0000,0x0180,0x0070,0x000C,0x0003,0x0000,0xE000,0x1800,0x0001,0xC000,0x0018,0x0007,0x0000,0xC000,0x000E,0x0000,0x001C,0x000E,0x0000,0x001C,


0x0000,0x00C0,0x0038,0x0007,
	0xC000,0x1020,0x0403,0x8000,0x0007,0x0080,0x1020,0x000E,0x0000,0x8100,0x201C,0x0000,0x0038,0x001C,0x0000,0x0038,0x0400,0x8100,0x0070,0x0004,0x0902,0x0000,0xE000,0x0810,0x0201,0xC000,0x0810,0x0007,0x0000,0x4080,0x100E,0x0000,0x001C,0x000E,0x0000,0x001C,


0x0200,0x4080,0x0038,0x0003,
	0xC000,0x0C00,0x0400,0x2000,0x0010,0x0080,0x00C0,0x000E,0x0000,0x6000,0x2001,0x0000,0x0080,0x0001,0x0000,0x0080,0x0400,0x0600,0x0070,0x0003,0x000C,0x0000,0xE000,0x0600,0x0200,0x1000,0x0060,0x0007,0x0000,0x3000,0x1000,0x8000,0x0040,0x0000,0x8000,0x0040,


0x0200,0x0300,0x0038,0x0001,
	0xE000,0x0014,0x0002,0x0000,0x0001,0x0000,0xA000,0x001F,0x0000,0x00A0,0x0010,0x0000,0x0008,0x0010,0x0000,0x0008,0x0005,0x0000,0x00F8,0x0000,0x0600,0x0001,0xF000,0x000A,0x0001,0x0000,0x1000,0x000F,0x8000,0x0050,0x0008,0x0000,0x0004,0x0008,0x0000,0x0004,


0x0002,0x8000,0x007C,0x0000,
	0xE000,0x001E,0x8200,0x0000,0x0000,0x0105,0xE000,0x001F,0x0000,0x00F4,0x1000,0x0000,0x0000,0x0000,0x0000,0x0000,0x082F,0x0000,0x00F8,0x0000,0x0600,0x0001,0xF000,0x000F,0x4100,0x0000,0x1000,0x000F,0x8000,0x007A,0x0800,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0417,0x8000,0x007C,0x0000,
	0xF000,0x000F,0xFF80,0x0000,0x0000,0x07FF,0xC000,0x003F,0x8000,0x007F,0xFC00,0x0000,0x0000,0x3C00,0x0000,0x0000,0x3FFE,0x0000,0x01FC,0x0000,0x0000,0x0003,0xF800,0x0007,0xFFC0,0x0000,0x0000,0x001F,0xC000,0x003F,0xFE00,0x0000,0x0000,0x1E00,0x0000,0x0000,


0x1FFF,0x0000,0x00FE,0x0000,
	0xF000,0x000F,0xFFC0,0x0000,0x0000,0x0FFF,0xC000,0x003F,0x8000,0x007F,0xFE00,0x0000,0x0000,0x7E00,0x0000,0x0000,0x7FFE,0x0000,0x01FC,0x0000,0x0000,0x0003,0xF800,0x0007,0xFFE0,0x0000,0x0000,0x001F,0xC000,0x003F,0xFF00,0x0000,0x0000,0x3F00,0x0000,0x0000,


0x3FFF,0x0000,0x00FE,0x0000,
	0x9800,0x00CE,0x0FC0,0x0000,0x0000,0x0FC1,0xCC00,0x0064,0xC000,0x0670,0x7E00,0x0000,0x0000,0x7E00,0x0000,0x0000,0x7E0E,0x6000,0x0326,0x0000,0x30C0,0x0006,0x4C00,0x0067,0x07E0,0x0000,0x0600,0x0032,0x6000,0x0338,0x3F00,0x0000,0x0000,0x3F00,0x0000,0x0000,


0x3F07,0x3000,0x0193,0x0000,
	0x9800,0x0066,0x0780,0x0800,0x0040,0x0781,0x9800,0x0064,0xC000,0x0330,0x3C00,0x4000,0x0200,0x3C00,0x4000,0x0200,0x3C0C,0xC000,0x0326,0x0000,0x1980,0x0006,0x4C00,0x0033,0x03C0,0x0400,0x0C00,0x0032,0x6000,0x0198,0x1E00,0x2000,0x0100,0x1E00,0x2000,0x0100,


0x1E06,0x6000,0x0193,0x0000,
	0x0800,0x0074,0x0701,0x0800,0x0042,0x0380,0xB800,0x0040,0x4000,0x03A0,0x3808,0x4000,0x0210,0x1808,0x4000,0x0210,0x1C05,0xC000,0x0202,0x0000,0x1F80,0x0004,0x0400,0x003A,0x0380,0x8400,0x1C00,0x0020,0x2000,0x01D0,0x1C04,0x2000,0x0108,0x0C04,0x2000,0x0108,


0x0E02,0xE000,0x0101,0x0000,
	0x1000,0x1830,0x0300,0x0000,0x0000,0x0300,0x3060,0x0020,0x8000,0xC180,0x1800,0x0000,0x0000,0x1800,0x0000,0x0000,0x1801,0x8300,0x0104,0x0006,0x0F06,0x0002,0x0800,0x0C18,0x0180,0x0000,0x1830,0x0010,0x4000,0x60C0,0x0C00,0x0000,0x0000,0x0C00,0x0000,0x0000,


0x0C00,0xC180,0x0082,0x0003,
	0x8000,0x1C30,0x0180,0x0000,0x0000,0x0600,0x30E0,0x0004,0x0000,0xE180,0x0C00,0x0000,0x0000,0x3C00,0x0000,0x0000,0x3001,0x8700,0x0020,0x0007,0x0F0E,0x0000,0x4000,0x0E18,0x00C0,0x0000,0x1870,0x0002,0x0000,0x70C0,0x0600,0x0000,0x0000,0x1E00,0x0000,0x0000,


0x1800,0xC380,0x0010,0x0003,
	0x0000,0x0000,0x0060,0x0000,0x0000,0x1800,0x0000,0x0000,0x0000,0x0000,0x0300,0x0000,0x0000,0xC300,0x0000,0x0000,0xC000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0030,0x0000,0x0000,0x0000,0x0000,0x0000,0x0180,0x0000,0x0000,0x6180,0x0000,0x0000,


0x6000,0x0000,0x0000,0x0000,
	0x0000,0x0008,0x0060,0x0000,0x0000,0x1800,0x4000,0x0000,0x0000,0x0040,0x0300,0x0000,0x0000,0xC300,0x0000,0x0000,0xC002,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0004,0x0030,0x0000,0x0000,0x0000,0x0000,0x0020,0x0180,0x0000,0x0000,0x6180,0x0000,0x0000,


0x6001,0x0000,0x0000,0x0000,
	0x0000,0x2008,0x1000,0x0000,0x0000,0x0020,0x4010,0x0000,0x0001,0x0040,0x8000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0102,0x0080,0x0000,0x0008,0x0001,0x0000,0x0000,0x1004,0x0800,0x0000,0x0008,0x0000,0x0000,0x8020,0x4000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0081,0x0040,0x0000,0x0004,
	0x00C0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0C00,0x0600,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x6000,0x3000,0x0000,0x00C0,0x0060,0x0000,0x0000,0x0000,0x0000,0x0600,0x0300,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0000,0x0000,0x3000,0x1800,
	0x00EC,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0xDC00,0x0760,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0006,0xE000,0x3B00,0x0000,0x0DC0,0x0076,0x0000,0x0000,0x0000,0x0000,0x6E00,0x03B0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0000,0x0003,0x7000,0x1D80,
	0x24FE,0x0000,0x0800,0x0004,0x8000,0x0040,0x0001,0xFC91,0x27F0,0x0000,0x4000,0x0024,0x0000,0x0000,0x0024,0x0000,0x0200,0x000F,0xE489,0x3F80,0x0000,0x1FC9,0x127F,0x0000,0x0400,0x0002,0x4000,0xFE48,0x93F8,0x0000,0x2000,0x0012,0x0000,0x0000,0x0012,0x0000,


0x0100,0x0007,0xF244,0x9FC0,
	0x2DF8,0x0000,0x8F00,0x0180,0x0600,0x03C4,0x0000,0x7ED1,0x6FC0,0x0004,0x7800,0x0C00,0x3000,0x1800,0x0C00,0x3000,0x1E20,0x0003,0xF68B,0x7E00,0x0000,0x07ED,0x16FC,0x0000,0x4780,0x00C0,0x0000,0x3F68,0xB7E0,0x0002,0x3C00,0x0600,0x1800,0x0C00,0x0600,0x1800,


0x0F10,0x0001,0xFB45,0xBF00,
	0x85F8,0x0000,0x0F80,0x00C0,0x0C00,0x07C0,0x0000,0x7E84,0x2FC0,0x0000,0x7C00,0x0600,0x6000,0x3C00,0x0600,0x6000,0x3E00,0x0003,0xF421,0x7E00,0x0000,0x07E8,0x42FC,0x0000,0x07C0,0x0060,0x0000,0x3F42,0x17E0,0x0000,0x3E00,0x0300,0x3000,0x1E00,0x0300,0x3000,


0x1F00,0x0001,0xFA10,0xBF00,
	0x81F8,0x0020,0x1B80,0x00E0,0x1C00,0x0760,0x1000,0x7E04,0x0FC0,0x0100,0xDC00,0x0700,0xE000,0x3C00,0x0700,0xE000,0x3B00,0x8003,0xF020,0x7E00,0x0900,0x07E0,0x40FC,0x0010,0x0DC0,0x0070,0x0800,0x3F02,0x07E0,0x0080,0x6E00,0x0380,0x7000,0x1E00,0x0380,0x7000,


0x1D80,0x4001,0xF810,0x3F00,
	0xC0F8,0x0030,0x3380,0x0060,0x1800,0x0730,0x3000,0x7C0E,0x07C0,0x0181,0x9C00,0x0300,0xC000,0x3C00,0x0300,0xC000,0x3981,0x8003,0xE070,0x3E00,0x0F00,0x07C0,0xE07C,0x0018,0x19C0,0x0030,0x1800,0x3E07,0x03E0,0x00C0,0xCE00,0x0180,0x6000,0x1E00,0x0180,0x6000,


0x1CC0,0xC001,0xF038,0x1F00,
	0x80FD,0x0010,0x1300,0x0023,0x1000,0x0320,0x2002,0xFC04,0x07E8,0x0080,0x9800,0x0118,0x8000,0x1800,0x0118,0x8000,0x1901,0x0017,0xE020,0x3F40,0x0600,0x2FC0,0x407E,0x8008,0x0980,0x0011,0x9001,0x7E02,0x03F4,0x0040,0x4C00,0x008C,0x4000,0x0C00,0x008C,0x4000,


0x0C80,0x800B,0xF010,0x1FA0,
	0x80F9,0x8010,0x1200,0x003C,0xF000,0x0120,0x2006,0x7C04,0x07CC,0x0080,0x9000,0x01E7,0x8000,0x0000,0x01E7,0x8000,0x0901,0x0033,0xE020,0x3E60,0x0600,0x67C0,0x407C,0xC008,0x0900,0x001E,0x7003,0x3E02,0x03E6,0x0040,0x4800,0x00F3,0xC000,0x0000,0x00F3,0xC000,


0x0480,0x8019,0xF010,0x1F30,
	0x00E1,0x8010,0x1A00,0x003C,0xF000,0x0160,0x2006,0x1C00,0x070C,0x0080,0xD000,0x01E7,0x8000,0x0000,0x01E7,0x8000,0x0B01,0x0030,0xE000,0x3860,0x0600,0x61C0,0x0070,0xC008,0x0D00,0x001E,0x7003,0x0E00,0x0386,0x0040,0x6800,0x00F3,0xC000,0x0000,0x00F3,0xC000,


0x0580,0x8018,0x7000,0x1C30,
	0x00E3,0x800C,0x0000,0x007C,0xF800,0x0000,0xC007,0x1C00,0x071C,0x0060,0x0000,0x03E7,0xC000,0x0000,0x03E7,0xC000,0x0006,0x0038,0xE000,0x38E0,0x0000,0x71C0,0x0071,0xC006,0x0000,0x003E,0x6003,0x8E00,0x038E,0x0030,0x0000,0x01F3,0xE000,0x0000,0x01F3,0xE000,


0x0003,0x001C,0x7000,0x1C70,
	0x8060,0x0008,0x0000,0x007C,0xF800,0x0000,0x4000,0x1804,0x0300,0x0040,0x0000,0x03E7,0xC000,0x0000,0x03E7,0xC000,0x0002,0x0000,0xC020,0x1800,0x0000,0x0180,0x4030,0x0004,0x0000,0x003E,0x6000,0x0C02,0x0180,0x0020,0x0000,0x01F3,0xE000,0x0000,0x01F3,0xE000,


0x0001,0x0000,0x6010,0x0C00,
	0x8060,0x000E,0x0000,0x00D8,0x6C00,0x0001,0xC000,0x1804,0x0300,0x0070,0x0000,0x06C3,0x6000,0x0000,0x06C3,0x6000,0x000E,0x0000,0xC020,0x1800,0x0000,0x0180,0x4030,0x0007,0x0000,0x006C,0x2000,0x0C02,0x0180,0x0038,0x0000,0x0361,0xB000,0x0000,0x0361,0xB000,


0x0007,0x0000,0x6010,0x0C00,
	0x0060,0x000E,0x0400,0x0018,0x6000,0x0081,0xC000,0x1800,0x0300,0x0070,0x2000,0x00C3,0x0000,0x0000,0x00C3,0x0000,0x040E,0x0000,0xC000,0x1800,0x0000,0x0180,0x0030,0x0007,0x0200,0x000C,0x2000,0x0C00,0x0180,0x0038,0x1000,0x0061,0x8000,0x0000,0x0061,0x8000,


0x0207,0x0000,0x6000,0x0C00,
	0x0070,0x0018,0x0400,0x0208,0x4100,0x0080,0x6000,0x3800,0x0380,0x00C0,0x2000,0x1042,0x0800,0x0000,0x1042,0x0800,0x0403,0x0001,0xC000,0x1C00,0x0600,0x0380,0x0038,0x000C,0x0200,0x0104,0x3000,0x1C00,0x01C0,0x0060,0x1000,0x0821,0x0400,0x0000,0x0821,0x0400,


0x0201,0x8000,0xE000,0x0E00,
	0x0070,0x0018,0x0004,0x0200,0x0100,0x8000,0x6000,0x3800,0x0380,0x00C0,0x0020,0x1000,0x0804,0x0020,0x1000,0x0804,0x0003,0x0001,0xC000,0x1C00,0x0600,0x0380,0x0038,0x000C,0x0002,0x0100,0x1000,0x1C00,0x01C0,0x0060,0x0010,0x0800,0x0402,0x0010,0x0800,0x0402,


0x0001,0x8000,0xE000,0x0E00,
	0x0078,0x0000,0x0004,0x0700,0x0380,0x8000,0x0000,0x7800,0x03C0,0x0000,0x0020,0x3800,0x1C04,0x0020,0x3800,0x1C04,0x0000,0x0003,0xC000,0x1E00,0x0000,0x0780,0x003C,0x0000,0x0002,0x0380,0x0000,0x3C00,0x01E0,0x0000,0x0010,0x1C00,0x0E02,0x0010,0x1C00,0x0E02,


0x0000,0x0001,0xE000,0x0F00,
	0x003C,0x0000,0x0086,0x0600,0x0181,0x8400,0x0000,0xF000,0x01E0,0x0000,0x0430,0x3000,0x0C0C,0x2430,0x3000,0x0C0C,0x2000,0x0007,0x8000,0x0F00,0x0000,0x0F00,0x001E,0x0000,0x0043,0x0300,0x0000,0x7800,0x00F0,0x0000,0x0218,0x1800,0x0606,0x1218,0x1800,0x0606,


0x1000,0x0003,0xC000,0x0780,
	0x003C,0x0070,0x0006,0x0000,0x0001,0x8000,0x3800,0xF000,0x01E0,0x0380,0x0030,0x0000,0x000C,0x0030,0x0000,0x000C,0x0001,0xC007,0x8000,0x0F00,0x1F80,0x0F00,0x001E,0x0038,0x0003,0x0000,0x1C00,0x7800,0x00F0,0x01C0,0x0018,0x0000,0x0006,0x0018,0x0000,0x0006,


0x0000,0xE003,0xC000,0x0780,
	0x0034,0x0070,0x300E,0x0400,0x0081,0xC030,0x3800,0xB000,0x01A0,0x0381,0x8070,0x2000,0x040E,0x0070,0x2000,0x040E,0x0181,0xC005,0x8000,0x0D00,0x1F80,0x0B00,0x001A,0x0038,0x1807,0x0200,0x1C00,0x5800,0x00D0,0x01C0,0xC038,0x1000,0x0207,0x0038,0x1000,0x0207,


0x00C0,0xE002,0xC000,0x0680,
	0x8014,0x0070,0x300E,0x3C00,0x00F1,0xC030,0x3800,0xA004,0x00A0,0x0381,0x8071,0xE000,0x078E,0x0071,0xE000,0x078E,0x0181,0xC005,0x0020,0x0500,0x1F80,0x0A00,0x400A,0x0038,0x1807,0x1E00,0x1C00,0x5002,0x0050,0x01C0,0xC038,0xF000,0x03C7,0x0038,0xF000,0x03C7,


0x00C0,0xE002,0x8010,0x0280,
	0xC010,0x0070,0x0000,0x3003,0x0030,0x0000,0x3800,0x200E,0x0080,0x0380,0x0001,0x8018,0x0180,0x0001,0x8018,0x0180,0x0001,0xC001,0x0070,0x0400,0x1F80,0x0200,0xE008,0x0038,0x0000,0x1801,0x9C00,0x1007,0x0040,0x01C0,0x0000,0xC00C,0x00C0,0x0000,0xC00C,0x00C0,


0x0000,0xE000,0x8038,0x0200,
	0xE018,0x0070,0x0000,0x2003,0x0010,0x0000,0x3800,0x601F,0x00C0,0x0380,0x0001,0x0018,0x0080,0x0001,0x0018,0x0080,0x0001,0xC003,0x00F8,0x0600,0x1F80,0x0601,0xF00C,0x0038,0x0000,0x1001,0x9C00,0x300F,0x8060,0x01C0,0x0000,0x800C,0x0040,0x0000,0x800C,0x0040,


0x0000,0xE001,0x807C,0x0300,
	0xE018,0x0044,0x0000,0x2003,0x0010,0x0000,0x8800,0x601F,0x00C0,0x0220,0x0001,0x0018,0x0080,0x0001,0x0018,0x0080,0x0004,0x4003,0x00F8,0x0600,0x1080,0x0601,0xF00C,0x0022,0x0000,0x1001,0x8400,0x300F,0x8060,0x0110,0x0000,0x800C,0x0040,0x0000,0x800C,0x0040,


0x0002,0x2001,0x807C,0x0300,
	0xD808,0x0004,0x0000,0x2000,0x0010,0x0000,0x8000,0x406E,0xC040,0x0020,0x0001,0x0000,0x0080,0x0001,0x0000,0x0080,0x0004,0x0002,0x0376,0x0200,0x0000,0x0406,0xEC04,0x0002,0x0000,0x1000,0x0000,0x2037,0x6020,0x0010,0x0000,0x8000,0x0040,0x0000,0x8000,0x0040,


0x0002,0x0001,0x01BB,0x0100,
	0xDC08,0x0084,0x0000,0x2000,0x0010,0x0000,0x8400,0x40EE,0xE040,0x0420,0x0001,0x0000,0x0080,0x0001,0x0000,0x0080,0x0004,0x2002,0x0777,0x0200,0x2040,0x040E,0xEE04,0x0042,0x0000,0x1000,0x0200,0x2077,0x7020,0x0210,0x0000,0x8000,0x0040,0x0000,0x8000,0x0040,


0x0002,0x1001,0x03BB,0x8100,
	0xC60C,0x0084,0x0000,0x3A00,0x0170,0x0000,0x8400,0xC18E,0x3060,0x0420,0x0001,0xD000,0x0B80,0x0001,0xD000,0x0B80,0x0004,0x2006,0x0C71,0x8300,0x2040,0x0C18,0xE306,0x0042,0x0000,0x1D00,0x0200,0x60C7,0x1830,0x0210,0x0000,0xE800,0x05C0,0x0000,0xE800,0x05C0,


0x0002,0x1003,0x0638,0xC180,
	0x820E,0x808C,0x0000,0x3380,0x0730,0x0000,0xC405,0xC104,0x1074,0x0460,0x0001,0x9C00,0x3980,0x0001,0x9C00,0x3980,0x0006,0x202E,0x0820,0x83A0,0x2040,0x5C10,0x4107,0x4046,0x0000,0x19C0,0x0202,0xE082,0x083A,0x0230,0x0000,0xCE00,0x1CC0,0x0000,0xCE00,0x1CC0,


0x0003,0x1017,0x0410,0x41D0,
	0x8187,0x8084,0x1000,0x0000,0x0000,0x0020,0x8407,0x8604,0x0C3C,0x0420,0x8000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0104,0x203C,0x3020,0x61E0,0x2040,0x7860,0x40C3,0xC042,0x0800,0x0000,0x0203,0xC302,0x061E,0x0210,0x4000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0082,0x101E,0x1810,0x30F0,
	0x81C7,0x80C0,0x1040,0x0000,0x0000,0x0820,0x0C07,0x8E04,0x0E3C,0x0600,0x8200,0x0000,0x0000,0x4200,0x0000,0x0000,0x4100,0x603C,0x7020,0x71E0,0x30C0,0x78E0,0x40E3,0xC060,0x0820,0x0000,0x0603,0xC702,0x071E,0x0300,0x4100,0x0000,0x0000,0x2100,0x0000,0x0000,


0x2080,0x301E,0x3810,0x38F0,
	0x81E7,0x80C0,0x3040,0x0000,0x0000,0x0830,0x0C07,0x9E04,0x0F3C,0x0601,0x8200,0x0000,0x0000,0x4200,0x0000,0x0000,0x4180,0x603C,0xF020,0x79E0,0x30C0,0x79E0,0x40F3,0xC060,0x1820,0x0000,0x0603,0xCF02,0x079E,0x0300,0xC100,0x0000,0x0000,0x2100,0x0000,0x0000,


0x20C0,0x301E,0x7810,0x3CF0,
	0x80CF,0x00C0,0x3000,0x0000,0x0000,0x0030,0x0C03,0xCC04,0x0678,0x0601,0x8000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0180,0x601E,0x6020,0x33C0,0x30C0,0x3CC0,0x4067,0x8060,0x1800,0x0000,0x0601,0xE602,0x033C,0x0300,0xC000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x00C0,0x300F,0x3010,0x19E0,
	0x808F,0x0088,0x1800,0x0000,0x0000,0x0060,0x4403,0xC404,0x0478,0x0440,0xC000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0302,0x201E,0x2020,0x23C0,0x2040,0x3C40,0x4047,0x8044,0x0C00,0x0000,0x0201,0xE202,0x023C,0x0220,0x6000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0181,0x100F,0x1010,0x11E0,
	0x800F,0x018C,0x1800,0x0040,0x0800,0x0060,0xC603,0xC004,0x0078,0x0C60,0xC000,0x0200,0x4000,0x0000,0x0200,0x4000,0x0306,0x301E,0x0020,0x03C0,0x6060,0x3C00,0x4007,0x80C6,0x0C00,0x0020,0x0301,0xE002,0x003C,0x0630,0x6000,0x0100,0x2000,0x0000,0x0100,0x2000,


0x0183,0x180F,0x0010,0x01E0,
	0xA027,0x438C,0x3800,0xC0C0,0x0C0C,0x0070,0xC70B,0x9015,0x013A,0x1C61,0xC006,0x0600,0x6060,0x0006,0x0600,0x6060,0x0386,0x385C,0x80A8,0x09D0,0xE070,0xB901,0x5013,0xA1C6,0x1C00,0x6060,0x0385,0xC80A,0x809D,0x0E30,0xE003,0x0300,0x3030,0x0003,0x0300,0x3030,


0x01C3,0x1C2E,0x4054,0x04E8,
	0x8013,0x8386,0x1800,0xC000,0x000C,0x0061,0x8707,0x2004,0x009C,0x1C30,0xC006,0x0000,0x0060,0x0006,0x0000,0x0060,0x030C,0x3839,0x0020,0x04E0,0xE070,0x7200,0x4009,0xC1C3,0x0C00,0x6000,0x0383,0x9002,0x004E,0x0E18,0x6003,0x0000,0x0030,0x0003,0x0000,0x0030,


0x0186,0x1C1C,0x8010,0x0270,
	0x8011,0x8386,0x0C00,0xC000,0x000C,0x00C1,0x8706,0x2004,0x008C,0x1C30,0x6006,0x0000,0x0060,0x0006,0x0000,0x0060,0x060C,0x3831,0x0020,0x0460,0xE070,0x6200,0x4008,0xC1C3,0x0600,0x6000,0x0383,0x1002,0x0046,0x0E18,0x3003,0x0000,0x0030,0x0003,0x0000,0x0030,


0x0306,0x1C18,0x8010,0x0230,
	0xB011,0xC784,0x0C00,0x8000,0x0004,0x00C0,0x878E,0x2035,0x808E,0x3C20,0x6004,0x0000,0x0020,0x0004,0x0000,0x0020,0x0604,0x3C71,0x01AC,0x0471,0xE078,0xE203,0x5808,0xE3C2,0x0600,0x4000,0x03C7,0x101A,0xC047,0x1E10,0x3002,0x0000,0x0010,0x0002,0x0000,0x0010,


0x0302,0x1E38,0x80D6,0x0238,
	0x3010,0xCF02,0x1802,0x0000,0x0001,0x0061,0x03CC,0x2031,0x8086,0x7810,0xC010,0x0000,0x0008,0x0010,0x0000,0x0008,0x0308,0x1E61,0x018C,0x0433,0xC03C,0xC203,0x1808,0x6781,0x0C01,0x0000,0x01E6,0x1018,0xC043,0x3C08,0x6008,0x0000,0x0004,0x0008,0x0000,0x0004,


0x0184,0x0F30,0x80C6,0x0219,
	0x1810,0xEF02,0x3883,0x0020,0x1003,0x0471,0x03DC,0x2060,0xC087,0x7811,0xC418,0x0100,0x8018,0x2418,0x0100,0x8018,0x2388,0x1EE1,0x0306,0x043B,0xC03D,0xC206,0x0C08,0x7781,0x1C41,0x8010,0x01EE,0x1030,0x6043,0xBC08,0xE20C,0x0080,0x400C,0x120C,0x0080,0x400C,


0x11C4,0x0F70,0x8183,0x021D,
	0x1810,0x6F02,0x3A83,0x0070,0x3803,0x0571,0x03D8,0x2060,0xC083,0x7811,0xD418,0x0381,0xC018,0x2418,0x0381,0xC018,0x2B88,0x1EC1,0x0306,0x041B,0xC03D,0x8206,0x0C08,0x3781,0x1D41,0x8038,0x01EC,0x1030,0x6041,0xBC08,0xEA0C,0x01C0,0xE00C,0x120C,0x01C0,0xE00C,


0x15C4,0x0F60,0x8183,0x020D,
	0x0010,0x7FC7,0x1A09,0x0070,0x3802,0x4163,0x8FF8,0x2000,0x0083,0xFE38,0xD048,0x0381,0xC012,0x0048,0x0381,0xC012,0x0B1C,0x7FC1,0x0000,0x041F,0xF0FF,0x8200,0x0008,0x3FE3,0x8D04,0x8038,0x07FC,0x1000,0x0041,0xFF1C,0x6824,0x01C0,0xE009,0x0024,0x01C0,0xE009,


0x058E,0x3FE0,0x8000,0x020F,
	0xA010,0x3FD7,0x9E08,0x0070,0x3800,0x41E7,0xAFF0,0x2015,0x0081,0xFEBC,0xF040,0x0381,0xC002,0x0040,0x0381,0xC002,0x0F3D,0x7F81,0x00A8,0x040F,0xF6FF,0x0201,0x5008,0x1FEB,0xCF04,0x0038,0x17F8,0x100A,0x8040,0xFF5E,0x7820,0x01C0,0xE001,0x0020,0x01C0,0xE001,


0x079E,0xBFC0,0x8054,0x0207,
	0x9018,0x3EFF,0xFE22,0x0030,0x3001,0x11FF,0xFDF0,0x6024,0x80C1,0xF7FF,0xF110,0x0181,0x8008,0x8110,0x0181,0x8008,0x8FFF,0xEF83,0x0124,0x060F,0xBFDF,0x0602,0x480C,0x1F7F,0xFF11,0x0018,0x1EF8,0x3012,0x4060,0xFBFF,0xF888,0x00C0,0xC004,0x4088,0x00C0,0xC004,


0x47FF,0xF7C1,0x8092,0x0307,
	0x9000,0x3E7F,0xFE3E,0x0010,0x2001,0xF1FF,0xF9F0,0x0024,0x8001,0xF3FF,0xF1F0,0x0081,0x000F,0x81F0,0x0081,0x000F,0x8FFF,0xCF80,0x0124,0x000F,0x9F9F,0x0002,0x4800,0x1F3F,0xFF1F,0x0008,0x1CF8,0x0012,0x4000,0xF9FF,0xF8F8,0x0040,0x8007,0xC0F8,0x0040,0x8007,


0xC7FF,0xE7C0,0x0092,0x0007,
	0x9800,0x3E3F,0xFE1F,0x0000,0x0003,0xE1FF,0xF1F0,0x0064,0xC001,0xF1FF,0xF0F8,0x0000,0x001F,0x00F8,0x0000,0x001F,0x0FFF,0x8F80,0x0326,0x000F,0x8F1F,0x0006,0x4C00,0x1F1F,0xFF0F,0x8000,0x18F8,0x0032,0x6000,0xF8FF,0xF87C,0x0000,0x000F,0x807C,0x0000,0x000F,


0x87FF,0xC7C0,0x0193,0x0007,
	0x9800,0x1F1D,0xFF1F,0x0000,0x0003,0xE3FE,0xE3E0,0x0064,0xC000,0xF8EF,0xF8F8,0x0000,0x001F,0x18F8,0x0000,0x001F,0x1FF7,0x1F00,0x0326,0x0007,0xC63E,0x0006,0x4C00,0x0F8E,0xFF8F,0x8000,0x11F0,0x0032,0x6000,0x7C77,0xFC7C,0x0000,0x000F,0x8C7C,0x0000,0x000F,


0x8FFB,0x8F80,0x0193,0x0003,
	0x9800,0x7784,0xFF1F,0x0000,0x0003,0xE3FC,0x87B8,0x0064,0xC003,0xBC27,0xF8F8,0x0000,0x001F,0x18F8,0x0000,0x001F,0x1FE4,0x3DC0,0x0326,0x001D,0xE07B,0x8006,0x4C00,0x3BC2,0x7F8F,0x8000,0x03DC,0x0032,0x6001,0xDE13,0xFC7C,0x0000,0x000F,0x8C7C,0x0000,0x000F,


0x8FF2,0x1EE0,0x0193,0x000E,
	0x1800,0xF3E4,0x7E1F,0x0010,0x2003,0xE1F8,0x9F3C,0x0060,0xC007,0x9F23,0xF0F8,0x0081,0x001F,0x00F8,0x0081,0x001F,0x0FC4,0xF9E0,0x0306,0x003C,0xF9F3,0xC006,0x0C00,0x79F2,0x3F0F,0x8008,0x0F9E,0x0030,0x6003,0xCF91,0xF87C,0x0040,0x800F,0x807C,0x0040,0x800F,


0x87E2,0x7CF0,0x0183,0x001E,
	0x0800,0xE1E0,0x3E3E,0x0008,0x4001,0xF1F0,0x1E1C,0x0040,0x4007,0x0F01,0xF1F0,0x0042,0x000F,0x81F0,0x0042,0x000F,0x8F80,0xF0E0,0x0202,0x0038,0x79E1,0xC004,0x0400,0x70F0,0x1F1F,0x0004,0x2F0E,0x0020,0x2003,0x8780,0xF8F8,0x0021,0x0007,0xC0F8,0x0021,0x0007,


0xC7C0,0x7870,0x0101,0x001C,
	0x0880,0x40E0,0x3E3E,0x0008,0x4001,0xF1F0,0x1C08,0x0440,0x4402,0x0701,0xF1F0,0x0042,0x000F,0x81F0,0x0042,0x000F,0x8F80,0xE040,0x2202,0x2010,0x39C0,0x8044,0x0440,0x2070,0x1F1F,0x0004,0x2E04,0x0220,0x2201,0x0380,0xF8F8,0x0021,0x0007,0xC0F8,0x0021,0x0007,


0xC7C0,0x7020,0x1101,0x1008,
	0x8080,0x00C0,0x3F3F,0x0000,0x0003,0xF3F0,0x0C00,0x0404,0x0400,0x0601,0xF9F8,0x0000,0x001F,0x99F8,0x0000,0x001F,0x9F80,0x6000,0x2020,0x2000,0x30C0,0x0040,0x4040,0x0060,0x1F9F,0x8000,0x0600,0x0202,0x0200,0x0300,0xFCFC,0x0000,0x000F,0xCCFC,0x0000,0x000F,


0xCFC0,0x3000,0x1010,0x1000,
	0x8000,0x00C0,0x1F1F,0x0004,0x8003,0xE3E0,0x0C00,0x0004,0x0000,0x0600,0xF8F8,0x0024,0x001F,0x18F8,0x0024,0x001F,0x1F00,0x6000,0x0020,0x0000,0x30C0,0x0000,0x4000,0x0060,0x0F8F,0x8002,0x4600,0x0002,0x0000,0x0300,0x7C7C,0x0012,0x000F,0x8C7C,0x0012,0x000F,


0x8F80,0x3000,0x0010,0x0000,
	0x8000,0x0000,0x0DFF,0x8008,0x4007,0xFEC0,0x0000,0x0004,0x0000,0x0000,0x6FFC,0x0042,0x003F,0xFFFC,0x0042,0x003F,0xF600,0x0000,0x0020,0x0000,0x0000,0x0000,0x4000,0x0000,0x06FF,0xC004,0x2000,0x0002,0x0000,0x0000,0x37FE,0x0021,0x001F,0xFFFE,0x0021,0x001F,


0xFB00,0x0000,0x0010,0x0000,
	0xC000,0x0000,0x007F,0x0000,0x0003,0xF800,0x0000,0x000E,0x0000,0x0000,0x03F8,0x0000,0x001F,0xC3F8,0x0000,0x001F,0xC000,0x0000,0x0070,0x0000,0x0000,0x0000,0xE000,0x0000,0x003F,0x8000,0x0000,0x0007,0x0000,0x0000,0x01FC,0x0000,0x000F,0xE1FC,0x0000,0x000F,


0xE000,0x0000,0x0038,0x0000,
	0xE000,0x0180,0x001F,0x0000,0x0003,0xE000,0x0600,0x001F,0x0000,0x0C00,0x00F8,0x0000,0x001F,0x00F8,0x0000,0x001F,0x0000,0x3000,0x00F8,0x0000,0x6060,0x0001,0xF000,0x00C0,0x000F,0x8000,0x0300,0x000F,0x8000,0x0600,0x007C,0x0000,0x000F,0x807C,0x0000,0x000F,


0x8000,0x1800,0x007C,0x0000,
	0xE000,0x0380,0x000F,0x0010,0x2003,0xC000,0x0700,0x001F,0x0000,0x1C00,0x0078,0x0081,0x001E,0x0078,0x0081,0x001E,0x0000,0x3800,0x00F8,0x0000,0xE070,0x0001,0xF000,0x01C0,0x0007,0x8008,0x0380,0x000F,0x8000,0x0E00,0x003C,0x0040,0x800F,0x003C,0x0040,0x800F,


0x0000,0x1C00,0x007C,0x0000,
	0xE004,0x0300,0x000F,0x8008,0x4007,0xC000,0x0300,0x801F,0x0020,0x1800,0x007C,0x0042,0x003E,0x007C,0x0042,0x003E,0x0000,0x1804,0x00F8,0x0100,0xC030,0x0801,0xF002,0x0180,0x0007,0xC004,0x2180,0x400F,0x8010,0x0C00,0x003E,0x0021,0x001F,0x003E,0x0021,0x001F,


0x0000,0x0C02,0x007C,0x0080,
	0xF004,0x0000,0x000B,0x8008,0x4007,0x4000,0x0000,0x803F,0x8020,0x0000,0x005C,0x0042,0x003A,0x005C,0x0042,0x003A,0x0000,0x0004,0x01FC,0x0100,0x0000,0x0803,0xF802,0x0000,0x0005,0xC004,0x2000,0x401F,0xC010,0x0000,0x002E,0x0021,0x001D,0x002E,0x0021,0x001D,


0x0000,0x0002,0x00FE,0x0080,
	0xB00E,0x0000,0x0003,0x8000,0x0007,0x0000,0x0001,0xC035,0x8070,0x0000,0x001C,0x0000,0x0038,0x001C,0x0000,0x0038,0x0000,0x000E,0x01AC,0x0380,0x0000,0x1C03,0x5807,0x0000,0x0001,0xC000,0x0000,0xE01A,0xC038,0x0000,0x000E,0x0000,0x001C,0x000E,0x0000,0x001C,


0x0000,0x0007,0x00D6,0x01C0,
	0x180C,0x0000,0x0021,0x8010,0x2006,0x1000,0x0000,0xC060,0xC060,0x0000,0x010C,0x0081,0x0030,0x810C,0x0081,0x0030,0x8000,0x0006,0x0306,0x0300,0x0000,0x0C06,0x0C06,0x0000,0x0010,0xC008,0x0000,0x6030,0x6030,0x0000,0x0086,0x0040,0x8018,0x4086,0x0040,0x8018,


0x4000,0x0003,0x0183,0x0180,
	0x183E,0x0000,0x0039,0xC070,0x380E,0x7000,0x0001,0xF060,0xC1F0,0x0000,0x01CE,0x0381,0xC073,0x81CE,0x0381,0xC073,0x8000,0x000F,0x8306,0x0F80,0x0000,0x1F06,0x0C1F,0x0000,0x001C,0xE038,0x0000,0xF830,0x60F8,0x0000,0x00E7,0x01C0,0xE039,0xC0E7,0x01C0,0xE039,


0xC000,0x0007,0xC183,0x07C0,
	0x083F,0x8000,0x001F,0xC020,0x100F,0xE000,0x0007,0xF040,0x41FC,0x0000,0x00FE,0x0100,0x807F,0x00FE,0x0100,0x807F,0x0000,0x003F,0x8202,0x0FE0,0x0000,0x7F04,0x041F,0xC000,0x000F,0xE010,0x0003,0xF820,0x20FE,0x0000,0x007F,0x0080,0x403F,0x807F,0x0080,0x403F,


0x8000,0x001F,0xC101,0x07F0,
	0x007F,0xA000,0x0007,0xE000,0x001F,0x8000,0x0017,0xF800,0x03FD,0x0000,0x003F,0x0000,0x00FC,0x003F,0x0000,0x00FC,0x0000,0x00BF,0xC000,0x1FE8,0x0001,0x7F80,0x003F,0xD000,0x0003,0xF000,0x000B,0xFC00,0x01FE,0x8000,0x001F,0x8000,0x007E,0x001F,0x8000,0x007E,


0x0000,0x005F,0xE000,0x0FF4,
	0x0053,0xE000,0x0001,0xE180,0x061E,0x0000,0x001F,0x2800,0x029F,0x0000,0x000F,0x0C00,0x30F0,0x000F,0x0C00,0x30F0,0x0000,0x00F9,0x4000,0x14F8,0x0001,0xF280,0x0029,0xF000,0x0000,0xF0C0,0x000F,0x9400,0x014F,0x8000,0x0007,0x8600,0x1878,0x0007,0x8600,0x1878,


0x0000,0x007C,0xA000,0x0A7C,
	0x0001,0xE000,0x0000,0xE600,0x019C,0x0000,0x001E,0x0000,0x000F,0x0000,0x0007,0x3000,0x0CE0,0x0007,0x3000,0x0CE0,0x0000,0x00F0,0x0000,0x0078,0x0001,0xE000,0x0000,0xF000,0x0000,0x7300,0x000F,0x0000,0x0007,0x8000,0x0003,0x9800,0x0670,0x0003,0x9800,0x0670,


0x0000,0x0078,0x0000,0x003C,
	0x0003,0xC000,0x0000,0xC600,0x018C,0x0000,0x000F,0x0000,0x001E,0x0000,0x0006,0x3000,0x0C60,0x0006,0x3000,0x0C60,0x0000,0x0078,0x0000,0x00F0,0x0000,0xF000,0x0001,0xE000,0x0000,0x6300,0x0007,0x8000,0x000F,0x0000,0x0003,0x1800,0x0630,0x0003,0x1800,0x0630,


0x0000,0x003C,0x0000,0x0078,
	0x0000,0xE000,0x0000,0x6600,0x0198,0x0000,0x001C,0x0000,0x0007,0x0000,0x0003,0x3000,0x0CC0,0x0003,0x3000,0x0CC0,0x0000,0x00E0,0x0000,0x0038,0x0001,0xC000,0x0000,0x7000,0x0000,0x3300,0x000E,0x0000,0x0003,0x8000,0x0001,0x9800,0x0660,0x0001,0x9800,0x0660,


0x0000,0x0070,0x0000,0x001C,
	0x8000,0x0000,0x0000,0x6600,0x0198,0x0000,0x0000,0x0004,0x0000,0x0000,0x0003,0x3000,0x0CC0,0x0003,0x3000,0x0CC0,0x0000,0x0000,0x0020,0x0000,0x0000,0x0000,0x4000,0x0000,0x0000,0x3300,0x0000,0x0002,0x0000,0x0000,0x0001,0x9800,0x0660,0x0001,0x9800,0x0660,


0x0000,0x0000,0x0010,0x0000,
	0x8000,0x0800,0x6000,0x2E00,0x01D0,0x0018,0x0040,0x0004,0x0000,0x4003,0x0001,0x7000,0x0E80,0x0001,0x7000,0x0E80,0x00C0,0x0200,0x0020,0x0002,0x0004,0x0000,0x4000,0x0400,0x3000,0x1700,0x0020,0x0002,0x0000,0x2001,0x8000,0xB800,0x0740,0x0000,0xB800,0x0740,


0x0060,0x0100,0x0010,0x0001,
	0x0000,0x0000,0x6000,0x1E00,0x01E0,0x0018,0x0000,0x0000,0x0000,0x0003,0x0000,0xF000,0x0F00,0x0000,0xF000,0x0F00,0x00C0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x3000,0x0F00,0x0000,0x0000,0x0000,0x0001,0x8000,0x7800,0x0780,0x0000,0x7800,0x0780,


0x0060,0x0000,0x0000,0x0000,
	0x0000,0x0010,0x0000,0x1E00,0x01E0,0x0000,0x2000,0x0000,0x0000,0x0080,0x0000,0xF000,0x0F00,0x0000,0xF000,0x0F00,0x0001,0x0000,0x0000,0x0000,0x0600,0x0000,0x0000,0x0008,0x0000,0x0F00,0x1000,0x0000,0x0000,0x0040,0x0000,0x7800,0x0780,0x0000,0x7800,0x0780,


0x0000,0x8000,0x0000,0x0000,
	0x0180,0x0018,0x0000,0x1E00,0x01E0,0x0000,0x6000,0x0600,0x0C00,0x00C0,0x0000,0xF000,0x0F00,0x0000,0xF000,0x0F00,0x0003,0x0000,0x3000,0x6000,0x0600,0x0060,0x00C0,0x000C,0x0000,0x0F00,0x1000,0x0300,0x0600,0x0060,0x0000,0x7800,0x0780,0x0000,0x7800,0x0780,


0x0001,0x8000,0x1800,0x3000,
	0x0080,0x0010,0x0000,0x1E00,0x01E0,0x0000,0x2000,0x0400,0x0400,0x0080,0x0000,0xF000,0x0F00,0x0000,0xF000,0x0F00,0x0001,0x0000,0x2000,0x2000,0x0600,0x0040,0x0040,0x0008,0x0000,0x0F00,0x1000,0x0200,0x0200,0x0040,0x0000,0x7800,0x0780,0x0000,0x7800,0x0780,


0x0000,0x8000,0x1000,0x1000,
	0x8080,0x0030,0x0000,0x0E00,0x01C0,0x0000,0x3000,0x0404,0x0400,0x0180,0x0000,0x7000,0x0E00,0x0000,0x7000,0x0E00,0x0001,0x8000,0x2020,0x2000,0x0F00,0x0040,0x4040,0x0018,0x0000,0x0700,0x1800,0x0202,0x0200,0x00C0,0x0000,0x3800,0x0700,0x0000,0x3800,0x0700,


0x0000,0xC000,0x1010,0x1000,
	0x8080,0x0400,0x0000,0x0F00,0x03C0,0x0000,0x0080,0x0404,0x0400,0x2000,0x0000,0x7800,0x1E00,0x0000,0x7800,0x1E00,0x0000,0x0400,0x2020,0x2001,0x0008,0x0040,0x4040,0x0200,0x0000,0x0780,0x0040,0x0202,0x0200,0x1000,0x0000,0x3C00,0x0F00,0x0000,0x3C00,0x0F00,


0x0000,0x0200,0x1010,0x1000,
	0x8000,0x0000,0x0000,0x0F00,0x03C0,0x0000,0x0000,0x0004,0x0000,0x0000,0x0000,0x7800,0x1E00,0x0000,0x7800,0x1E00,0x0000,0x0000,0x0020,0x0000,0x0000,0x0000,0x4000,0x0000,0x0000,0x0780,0x0000,0x0002,0x0000,0x0000,0x0000,0x3C00,0x0F00,0x0000,0x3C00,0x0F00,


0x0000,0x0000,0x0010,0x0000,
	0x8000,0x0000,0x0000,0x0F80,0x07C0,0x0000,0x0000,0x0004,0x0000,0x0000,0x0000,0x7C00,0x3E00,0x0000,0x7C00,0x3E00,0x0000,0x0000,0x0020,0x0000,0x0000,0x0000,0x4000,0x0000,0x0000,0x07C0,0x0000,0x0002,0x0000,0x0000,0x0000,0x3E00,0x1F00,0x0000,0x3E00,0x1F00,


0x0000,0x0000,0x0010,0x0000,
	0x8000,0x1100,0x0000,0x0F80,0x07C0,0x0000,0x0220,0x0004,0x0000,0x8800,0x0000,0x7C00,0x3E00,0x0000,0x7C00,0x3E00,0x0000,0x1100,0x0020,0x0004,0x4022,0x0000,0x4000,0x0880,0x0000,0x07C0,0x0110,0x0002,0x0000,0x4400,0x0000,0x3E00,0x1F00,0x0000,0x3E00,0x1F00,


0x0000,0x0880,0x0010,0x0002,
	0x8000,0x137E,0x0000,0x0880,0x0440,0x0001,0xFB20,0x0004,0x0000,0x9BF0,0x0000,0x4400,0x2200,0x0000,0x4400,0x2200,0x000F,0xD900,0x0020,0x0004,0xDFB2,0x0000,0x4000,0x09BF,0x0000,0x0440,0x1D90,0x0002,0x0000,0x4DF8,0x0000,0x2200,0x1100,0x0000,0x2200,0x1100,


0x0007,0xEC80,0x0010,0x0002,
	0x8000,0x3206,0x0000,0x0000,0x0000,0x0001,0x8130,0x0004,0x0001,0x9030,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x000C,0x0980,0x0020,0x000C,0x8013,0x0000,0x4000,0x1903,0x0000,0x0000,0x0098,0x0002,0x0000,0xC818,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0006,0x04C0,0x0010,0x0006,
	0x0000,0x0300,0x0000,0x0000,0x0000,0x0000,0x0300,0x0000,0x0000,0x1800,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x1800,0x0000,0x0000,0xC030,0x0000,0x0000,0x0180,0x0000,0x0000,0x0180,0x0000,0x0000,0x0C00,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0000,0x0C00,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0400,0x0080,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x2000,0x0400,0x0000,0x2000,0x0400,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0200,0x0000,0x0000,0x0000,0x0000,0x0000,0x1000,0x0200,0x0000,0x1000,0x0200,


0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0080,0x0100,0x0200,0x0400,0x0000,0x0000,0x0000,0x0000,0x0400,0x0800,0x1000,0x2400,0x0800,0x1000,0x2000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0040,0x0080,0x0000,0x0000,0x0000,0x0000,0x0200,0x0400,0x0800,0x1200,0x0400,0x0800,


0x1000,0x0000,0x0000,0x0000,
	0x8000,0x1800,0x0020,0x0108,0x4200,0x1000,0x0060,0x0004,0x0000,0xC000,0x0100,0x0842,0x1000,0x8100,0x0842,0x1000,0x8000,0x0300,0x0020,0x0006,0x0006,0x0000,0x4000,0x0C00,0x0010,0x0084,0x2030,0x0002,0x0000,0x6000,0x0080,0x0421,0x0800,0x4080,0x0421,0x0800,


0x4000,0x0180,0x0010,0x0003,
	0x8002,0x1800,0x0000,0x0018,0x6000,0x0000,0x0061,0x0004,0x0010,0xC000,0x0000,0x00C3,0x0000,0x0000,0x00C3,0x0000,0x0000,0x0308,0x0020,0x0086,0x0006,0x1000,0x4001,0x0C00,0x0000,0x000C,0x2030,0x8002,0x0008,0x6000,0x0000,0x0061,0x8000,0x0000,0x0061,0x8000,


0x0000,0x0184,0x0010,0x0043,
	0x8000,0x1800,0x0000,0x0018,0x6000,0x0000,0x0060,0x0004,0x0000,0xC000,0x0000,0x00C3,0x0000,0x0000,0x00C3,0x0000,0x0000,0x0300,0x0020,0x0006,0x0006,0x0000,0x4000,0x0C00,0x0000,0x000C,0x2030,0x0002,0x0000,0x6000,0x0000,0x0061,0x8000,0x0000,0x0061,0x8000,


0x0000,0x0180,0x0010,0x0003,
	0x8000,0x1A00,0x0008,0x413C,0xF208,0x4000,0x0160,0x0004,0x0000,0xD000,0x0042,0x09E7,0x9042,0x0042,0x09E7,0x9042,0x0000,0x0B00,0x0020,0x0006,0x8016,0x0000,0x4000,0x0D00,0x0004,0x209E,0x60B0,0x0002,0x0000,0x6800,0x0021,0x04F3,0xC821,0x0021,0x04F3,0xC821,


0x0000,0x0580,0x0010,0x0003,
	0x0001,0xF800,0x0000,0xFC3C,0xF0FC,0x0000,0x007E,0x0000,0x000F,0xC000,0x0007,0xE1E7,0x87E0,0x0007,0xE1E7,0x87E0,0x0000,0x03F0,0x0000,0x007E,0x0007,0xE000,0x0000,0xFC00,0x0000,0x7E1E,0x603F,0x0000,0x0007,0xE000,0x0003,0xF0F3,0xC3F0,0x0003,0xF0F3,0xC3F0,


0x0000,0x01F8,0x0000,0x003F,
	0x0001,0xF000,0x001D,0xFC3F,0xF0FE,0xE000,0x003E,0x0000,0x000F,0x8000,0x00EF,0xE1FF,0x87F7,0x00EF,0xE1FF,0x87F7,0x0000,0x01F0,0x0000,0x007C,0x0003,0xE000,0x0000,0xF800,0x000E,0xFE1F,0xE01F,0x0000,0x0007,0xC000,0x0077,0xF0FF,0xC3FB,0x8077,0xF0FF,0xC3FB,


0x8000,0x00F8,0x0000,0x003E,
	0x8000,0x7000,0x003F,0xE03F,0xF01F,0xF000,0x0038,0x0004,0x0003,0x8000,0x01FF,0x01FF,0x80FF,0x81FF,0x01FF,0x80FF,0x8000,0x01C0,0x0020,0x001C,0x0003,0x8000,0x4000,0x3800,0x001F,0xF01F,0xE01C,0x0002,0x0001,0xC000,0x00FF,0x80FF,0xC07F,0xC0FF,0x80FF,0xC07F,


0xC000,0x00E0,0x0010,0x000E,
	0xE000,0x3000,0x011F,0xC03F,0xF00F,0xE200,0x0030,0x001F,0x0001,0x8000,0x08FE,0x01FF,0x807F,0x18FE,0x01FF,0x807F,0x1000,0x0180,0x00F8,0x000C,0x0003,0x0001,0xF000,0x1800,0x008F,0xE01F,0xE018,0x000F,0x8000,0xC000,0x047F,0x00FF,0xC03F,0x8C7F,0x00FF,0xC03F,


0x8800,0x00C0,0x007C,0x0006,
	0xE000,0x0000,0x011F,0x803F,0xF007,0xE200,0x0000,0x001F,0x0000,0x0000,0x08FC,0x01FF,0x803F,0x18FC,0x01FF,0x803F,0x1000,0x0000,0x00F8,0x0000,0x0000,0x0001,0xF000,0x0000,0x008F,0xC01F,0xE000,0x000F,0x8000,0x0000,0x047E,0x00FF,0xC01F,0x8C7E,0x00FF,0xC01F,


0x8800,0x0000,0x007C,0x0000,
	0x1000,0x0000,0x001F,0x803F,0xF007,0xE000,0x0000,0x0020,0x8000,0x0000,0x00FC,0x01FF,0x803F,0x00FC,0x01FF,0x803F,0x0000,0x0000,0x0104,0x0000,0x0000,0x0002,0x0800,0x0000,0x000F,0xC01F,0xE000,0x0010,0x4000,0x0000,0x007E,0x00FF,0xC01F,0x807E,0x00FF,0xC01F,


0x8000,0x0000,0x0082,0x0000,
	0x0400,0x0000,0x061F,0x003F,0xF003,0xE180,0x0000,0x0080,0x2000,0x0000,0x30F8,0x01FF,0x801F,0x00F8,0x01FF,0x801F,0x0C00,0x0000,0x0401,0x0000,0x0000,0x0008,0x0200,0x0000,0x030F,0x801F,0xE000,0x0040,0x1000,0x0000,0x187C,0x00FF,0xC00F,0x807C,0x00FF,0xC00F,


0x8600,0x0000,0x0200,0x8000,
	0x8401,0x0000,0x061F,0x403F,0xF00B,0xE180,0x0002,0x0084,0x2008,0x0000,0x30FA,0x01FF,0x805F,0x00FA,0x01FF,0x805F,0x0C00,0x0010,0x0421,0x0040,0x0000,0x2008,0x4200,0x8000,0x030F,0xA01F,0xC001,0x0042,0x1004,0x0000,0x187D,0x00FF,0xC02F,0x807D,0x00FF,0xC02F,


0x8600,0x0008,0x0210,0x8020,
	0x0000,0x0000,0x0C1E,0x006F,0xD801,0xE0C0,0x0000,0x0000,0x0000,0x0000,0x60F0,0x037E,0xC00F,0x00F0,0x037E,0xC00F,0x0600,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x060F,0x0037,0xE000,0x0000,0x0000,0x0000,0x3078,0x01BF,0x6007,0x8078,0x01BF,0x6007,


0x8300,0x0000,0x0000,0x0000,
	0x0080,0x0000,0x001E,0x006F,0xD801,0xE000,0x0000,0x0400,0x0400,0x0000,0x00F0,0x037E,0xC00F,0x00F0,0x037E,0xC00F,0x0000,0x0000,0x2000,0x2000,0x0000,0x0040,0x0040,0x0000,0x000F,0x0037,0xE000,0x0200,0x0200,0x0000,0x0078,0x01BF,0x6007,0x8078,0x01BF,0x6007,


0x8000,0x0000,0x1000,0x1000,
	0x0080,0x1000,0x1003,0x004F,0xC803,0x0020,0x0020,0x0400,0x0400,0x8000,0x8018,0x027E,0x4018,0x0018,0x027E,0x4018,0x0100,0x0100,0x2000,0x2004,0x0002,0x0040,0x0040,0x0800,0x0801,0x8027,0xE010,0x0200,0x0200,0x4000,0x400C,0x013F,0x200C,0x000C,0x013F,0x200C,


0x0080,0x0080,0x1000,0x1002,
	0x0040,0x7000,0x1001,0x800F,0xC006,0x0020,0x0038,0x0800,0x0203,0x8000,0x800C,0x007E,0x0030,0x000C,0x007E,0x0030,0x0100,0x01C0,0x4000,0x101C,0x0003,0x8080,0x0020,0x3800,0x0800,0xC007,0xE01C,0x0400,0x0101,0xC000,0x4006,0x003F,0x0018,0x0006,0x003F,0x0018,


0x0080,0x00E0,0x2000,0x080E,
	0x803F,0x1801,0x1005,0xE007,0x801E,0x8022,0x0063,0xF004,0x01F8,0xC008,0x802F,0x003C,0x00F4,0x002F,0x003C,0x00F4,0x0110,0x031F,0x8020,0x0FC6,0x0006,0x3F00,0x401F,0x8C00,0x8802,0xF003,0xC031,0xF802,0x00FC,0x6004,0x4017,0x801E,0x007A,0x0017,0x801E,0x007A,


0x0088,0x018F,0xC010,0x07E3,
	0x803F,0x0801,0x6004,0xE007,0x801C,0x801A,0x0043,0xF004,0x01F8,0x400B,0x0027,0x003C,0x00E4,0x0027,0x003C,0x00E4,0x00D0,0x021F,0x8020,0x0FC2,0x0004,0x3F00,0x401F,0x8400,0xB002,0x7003,0xC021,0xF802,0x00FC,0x2005,0x8013,0x801E,0x0072,0x0013,0x801E,0x0072,


0x0068,0x010F,0xC010,0x07E1,
	0x801F,0xEC01,0xE004,0xF003,0x003C,0x801E,0x00DF,0xE004,0x00FF,0x600F,0x0027,0x8018,0x01E4,0x0027,0x8018,0x01E4,0x00F0,0x06FF,0x0020,0x07FB,0x000D,0xFE00,0x400F,0xF600,0xF002,0x7801,0x806F,0xF002,0x007F,0xB007,0x8013,0xC00C,0x00F2,0x0013,0xC00C,0x00F2,


0x0078,0x037F,0x8010,0x03FD,
	0x800F,0xBE00,0xF002,0x7C03,0x00F9,0x003C,0x01F7,0xC004,0x007D,0xF007,0x8013,0xE018,0x07C8,0x0013,0xE018,0x07C8,0x01E0,0x0FBE,0x0020,0x03EF,0x801F,0x7C00,0x4007,0xDF00,0x7801,0x3E01,0x80FB,0xE002,0x003E,0xF803,0xC009,0xF00C,0x03E4,0x0009,0xF00C,0x03E4,


0x00F0,0x07DF,0x0010,0x01FF,
	0x8003,0x3E00,0xF003,0x3C00,0x00F3,0x003C,0x01F3,0x0004,0x0019,0xF007,0x8019,0xE000,0x0798,0x0019,0xE000,0x0798,0x01E0,0x0F98,0x0020,0x00CF,0x801F,0x3000,0x4001,0x9F00,0x7801,0x9E00,0x00F9,0x8002,0x000C,0xF803,0xC00C,0xF000,0x03CC,0x000C,0xF000,0x03CC,


0x00F0,0x07CC,0x0010,0x007F,
	0x8001,0x8200,0xE003,0x3E03,0x01F3,0x001C,0x0106,0x0004,0x000C,0x1007,0x0019,0xF018,0x0F98,0x0019,0xF018,0x0F98,0x00E0,0x0830,0x0020,0x0060,0x8010,0x6000,0x4000,0xC100,0x7001,0x9F01,0x8083,0x0002,0x0006,0x0803,0x800C,0xF80C,0x07CC,0x000C,0xF80C,0x07CC,


0x0070,0x0418,0x0010,0x003F,
	0x8000,0xC000,0x6001,0x0E07,0x81C2,0x0018,0x000C,0x0004,0x0006,0x0003,0x0008,0x703C,0x0E10,0x0008,0x703C,0x0E10,0x00C0,0x0060,0x0020,0x0030,0x0000,0xC000,0x4000,0x6000,0x3000,0x8703,0xC006,0x0002,0x0003,0x0001,0x8004,0x381E,0x0708,0x0004,0x381E,0x0708,


0x0060,0x0030,0x0010,0x001F,
	0x8000,0xC040,0x6001,0x0207,0x8102,0x0018,0x080C,0x0004,0x0006,0x0203,0x0008,0x103C,0x0810,0x0008,0x103C,0x0810,0x00C0,0x4060,0x0020,0x0030,0x1080,0xC000,0x4000,0x6020,0x3000,0x8103,0xC406,0x0002,0x0003,0x0101,0x8004,0x081E,0x0408,0x0004,0x081E,0x0408,


0x0060,0x2030,0x0010,0x001F,
	0x8000,0x6040,0x3001,0x8100,0x0206,0x0030,0x0818,0x0004,0x0003,0x0201,0x800C,0x0800,0x1030,0x000C,0x0800,0x1030,0x0180,0x40C0,0x0020,0x0018,0x1081,0x8000,0x4000,0x3020,0x1800,0xC080,0x040C,0x0002,0x0001,0x8100,0xC006,0x0400,0x0818,0x0006,0x0400,0x0818,


0x00C0,0x2060,0x0010,0x000F,
	0x8000,0x2000,0x0003,0x8100,0x0207,0x0000,0x0010,0x0004,0x0001,0x0000,0x001C,0x0800,0x1038,0x001C,0x0800,0x1038,0x0000,0x0080,0x0020,0x0008,0x0001,0x0000,0x4000,0x1000,0x0001,0xC080,0x0008,0x0002,0x0000,0x8000,0x000E,0x0400,0x081C,0x000E,0x0400,0x081C,


0x0000,0x0040,0x0010,0x0007,
	0xC000,0x3000,0x0003,0x8000,0x0007,0x0000,0x0030,0x000E,0x0001,0x8000,0x001C,0x0000,0x0038,0x001C,0x0000,0x0038,0x0000,0x0180,0x0070,0x000C,0x0003,0x0000,0xE000,0x1800,0x0001,0xC000,0x0018,0x0007,0x0000,0xC000,0x000E,0x0000,0x001C,0x000E,0x0000,0x001C,


0x0000,0x00C0,0x0038,0x0007,
	0xC000,0x1020,0x0403,0x8000,0x0007,0x0080,0x1020,0x000E,0x0000,0x8100,0x201C,0x0000,0x0038,0x001C,0x0000,0x0038,0x0400,0x8100,0x0070,0x0004,0x0902,0x0000,0xE000,0x0810,0x0201,0xC000,0x0810,0x0007,0x0000,0x4080,0x100E,0x0000,0x001C,0x000E,0x0000,0x001C,


0x0200,0x4080,0x0038,0x0003,
	0xC000,0x0C00,0x0400,0x2000,0x0010,0x0080,0x00C0,0x000E,0x0000,0x6000,0x2001,0x0000,0x0080,0x0001,0x0000,0x0080,0x0400,0x0600,0x0070,0x0003,0x000C,0x0000,0xE000,0x0600,0x0200,0x1000,0x0060,0x0007,0x0000,0x3000,0x1000,0x8000,0x0040,0x0000,0x8000,0x0040,


0x0200,0x0300,0x0038,0x0001,
	0xE000,0x0014,0x0002,0x0000,0x0001,0x0000,0xA000,0x001F,0x0000,0x00A0,0x0010,0x0000,0x0008,0x0010,0x0000,0x0008,0x0005,0x0000,0x00F8,0x0000,0x0600,0x0001,0xF000,0x000A,0x0001,0x0000,0x1000,0x000F,0x8000,0x0050,0x0008,0x0000,0x0004,0x0008,0x0000,0x0004,


0x0002,0x8000,0x007C,0x0000,
	0xE000,0x001E,0x8200,0x0000,0x0000,0x0105,0xE000,0x001F,0x0000,0x00F4,0x1000,0x0000,0x0000,0x0000,0x0000,0x0000,0x082F,0x0000,0x00F8,0x0000,0x0600,0x0001,0xF000,0x000F,0x4100,0x0000,0x1000,0x000F,0x8000,0x007A,0x0800,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0417,0x8000,0x007C,0x0000,
	0xF000,0x000F,0xFF80,0x0000,0x0000,0x07FF,0xC000,0x003F,0x8000,0x007F,0xFC00,0x0000,0x0000,0x3C00,0x0000,0x0000,0x3FFE,0x0000,0x01FC,0x0000,0x0000,0x0003,0xF800,0x0007,0xFFC0,0x0000,0x0000,0x001F,0xC000,0x003F,0xFE00,0x0000,0x0000,0x1E00,0x0000,0x0000,


0x1FFF,0x0000,0x00FE,0x0000,
	0xF000,0x000F,0xFFC0,0x0000,0x0000,0x0FFF,0xC000,0x003F,0x8000,0x007F,0xFE00,0x0000,0x0000,0x7E00,0x0000,0x0000,0x7FFE,0x0000,0x01FC,0x0000,0x0000,0x0003,0xF800,0x0007,0xFFE0,0x0000,0x0000,0x001F,0xC000,0x003F,0xFF00,0x0000,0x0000,0x3F00,0x0000,0x0000,


0x3FFF,0x0000,0x00FE,0x0000,
	0x9800,0x00CE,0x0FC0,0x0000,0x0000,0x0FC1,0xCC00,0x0064,0xC000,0x0670,0x7E00,0x0000,0x0000,0x7E00,0x0000,0x0000,0x7E0E,0x6000,0x0326,0x0000,0x30C0,0x0006,0x4C00,0x0067,0x07E0,0x0000,0x0600,0x0032,0x6000,0x0338,0x3F00,0x0000,0x0000,0x3F00,0x0000,0x0000,


0x3F07,0x3000,0x0193,0x0000,
	0x9800,0x0066,0x0780,0x0800,0x0040,0x0781,0x9800,0x0064,0xC000,0x0330,0x3C00,0x4000,0x0200,0x3C00,0x4000,0x0200,0x3C0C,0xC000,0x0326,0x0000,0x1980,0x0006,0x4C00,0x0033,0x03C0,0x0400,0x0C00,0x0032,0x6000,0x0198,0x1E00,0x2000,0x0100,0x1E00,0x2000,0x0100,


0x1E06,0x6000,0x0193,0x0000,
	0x0800,0x0074,0x0701,0x0800,0x0042,0x0380,0xB800,0x0040,0x4000,0x03A0,0x3808,0x4000,0x0210,0x1808,0x4000,0x0210,0x1C05,0xC000,0x0202,0x0000,0x1F80,0x0004,0x0400,0x003A,0x0380,0x8400,0x1C00,0x0020,0x2000,0x01D0,0x1C04,0x2000,0x0108,0x0C04,0x2000,0x0108,


0x0E02,0xE000,0x0101,0x0000,
	0x1000,0x1830,0x0300,0x0000,0x0000,0x0300,0x3060,0x0020,0x8000,0xC180,0x1800,0x0000,0x0000,0x1800,0x0000,0x0000,0x1801,0x8300,0x0104,0x0006,0x0F06,0x0002,0x0800,0x0C18,0x0180,0x0000,0x1830,0x0010,0x4000,0x60C0,0x0C00,0x0000,0x0000,0x0C00,0x0000,0x0000,


0x0C00,0xC180,0x0082,0x0003,
	0x8000,0x1C30,0x0180,0x0000,0x0000,0x0600,0x30E0,0x0004,0x0000,0xE180,0x0C00,0x0000,0x0000,0x3C00,0x0000,0x0000,0x3001,0x8700,0x0020,0x0007,0x0F0E,0x0000,0x4000,0x0E18,0x00C0,0x0000,0x1870,0x0002,0x0000,0x70C0,0x0600,0x0000,0x0000,0x1E00,0x0000,0x0000,


0x1800,0xC380,0x0010,0x0003,
	0x0000,0x0000,0x0060,0x0000,0x0000,0x1800,0x0000,0x0000,0x0000,0x0000,0x0300,0x0000,0x0000,0xC300,0x0000,0x0000,0xC000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0030,0x0000,0x0000,0x0000,0x0000,0x0000,0x0180,0x0000,0x0000,0x6180,0x0000,0x0000,


0x6000,0x0000,0x0000,0x0000,
	0x0000,0x0008,0x0060,0x0000,0x0000,0x1800,0x4000,0x0000,0x0000,0x0040,0x0300,0x0000,0x0000,0xC300,0x0000,0x0000,0xC002,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0004,0x0030,0x0000,0x0000,0x0000,0x0000,0x0020,0x0180,0x0000,0x0000,0x6180,0x0000,0x0000,


0x6001,0x0000,0x0000,0x0000,
	0x0000,0x2008,0x1000,0x0000,0x0000,0x0020,0x4010,0x0000,0x0001,0x0040,0x8000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0102,0x0080,0x0000,0x0008,0x0001,0x0000,0x0000,0x1004,0x0800,0x0000,0x0008,0x0000,0x0000,0x8020,0x4000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0081,0x0040,0x0000,0x0004,
	0x00C0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0C00,0x0600,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x6000,0x3000,0x0000,0x00C0,0x0060,0x0000,0x0000,0x0000,0x0000,0x0600,0x0300,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0000,0x0000,0x3000,0x1800,
	0x00EC,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0xDC00,0x0760,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0006,0xE000,0x3B00,0x0000,0x0DC0,0x0076,0x0000,0x0000,0x0000,0x0000,0x6E00,0x03B0,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0000,0x0003,0x7000,0x1D80,
	0x24FE,0x0000,0x0800,0x0004,0x8000,0x0040,0x0001,0xFC91,0x27F0,0x0000,0x4000,0x0024,0x0000,0x0000,0x0024,0x0000,0x0200,0x000F,0xE489,0x3F80,0x0000,0x1FC9,0x127F,0x0000,0x0400,0x0002,0x4000,0xFE48,0x93F8,0x0000,0x2000,0x0012,0x0000,0x0000,0x0012,0x0000,


0x0100,0x0007,0xF244,0x9FC0,
	0x2DF8,0x0000,0x8F00,0x0180,0x0600,0x03C4,0x0000,0x7ED1,0x6FC0,0x0004,0x7800,0x0C00,0x3000,0x1800,0x0C00,0x3000,0x1E20,0x0003,0xF68B,0x7E00,0x0000,0x07ED,0x16FC,0x0000,0x4780,0x00C0,0x0000,0x3F68,0xB7E0,0x0002,0x3C00,0x0600,0x1800,0x0C00,0x0600,0x1800,


0x0F10,0x0001,0xFB45,0xBF00,
	0x85F8,0x0000,0x0F80,0x00C0,0x0C00,0x07C0,0x0000,0x7E84,0x2FC0,0x0000,0x7C00,0x0600,0x6000,0x3C00,0x0600,0x6000,0x3E00,0x0003,0xF421,0x7E00,0x0000,0x07E8,0x42FC,0x0000,0x07C0,0x0060,0x0000,0x3F42,0x17E0,0x0000,0x3E00,0x0300,0x3000,0x1E00,0x0300,0x3000,


0x1F00,0x0001,0xFA10,0xBF00,
	0x81F8,0x0020,0x1B80,0x00E0,0x1C00,0x0760,0x1000,0x7E04,0x0FC0,0x0100,0xDC00,0x0700,0xE000,0x3C00,0x0700,0xE000,0x3B00,0x8003,0xF020,0x7E00,0x0900,0x07E0,0x40FC,0x0010,0x0DC0,0x0070,0x0800,0x3F02,0x07E0,0x0080,0x6E00,0x0380,0x7000,0x1E00,0x0380,0x7000,


0x1D80,0x4001,0xF810,0x3F00,
	0xC0F8,0x0030,0x3380,0x0060,0x1800,0x0730,0x3000,0x7C0E,0x07C0,0x0181,0x9C00,0x0300,0xC000,0x3C00,0x0300,0xC000,0x3981,0x8003,0xE070,0x3E00,0x0F00,0x07C0,0xE07C,0x0018,0x19C0,0x0030,0x1800,0x3E07,0x03E0,0x00C0,0xCE00,0x0180,0x6000,0x1E00,0x0180,0x6000,


0x1CC0,0xC001,0xF038,0x1F00,
	0x80FD,0x0010,0x1300,0x0023,0x1000,0x0320,0x2002,0xFC04,0x07E8,0x0080,0x9800,0x0118,0x8000,0x1800,0x0118,0x8000,0x1901,0x0017,0xE020,0x3F40,0x0600,0x2FC0,0x407E,0x8008,0x0980,0x0011,0x9001,0x7E02,0x03F4,0x0040,0x4C00,0x008C,0x4000,0x0C00,0x008C,0x4000,


0x0C80,0x800B,0xF010,0x1FA0,
	0x80F9,0x8010,0x1200,0x003C,0xF000,0x0120,0x2006,0x7C04,0x07CC,0x0080,0x9000,0x01E7,0x8000,0x0000,0x01E7,0x8000,0x0901,0x0033,0xE020,0x3E60,0x0600,0x67C0,0x407C,0xC008,0x0900,0x001E,0x7003,0x3E02,0x03E6,0x0040,0x4800,0x00F3,0xC000,0x0000,0x00F3,0xC000,


0x0480,0x8019,0xF010,0x1F30,
	0x00E1,0x8010,0x1A00,0x003C,0xF000,0x0160,0x2006,0x1C00,0x070C,0x0080,0xD000,0x01E7,0x8000,0x0000,0x01E7,0x8000,0x0B01,0x0030,0xE000,0x3860,0x0600,0x61C0,0x0070,0xC008,0x0D00,0x001E,0x7003,0x0E00,0x0386,0x0040,0x6800,0x00F3,0xC000,0x0000,0x00F3,0xC000,


0x0580,0x8018,0x7000,0x1C30,
	0x00E3,0x800C,0x0000,0x007C,0xF800,0x0000,0xC007,0x1C00,0x071C,0x0060,0x0000,0x03E7,0xC000,0x0000,0x03E7,0xC000,0x0006,0x0038,0xE000,0x38E0,0x0000,0x71C0,0x0071,0xC006,0x0000,0x003E,0x6003,0x8E00,0x038E,0x0030,0x0000,0x01F3,0xE000,0x0000,0x01F3,0xE000,


0x0003,0x001C,0x7000,0x1C70,
	0x8060,0x0008,0x0000,0x007C,0xF800,0x0000,0x4000,0x1804,0x0300,0x0040,0x0000,0x03E7,0xC000,0x0000,0x03E7,0xC000,0x0002,0x0000,0xC020,0x1800,0x0000,0x0180,0x4030,0x0004,0x0000,0x003E,0x6000,0x0C02,0x0180,0x0020,0x0000,0x01F3,0xE000,0x0000,0x01F3,0xE000,


0x0001,0x0000,0x6010,0x0C00,
	0x8060,0x000E,0x0000,0x00D8,0x6C00,0x0001,0xC000,0x1804,0x0300,0x0070,0x0000,0x06C3,0x6000,0x0000,0x06C3,0x6000,0x000E,0x0000,0xC020,0x1800,0x0000,0x0180,0x4030,0x0007,0x0000,0x006C,0x2000,0x0C02,0x0180,0x0038,0x0000,0x0361,0xB000,0x0000,0x0361,0xB000,


0x0007,0x0000,0x6010,0x0C00,
	0x0060,0x000E,0x0400,0x0018,0x6000,0x0081,0xC000,0x1800,0x0300,0x0070,0x2000,0x00C3,0x0000,0x0000,0x00C3,0x0000,0x040E,0x0000,0xC000,0x1800,0x0000,0x0180,0x0030,0x0007,0x0200,0x000C,0x2000,0x0C00,0x0180,0x0038,0x1000,0x0061,0x8000,0x0000,0x0061,0x8000,


0x0207,0x0000,0x6000,0x0C00,
	0x0070,0x0018,0x0400,0x0208,0x4100,0x0080,0x6000,0x3800,0x0380,0x00C0,0x2000,0x1042,0x0800,0x0000,0x1042,0x0800,0x0403,0x0001,0xC000,0x1C00,0x0600,0x0380,0x0038,0x000C,0x0200,0x0104,0x3000,0x1C00,0x01C0,0x0060,0x1000,0x0821,0x0400,0x0000,0x0821,0x0400,


0x0201,0x8000,0xE000,0x0E00,
	0x0070,0x0018,0x0004,0x0200,0x0100,0x8000,0x6000,0x3800,0x0380,0x00C0,0x0020,0x1000,0x0804,0x0020,0x1000,0x0804,0x0003,0x0001,0xC000,0x1C00,0x0600,0x0380,0x0038,0x000C,0x0002,0x0100,0x1000,0x1C00,0x01C0,0x0060,0x0010,0x0800,0x0402,0x0010,0x0800,0x0402,


0x0001,0x8000,0xE000,0x0E00,
	0x0078,0x0000,0x0004,0x0700,0x0380,0x8000,0x0000,0x7800,0x03C0,0x0000,0x0020,0x3800,0x1C04,0x0020,0x3800,0x1C04,0x0000,0x0003,0xFFFF,0xFFFF,0xFFFF,0xFF80,0x003C,0x0000,0x0002,0x0380,0x0000,0x3C00,0x01E0,0x0000,0x0010,0x1C00,0x0E02,0x0010,0x1C00,0x0E02,


0x0000,0x0001,0xE000,0x0F00,
	0x003C,0x0000,0x0086,0x0600,0x0181,0x8400,0x0000,0xF000,0x01E0,0x0000,0x0430,0x3000,0x0C0C,0x2430,0x3000,0x0C0C,0x2000,0x0007,0xFFFF,0xFFFF,0xFFFF,0xFF00,0x001E,0x0000,0x0043,0x0300,0x0000,0x7800,0x00F0,0x0000,0x0218,0x1800,0x0606,0x1218,0x1800,0x0606,


0x1000,0x0003,0xC000,0x0780,
	0x003C,0x0070,0x0006,0x0000,0x0001,0x8000,0x3800,0xF000,0x01E0,0x0380,0x0030,0x0000,0x000C,0x0030,0x0000,0x000C,0x0001,0xC007,0xFFFF,0xFFFF,0xFFFF,0xFF00,0x001E,0x0038,0x0003,0x0000,0x1C00,0x7800,0x00F0,0x01C0,0x0018,0x0000,0x0006,0x0018,0x0000,0x0006,


0x0000,0xE003,0xC000,0x0780,
	0x0034,0x0070,0x300E,0x0400,0x0081,0xC030,0x3800,0xB000,0x01A0,0x0381,0x8070,0x2000,0x040E,0x0070,0x2000,0x040E,0x0181,0xC005,0xFFFF,0xFFFF,0xFFFF,0xFF00,0x001A,0x0038,0x1807,0x0200,0x1C00,0x5800,0x00D0,0x01C0,0xC038,0x1000,0x0207,0x0038,0x1000,0x0207,


0x00C0,0xE002,0xC000,0x0680,
	0x8014,0x0070,0x300E,0x3C00,0x00F1,0xC030,0x3800,0xA004,0x00A0,0x0381,0x8071,0xE000,0x078E,0x0071,0xE000,0x078E,0x0181,0xC005,0xFFFF,0xFFFF,0xFFFF,0xFE00,0x400A,0x0038,0x1807,0x1E00,0x1C00,0x5002,0x0050,0x01C0,0xC038,0xF000,0x03C7,0x0038,0xF000,0x03C7,


0x00C0,0xE002,0x8010,0x0280,
	0xC010,0x0070,0x0000,0x3003,0x0030,0x0000,0x3800,0x200E,0x0080,0x0380,0x0001,0x8018,0x0180,0x0001,0x8018,0x0180,0x0001,0xC001,0xFFFF,0xFFFF,0xFFFF,0xFE00,0xE008,0x0038,0x0000,0x1801,0x9C00,0x1007,0x0040,0x01C0,0x0000,0xC00C,0x00C0,0x0000,0xC00C,0x00C0,


0x0000,0xE000,0x8038,0x0200,
	0xE018,0x0070,0x0000,0x2003,0x0010,0x0000,0x3800,0x601F,0x00C0,0x0380,0x0001,0x0018,0x0080,0x0001,0x0018,0x0080,0x0001,0xC003,0xFFFF,0xFFFF,0xFFFF,0xFE01,0xF00C,0x0038,0x0000,0x1001,0x9C00,0x300F,0x8060,0x01C0,0x0000,0x800C,0x0040,0x0000,0x800C,0x0040,


0x0000,0xE001,0x807C,0x0300,
	0xE018,0x0044,0x0000,0x2003,0x0010,0x0000,0x8800,0x601F,0x00C0,0x0220,0x0001,0x0018,0x0080,0x0001,0x0018,0x0080,0x0004,0x4003,0xFFFF,0xFFFF,0xFFFF,0xFE01,0xF00C,0x0022,0x0000,0x1001,0x8400,0x300F,0x8060,0x0110,0x0000,0x800C,0x0040,0x0000,0x800C,0x0040,


0x0002,0x2001,0x807C,0x0300,
	0xF808,0x0004,0x0000,0x2000,0x0010,0x0000,0x8000,0x407F,0xC040,0x0020,0x0001,0x0000,0x0080,0x0001,0x0000,0x0080,0x0004,0x0003,0xFFFF,0xFFFF,0xFFFF,0xFC07,0xFC04,0x0002,0x0000,0x1000,0x0000,0x203F,0xE020,0x0010,0x0000,0x8000,0x0040,0x0000,0x8000,0x0040,


0x0002,0x0001,0x01FF,0x0100,
	0xFE04,0x0042,0x0000,0x1000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,


0x0000,0x0000,0x0000,0x0000
};

// PRIVATE STRUCTURES -----------------------------------------------------

struct Image IsoImage =
{	0, 0,
	11, ISOSQUAREY, 4,
	IsoImageData[0],
	0xff, 0x0,
	NULL
};
struct Image Image =
{	0, 0,
	SQUAREX, SQUAREY, 4,
	ImageData[0],
	0xff, 0x0,
	NULL
};
struct Image Logo =
{	0, 0,
	237, 72, 4,
	LogoData,
	0xff, 0x0,
	NULL
};
struct Image About =
{	0, 0,
	45, 20, 4,
	AboutData,
	0xff, 0x0,
	NULL
};
struct Image Background =
{	0, 0,
	640, 256, 1,
	BackgroundData,
	1, 12,
	NULL
};

struct EasyStruct EasyStruct =
{	sizeof(struct EasyStruct),
	0,
	TITLEBAR,
	"Fieldset/highscores have been modified.",
	"Continue|Cancel"
};

struct
{	WORD x, y;
	STRPTR text;
} about[ABOUTLINES + 1] =
{   {    80,        27,     "Worm Wars 5.21 for Amiga"          },
    {    80,        35,     "Wednesday 5 January 2000"          },
	{	 80,		43,		" 2000 Amigan Software"			},
    {    16,        59,		"Programming, Graphics:"        	},
    {    48,        67,		"James R. Jacobs"               	},
	{	 16,		83,		"Sound, Design, Playtesting:"		},
    {    48,        91,		"James R. Jacobs & Lisa N. Woon"	},
	{	 16,	   107,		"Project size:       bytes"			}
};
struct
{	UBYTE		scancode;
	BYTE		player, deltax, deltay, special;
	BOOL		down;
} key[NUMKEYS + 1] =
{	{Q,					0,	-1,	-1,	MOVE,		FALSE},
	{W,					0,	 0,	-1,	MOVE,		FALSE},
	{E,					0,	 1,	-1,	MOVE,		FALSE},
	{A,					0,	-1,	 0,	MOVE,		FALSE},
	{S,					0,	 0,	 1,	MOVE,		FALSE},
	{D,					0,	 1,	 0,	MOVE,		FALSE},
	{Z,					0,	-1,	 1,	MOVE,		FALSE},
	{KEY_X,				0,	 1,	 1,	MOVE,		FALSE},
	{C,					0,	 1,	 1,	MOVE,		FALSE},
	{NUMERICSEVEN,		1,	-1,	-1,	MOVE,		FALSE},
	{NUMERICEIGHT,		1,	 0,	-1,	MOVE,		FALSE},
	{NUMERICNINE,		1,	 1,	-1,	MOVE,		FALSE},
	{NUMERICFOUR,		1,	-1,	 0,	MOVE,		FALSE},
	{NUMERICFIVE,		1,	 0,	 1,	MOVE,		FALSE},
	{NUMERICSIX,		1,	 1,	 0,	MOVE,		FALSE},
	{NUMERICONE,		1,	-1,	 1,	MOVE,		FALSE},
	{NUMERICTWO,		1,	 0,	 1,	MOVE,		FALSE},
	{NUMERICTHREE,		1,	 1,	 1,	MOVE,		FALSE},
	{UP,				1,	 0,	-1,	ONEHUMAN,	FALSE},
	{DOWN,				1,	 0,	 1,	ONEHUMAN,	FALSE},
	{RIGHT,				1,	 1,	 0,	ONEHUMAN,	FALSE},
	{LEFT,				1,	-1,	 0,	ONEHUMAN,	FALSE},
	{SPACEBAR,			0,	 0,	 0,	AMMO,		FALSE},
	{NUMERICZERO,		1,	 0,	 0,	AMMO,		FALSE},
	{ENTER,				1,	 0,	 0,	AMMO,		FALSE},
	{HELP,				0,	 0,	 0,	TRAINER,	FALSE},
	{NUMERICSLASH,		0,	 0,	 0,	TRAINER,	FALSE},
	{NUMERICASTERISK,	0,	 0,	 0,	TRAINER,	FALSE},
	{NUMERICPLUS,		0,	 0,	 0,	TRAINER,	FALSE}
}; // Never leave unused keys in this structure.

// #DEFINES ---------------------------------------------------------------

#define ID_8SVX					MAKE_ID('8','S','V','X')
#define ID_BODY					MAKE_ID('B','O','D','Y')
#define ID_VHDR					MAKE_ID('V','H','D','R')
#define Unity					0x10000L	// Unity = Fixed 1.0 = maximum volume
#define PALCLOCK				3546895L
#define SOFT					32
#define LOUD					64			// = Unity

// PRIVATE STRUCTURES -----------------------------------------------------

#define MENUENTRIES			   19
struct NewMenu NewMenu[MENUENTRIES] =
{	{	NM_TITLE,	"Project",		 0 , 0,					0, 0},
	{	 NM_ITEM,	"New",			"N", 0,					0, 0},
	{	 NM_ITEM,	"Open...",		"O", 0,					0, 0},
	{	 NM_ITEM,	"Save",			"S", 0,					0, 0},
	{	 NM_ITEM,	"Save As...",	"A", 0,					0, 0},
	{	 NM_ITEM,	NM_BARLABEL,	 0 , 0,					0, 0},
	{	 NM_ITEM,	"About...",		"?", 0,					0, 0},
	{	 NM_ITEM,	NM_BARLABEL,	 0 , 0,					0, 0},
	{	 NM_ITEM,	"Quit",			"Q", 0,					0, 0},
	{   NM_TITLE,	"Edit",			 0 , NM_MENUDISABLED,	0, 0},
	{    NM_ITEM,	"Cut",			"X", 0,					0, 0},
	{    NM_ITEM,	"Copy",			"C", 0,					0, 0},
	{    NM_ITEM,	"Paste",		"V", 0,					0, 0},
	{	NM_TITLE,	"Level",		 0 , NM_MENUDISABLED,	0, 0},
	{	 NM_ITEM,	"Insert",		"I", 0,					0, 0},
	{	 NM_ITEM,	"Delete",		"D", 0,					0, 0},
	{	 NM_ITEM,	"Erase",		"E", 0,					0, 0},
	{	 NM_ITEM,	"Append",		"K", 0,					0, 0},
	{	  NM_END,	NULL,			 0 , 0,					0, 0}
};

struct GamePortTrigger Trigger =
{	GPTF_DOWNKEYS,
	65535,
	1, 1
};
struct NewGadget CycleGadget =
{	360, 0,
	128, 13,
	ILLEGAL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};
struct NewGadget CheckboxGadget =
{	250, 234,
	0, 0,
	"_Isometric 3D?",
	NULL,
	NULL,
	NULL,
	NULL
};
struct NewGadget StringGadget =
{	235, 0,
	181, 12,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};
struct
{	STRPTR		filename;
	BYTE		volume;
} samp[SAMPLES + 1] =
{   {     "PROGDIR:fx/goatfire.8svx",   SOFT    },  // FXGOATFIRE
	{"PROGDIR:fx/protectorborn.8svx",	SOFT	},	// FXPROTECTORBORN
	{ "PROGDIR:fx/timebombtick.8svx",	LOUD	},	// FXTIMEBOMBTICK
	{    "PROGDIR:fx/getgrower.8svx",	SOFT	},	// FXGETGROWER
	{"PROGDIR:fx/missileactive.8svx",	LOUD	},	// FXMISSILEACTIVE
	{   "PROGDIR:fx/getpowerup.8svx",	LOUD	},	// FXGETPOWERUP
	{  "PROGDIR:fx/useteleport.8svx",	SOFT	},	// FXUSETELEPORT
	{    "PROGDIR:fx/wormdeath.8svx",	LOUD	},	// FXWORMDEATH
	{      "PROGDIR:fx/getammo.8svx",	LOUD	},	// FXGETAMMO
	{      "PROGDIR:fx/explode.8svx",	SOFT	},	// FXEXPLODE
    {     "PROGDIR:fx/goatborn.8svx",   SOFT    },  // FXGOATBORN
	{      "PROGDIR:fx/orbborn.8svx",	SOFT	},	// FXORBBORN
	{        "PROGDIR:fx/shoot.8svx",	SOFT	},	// FXSHOOT
	{  "PROGDIR:fx/titlescreen.8svx",	SOFT	},	// FXTITLESCREEN
	{      "PROGDIR:fx/victory.8svx",	SOFT	},	// FXVICTORY
	{      "PROGDIR:fx/filenew.8svx",	LOUD	},	// FXFILENEW
	{     "PROGDIR:fx/fileopen.8svx",	LOUD	},	// FXFILEOPEN
	{     "PROGDIR:fx/filesave.8svx",	LOUD	},	// FXFILESAVE
	{    "PROGDIR:fx/bombblast.8svx",	SOFT	},	// FXBOMBBLAST
	{       "PROGDIR:fx/defeat.8svx",	LOUD	},	// FXDEFEAT
	{   "PROGDIR:fx/endoflevel.8svx",	LOUD	},	// FXENDOFLEVEL
	{        "PROGDIR:fx/click.8svx",	SOFT	},	// FXCLICK
	{    "PROGDIR:fx/usearmour.8svx",	SOFT	},	// FXUSEARMOUR
	{    "PROGDIR:fx/getobject.8svx",	SOFT	},	// FXGETOBJECT
	{    "PROGDIR:fx/timealert.8svx",	SOFT	},	// FXTIMEALERT
	{     "PROGDIR:fx/orbsplit.8svx",	SOFT	},	// FXORBSPLIT
	{     "PROGDIR:fx/orbdeath.8svx",	SOFT	},	// FXORBDEATH
	{   "PROGDIR:fx/filesaveas.8svx",	LOUD	},	// FXFILESAVEAS
	{           "PROGDIR:fx/fx.8svx",	SOFT	},	// FXFX
	{     "PROGDIR:fx/getnitro.8svx",	LOUD	},	// FXGETNITRO
	{         "PROGDIR:fx/jump.8svx",	SOFT	},	// FXJUMP
	{            "PROGDIR:fx/C.8svx",	LOUD	},	// FX_C
	{            "PROGDIR:fx/O.8svx",	LOUD	},	// FX_O
	{            "PROGDIR:fx/M.8svx",	LOUD	},	// FX_M
	{            "PROGDIR:fx/P.8svx",	LOUD	},	// FX_P
	{            "PROGDIR:fx/L.8svx",	LOUD	},	// FX_L
	{            "PROGDIR:fx/E.8svx",	LOUD	},	// FX_E
	{            "PROGDIR:fx/T.8svx",	LOUD	},	// FX_T
	{    "PROGDIR:fx/highscore.8svx",	SOFT	},	// FXHISCORE
	{         "PROGDIR:fx/thud.8svx",	SOFT	},	// FXTHUD
    {    "PROGDIR:fx/goatdeath.8svx",   SOFT    },  // FXGOATDEATH
	{     "PROGDIR:fx/getskull.8svx",	SOFT	},	// FXGETSKULL
	{    "PROGDIR:fx/usetongue.8svx",	SOFT	},	// FXUSETONGUE
	{     "PROGDIR:fx/dripborn.8svx",	SOFT	},	// FXDRIPBORN
	{       "PROGDIR:fx/amigan.8svx",	SOFT	},	// FXAMIGAN
	{        "PROGDIR:fx/green.8svx",	SOFT	},	// FXPAIN
	{          "PROGDIR:fx/red.8svx",	SOFT	},
	{         "PROGDIR:fx/blue.8svx",	SOFT	},
	{       "PROGDIR:fx/yellow.8svx",	SOFT	},
	{  "PROGDIR:fx/penguinborn.8svx",	SOFT	},	// FXPENGUINBORN
	{ "PROGDIR:fx/penguindeath.8svx",	SOFT	},	// FXPENGUINDEATH
	{      "PROGDIR:fx/rundown.8svx",	SOFT	},	// FXRUNDOWN
	{      "PROGDIR:fx/editcut.8svx",	SOFT	},	// FXEDITCUT
	{     "PROGDIR:fx/editcopy.8svx",	SOFT	},	// FXEDITCOPY
	{    "PROGDIR:fx/editpaste.8svx",	SOFT	},	// FXEDITPASTE
	{  "PROGDIR:fx/leveldelete.8svx",	SOFT	}	// FXLEVELDELETE
};

// ARRAYS -----------------------------------------------------------------

MODULE STRPTR CycleText[] =
{	"Worm _1:",
	"Worm _2:",
	"Worm _3:",
	"Worm _4:"
};
MODULE UBYTE* CycleOptions[4][4] =
{ {	(UBYTE *) "None",
	(UBYTE *) "Amiga",
	(UBYTE *) "Lt. Kybd",
	NULL
  },
  {	(UBYTE *) "None",
	(UBYTE *) "Amiga",
	(UBYTE *) "Rt. Kybd",
	NULL
  },
  {	(UBYTE *) "None",
	(UBYTE *) "Amiga",
	(UBYTE *) "Joystick 2",
	NULL
  },
  {	(UBYTE *) "None",
	(UBYTE *) "Amiga",
	(UBYTE *) "Joystick 1",
	NULL
} };

MODULE STRPTR sfxerror[] =
{	"No errors.",
	"Can't open file!",
	"Can't read file!",
	"Not an IFF 8SVX; too short!",
	"Not an IFF FORM!",
	"No memory for read!",
	"Read error!",
	"Malformed IFF; too short!",
	"Not an IFF 8SVX!",
	"No chip memory!"
};

// PUBLIC STRUCTURES -----------------------------------------------------

struct Chunk
{	LONG ckID;
	LONG ckSize;
	LONG ckType;
	void* ckData;
	struct Chunk* ckNext;
};

// TYPEDEFS ---------------------------------------------------------------

typedef LONG Fixed;				/* A fixed-point value, 16 bits to the left
								of the point and 16 to the right. A Fixed
								is a number of 2**16ths, ie. 65536ths. */
typedef struct {
	ULONG oneShotHiSamples,		// # samples in the high octave 1-shot part
		repeatHiSamples,		// # samples in the high octave repeat part
		samplesPerHiCycle;		// # samples/cycle in high octave, else 0
	UWORD samplesPerSec;		// data sampling rate
	UBYTE ctOctave,				// # of octaves of waveforms
		sCompression;			// data compression technique used
	Fixed volume;				/* playback nominal volume from 0 to Unity
								(full volume). Map this value into the
								output hardware's dynamic range (0-64). */
} Voice8Header;

// Must have blank line at EOF.
